/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				JM_MANAGE.C												*/
/*																		*/
/*																		*/
/************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include "../inc/types.h"
#ifdef AIX
#include <sys/select.h>
#endif
#include <fcntl.h>
#include <errno.h>
#include <sys/ipc.h>
#include <sys/shm.h>
#include "../inc/time.h"

extern int pm_sock;
extern int suppress_errors;

static fd_set fdset;


/******************************************************************/
/*
	FUNCTION NAME:

	  send_to_user

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_to_user(lpn,m)
int lpn;
cps_message *m;
{
	if (send_binary(shm_map[usr_map[lpn-FIRST_USER_LPN].shm].sock,
		m,4*sizeof(long)+m->size) == -1)
	{
		jm_dead_shm_found(usr_map[lpn-FIRST_USER_LPN].shm,1);
	}
}

send_to_shm(shm_lpn,m)
int shm_lpn;
cps_message *m;
{
	if (send_binary(shm_map[shm_lpn-FIRST_SHM_LPN].sock,
		m,4*sizeof(long)+m->size) == -1)
	{

		jm_dead_shm_found(shm_lpn-FIRST_SHM_LPN,2);

	}
}

shm_broadcast(m)
cps_message *m;
{
	int i;
	for (i=0; i<num_shm_procs; i++)
	{
		if (shm_map[i].status)
		{
			m->dst_lpn = i+FIRST_SHM_LPN;
			send_to_shm(i+FIRST_SHM_LPN,m);
		}
	}
}

set_broadcast(s,size,m)
unsigned long *s;
int size;
cps_message *m;
{
	int i,*p;
	p = set_query(s,size);
	i = 0;
	while (p[i] != 0)
	{
		m->dst_lpn = p[i];

		send_to_user(p[i],m);
		i++;
	}
}

usr_broadcast(m)
cps_message *m;
{
	int c;
	m->src_lpn = 0;
	for (c=1; c<=MAX_CLASSES; c++) {
		if (class[c].valid) {
			set_broadcast(class[c].alive,MAX_PROCESSES,m);
		}
	}
}

shm_usr_broadcast(m)
cps_message *m;
{
	int c;
	for( c = 0; c < num_shm_procs; c++ ) {
		if (shm_map[c].status) {
			m->dst_lpn = -1;
			send_to_shm( c + FIRST_SHM_LPN, m );
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  check_queue_full

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

check_queue_full(q)
int q;
{
	int s1,s2;

	int i,*lpn;

	cps_message m;

	if (set_size(jmq[q].wait_full,MAX_PROCESSES) == 0)
		return;

	if (jmq[q].eoq == 0) {
		s1 = set_size(jmq[q].declared,MAX_PROCESSES);
		s2 = set_size(jmq[q].q_set,MAX_PROCESSES);
	}

	if (jmq[q].eoq || (s1 == s2)) {
		lpn = set_query(jmq[q].wait_full,MAX_PROCESSES);
		i = 0;
		while (lpn[i] != 0) {
			cps_bld_wait_complete(&m,0,lpn[i]);
			send_to_user(lpn[i],&m);
			mon_shm->p_state[lpn[i]-FIRST_USER_LPN].state = PSTATE_RUNNING;
			i++;
		}
		set_clear(jmq[q].wait_full,MAX_PROCESSES);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  check_queue_empty

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

check_queue_empty(q)
int q;
{
	int i,*lpn;

	cps_message m;

	if (set_size(jmq[q].wait_empty,MAX_PROCESSES) == 0)
		return;

	if (set_size(jmq[q].q_set,MAX_PROCESSES) == 0) {
		lpn = set_query(jmq[q].wait_empty,MAX_PROCESSES);
		i = 0;
		while (lpn[i] != 0) {
			cps_bld_wait_complete(&m,0,lpn[i]);
			send_to_user(lpn[i],&m);
			mon_shm->p_state[lpn[i]-FIRST_USER_LPN].state = PSTATE_RUNNING;
			i++;
		}
		set_clear(jmq[q].wait_empty,MAX_PROCESSES);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  check_dequeue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

check_dequeue(q)
int q;
{

	int i,n,nq,ndq;

	struct cps_queue_element *e1,*e2,*e;

	cps_message m;

	nq = set_size(jmq[q].q_set,MAX_PROCESSES);
	if (nq == 0)
	{
		if (jmq[q].eoq)
			goto end_of_queue;
		else
			return;
	}

	while (1)
	{
		nq = set_size(jmq[q].q_set,MAX_PROCESSES);
		if (nq == 0)
		{
			if (jmq[q].eoq)
				goto end_of_queue;
			else
				return;
		}

		ndq = set_size(jmq[q].dq_set,MAX_PROCESSES);
		if (ndq == 0)
			return;

/* JMDB BREAKPOINT */
		jm_rpc_breakpoint( queue_stop[q], q, &queue_point );

		e1 = cps_remove_q(jmq[q].queued);
		e2 = cps_remove_q(jmq[q].dequeue);
		if ((e1 == NULL) || (e2 == NULL))
		{
			printerr(FACILITY,FATAL_MSG,"Process queues corrupt",NULL,11);
			abort_job(1);
		}
		remove_from_set(e1->lpn,jmq[q].q_set);

/* Remove the dequeueing lpn from the set only if it has no other dequeue
   requests */

		e = jmq[q].dequeue->front;
		while (e != NULL)
		{
			if (e->lpn == e2->lpn)
				break;
			e = e->next;
   		}
		if (e == NULL)
			remove_from_set(e2->lpn,jmq[q].dq_set);

/* An attribute value of zero indicates that a normal acp_dequeue_process was
   called.  Non-zero values are the id number of an asynchronous dequeue
   request. */

		if (e2->attr == 0)
		{
			cps_bld_dequeue_reply(&m,0,e2->lpn,e1->lpn,q,jmq[q].bcount,e1->msg);
			send_to_user(e2->lpn,&m);
			mon_shm->p_state[e1->lpn-FIRST_USER_LPN].state = PSTATE_RUNNING;
		}
		else
		{

/* Only send data ready messages for asynchronous gets */

			if (e2->attr < 0)
				send_data_rdy_msgs(e1->lpn,e2->lpn);
			cps_bld_asynch_dq_reply(&m,0,e2->lpn,e1->lpn,q,e2->attr);
			send_to_user(e2->lpn,&m);
		}

		remove_from_set(e1->lpn,(unsigned long *) &mon_shm->qset[q]);
		cps_enter_q(zfree,e2);
		cps_enter_q(jmq[q].free,e1);

	}

	end_of_queue:

/* Check for end of queue with no processes left on queue */

	ndq = set_size(jmq[q].dq_set,MAX_PROCESSES);
	while (ndq > 0)
	{

/* JMDB BREAKPOINT */
		jm_rpc_breakpoint( queue_stop[q], q, &queue_point );

		e2 = cps_remove_q(jmq[q].dequeue);
		if (e2 == NULL)
		{
			printerr(FACILITY,FATAL_MSG,"Process queues corrupt",NULL,12);
			abort_job(2);
		}

/* Remove the dequeueing lpn from the set only if it has no other dequeue
   requests */

		e = jmq[q].dequeue->front;
		while (e != NULL)
		{
			if (e->lpn == e2->lpn) break;
			e = e->next;
   		}
		if (e == NULL)
			remove_from_set(e2->lpn,jmq[q].dq_set);

		if (e2->attr == 0)
		{
			cps_bld_dequeue_reply(&m,0,e2->lpn,CPS_END_OF_QUEUE,q,0,0);
			send_to_user(e2->lpn,&m);
		}
		else
		{
			cps_bld_asynch_dq_reply(&m,0,e2->lpn,CPS_END_OF_QUEUE,q,e2->attr);
			send_to_user(e2->lpn,&m);
		}
		cps_enter_q(zfree,e2);

/* Re-calculate the number of processes waiting to dequeue */

		ndq = set_size(jmq[q].dq_set,MAX_PROCESSES);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_dequeue_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_dequeue_process(lpn,q,wait_flag)
int lpn,q,wait_flag;
{
	cps_msg m;
	struct cps_queue_element *e;
	static int lpn_s, q_s, wait_flag_s;

	if ((wait_flag == 0) && (set_analyze(jmq[q].q_set,MAX_PROCESSES) == 0))
	{
/* Reply with CPS_EMPTY */
		cps_bld_dequeue_reply(&m,0,lpn,CPS_EMPTY,q,0,0);
		send_to_user(lpn,&m);
		return;
	}

/* Put the process on a waiting list (even if the queue has ended) */

	e = cps_alloc_qelem(zfree,0);
	e->attr = 0;
	e->lpn = lpn;
	cps_enter_q(jmq[q].dequeue,e);
	add_to_set(lpn,jmq[q].dq_set);

/* Check to see if it can be dequeued */

	check_dequeue(q);
	check_queue_empty(q);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_asynch_dq

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_asynch_dq(lpn,q,id)
int lpn,q,id;
{
	cps_msg m;

	struct cps_queue_element *e;

/* Put the process on a waiting list (even if the queue has ended) */

	e = cps_alloc_qelem(zfree,0);
	e->attr = id;
	e->lpn = lpn;
	cps_enter_q(jmq[q].dequeue,e);
	add_to_set(lpn,jmq[q].dq_set);

/* Check to see if it can be dequeued */

	check_dequeue(q);
	check_queue_empty(q);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_queue_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_queue_process(lpn,q,args)
int lpn,q;
char *args;
{
	struct cps_queue_element *e;
	char errmsg[132];
	static int lpn_s, q_s;
	char *args_s;

/* Make sure this process is not already on the queue */

	if (set_member(lpn,jmq[q].q_set))
	{
		sprintf(errmsg,"Process %d already in queue %d",lpn,q);
		printerr(FACILITY,WARNING_MSG,errmsg,NULL,13);
		return;
	}

/* Check to see if this process is allowed to be on the queue */

	if (!set_member(lpn,jmq[q].declared))
	{
		sprintf(errmsg,"Process %d not declared for queue %d",lpn,q);
		printerr(FACILITY,WARNING_MSG,errmsg,NULL,14);
		return;
	}

/* Make sure the queue hasn't ended */

	if (jmq[q].eoq == -1)
	{
		sprintf(errmsg,
			"Attempted to place process %d onto queue %d after end-of-queue",
			lpn,q);
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,15);
		jm_dead_proc_found(lpn);
		return;
	}

/* JMDB BREAKPOINT */
	jm_rpc_breakpoint( queue_stop[q], q + MAX_QUEUES, &queue_point );

/* Queue the process */

	e = cps_alloc_qelem(jmq[q].free,jmq[q].bcount);
	e->lpn = lpn;
	bcopy(args,e->msg,jmq[q].bcount);
	cps_enter_q(jmq[q].queued,e);
	add_to_set(lpn,jmq[q].q_set);

/* Update monitors */

	mon_shm->p_state[lpn-FIRST_USER_LPN].state = PSTATE_QUEUED;
	mon_shm->p_state[lpn-FIRST_USER_LPN].attrib = q;
	add_to_set(lpn,(unsigned long *) &mon_shm->qset[q]);

/* Check for waiting processes */

	check_dequeue(q);
	check_queue_full(q);

}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_declare_queue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_declare_queue(lpn,q,n,b)
int lpn,q,n,*b;
{
	char errmsg[132];
	int not_warned = -1;

	int bcount;

	int i;

/* Declare the queue */

	if (jmq[q].active == 0) {
		not_warned = 0;
		jmq[q].active = -1;
	}

	add_to_set(lpn,jmq[q].declared);

	if (n != jmq[q].nargs) {
		if (not_warned) {
			sprintf(errmsg,
			"Process %d declared queue %d with different arguments",
			lpn,q);
			printerr(FACILITY,WARNING_MSG,errmsg,NULL,16);
			not_warned = 0;
		}
		if (n > jmq[q].nargs)
			jmq[q].nargs = n;
	}

	bcount = 0;
	for (i=0; i<n; i++) {
		if (not_warned && (jmq[q].argcnt[i] != b[i])) {
			sprintf(errmsg,
				"Process %d declared queue %d with different arguments",lpn,q);
			printerr(FACILITY,WARNING_MSG,errmsg,NULL,17);
			not_warned = 0;
		}
		if (b[i] > jmq[q].argcnt[i]) {
			bcount += b[i];
			jmq[q].argcnt[i] = b[i];
		}
		else {
			bcount += jmq[q].argcnt[i];
		}
	}
	jmq[q].bcount = bcount;

}

/************************************************************************/
/*

	Function: send_data_rdy_msgs

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int send_data_rdy_msgs(src_lpn,dst_lpn)
int src_lpn,dst_lpn;
{
	int i,j;
	cps_message m;
	struct acp_data_rdy *d;

	i = src_lpn - FIRST_USER_LPN;

	d = acp_dblocks[i];
	while (d != NULL) {
		if (d->valid) {
			cps_bld_data_ready(&m,0,dst_lpn,src_lpn,d->blockno,d->bcount);
			send_to_user(dst_lpn,&m);
			d->valid = 0;
		}
		d = d->next;
	}
}

/************************************************************************/
/*

	Function: handle_data_ready

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

static int handle_data_ready(lpn,blockno,bcount)
int lpn,blockno,bcount;
{
	int i;

	struct acp_data_rdy *d;

	i = lpn - FIRST_USER_LPN;

/* Find the record for this block number */

	d = acp_dblocks[i];
	while (d != NULL) {
		if (d->blockno == blockno) break;
		d = d->next;
	}
	if (d == NULL) {
		d = (struct acp_data_rdy *) hmalloc(sizeof(struct acp_data_rdy));
		d->blockno = blockno;
		d->next = acp_dblocks[i];
		acp_dblocks[i] = d;
	}

/* Store the byte count and set the valid flag */

	d->bcount = bcount;
	d->valid = -1;	/* This record will be sent as a DATA_READY message
			   to whatever process does an asynchronous dq & get */
}

/******************************************************************/
/*
	FUNCTION NAME:

	  abort_job

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

abort_job(x)
int x;
{

	cps_message msg;

	int i,pid;

	char shell_cmd[80];

/* Tell each SHM process to shutdown */

	for (i=0; i<num_shm_procs; i++) {
		if (shm_map[i].status) {
			cps_bld_shm_shutdown(&msg,0,i+FIRST_SHM_LPN);
			send_binary(shm_map[i].sock,&msg,4*sizeof(long)+msg.size);
		}
	}

/* Remove the shared memory */

	mon_shm->job_done = -1;
	shmctl(mon_shmid,IPC_RMID,NULL);

/* Wait for our parent to die */

	while(1) {
		pid = getppid();
		if (pid == 1)
			break;
		sleep(1);
	}

	jm_exit();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  check_sync

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

check_sync(sync_num)
int sync_num;
{
	int s1,s2;

	int i;

	cps_message m;

/* Check to see if the condition has been met */

	s1 = set_size(need_sync[sync_num],MAX_PROCESSES);
	s2 = set_size(at_sync[sync_num],MAX_PROCESSES);
	if (s1 != s2) return;

/* Send a request complete message to each process */

	for (i=0; i<num_usr_procs; i++)
	{
		if (proc_sync[i] == sync_num)
		{
			cps_bld_sync_complete(&m,0,i+FIRST_USER_LPN);
			send_to_user(i+FIRST_USER_LPN,&m);
			mon_shm->p_state[i].state = PSTATE_RUNNING;
			proc_sync[i] = -1;
		}
	}

/* Indicate that the sync number is no longer active */

	sync_active[sync_num] = 0;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_wait_sync

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_wait_sync(lpn,sync_num,nclass,c)
int lpn,sync_num,nclass,*c;
{
	int sync_class[MAX_CLASSES];

	int n;
	int *cl;

	int i,j;

	int nump,p[MAX_PROCESSES],*p1;

/* If nclass is zero, then synchronize with all processes */

	n = nclass;
	if (n == 0) 
	{
		for (i=1; i<=MAX_CLASSES; i++)
		{
			if (class[i].valid)
			{
				sync_class[n] = i;
				n++;
			}
		}
		cl = sync_class;
	}
	else
	{
		cl = c;
	}

/* Build an array of processes in the set */

	nump = 0;
	for (i=0; i<n; i++)
	{
		p1 = set_query(class[cl[i]].alive,MAX_PROCESSES);
		j = 0;
		while (p1[j] != 0)
		{
			p[nump++] = p1[j++];
		}
	}

/* Make sure the sync sets are allocated */

	if (need_sync[sync_num] == NULL)
	{
		need_sync[sync_num] = create_empty_set(MAX_PROCESSES);
		at_sync[sync_num] = create_empty_set(MAX_PROCESSES);
	}

/* Check to see if this is the first process reaching the sync point */

	if (!sync_active[sync_num])
	{
		set_clear(need_sync[sync_num],MAX_PROCESSES);
		set_clear(at_sync[sync_num],MAX_PROCESSES);
		sync_active[sync_num] = -1;
	}

/* Add processes needed to complete the sync */

	for (i=0; i<nump; i++)
		add_to_set(p[i],need_sync[sync_num]);

/* Record the sync number this process is waiting for */

	proc_sync[lpn-1] = sync_num;
	if (set_member(lpn,need_sync[sync_num]))
		add_to_set(lpn,at_sync[sync_num]);

/* Check to see if the sync point has been satisfied */

	check_sync(sync_num);

/* Check for a possible deadlock */
/*
	check_deadlock(lpn);
*/
}

/******************************************************************/
/*
	FUNCTION NAME:

	  add_remote_sub

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

add_remote_sub(lpn,rsub)
int lpn;
struct acp_remote_sub *rsub;
{
	struct acp_remote_sub *r;

	r = remote_subs;
	while (r != NULL) {
		if (r->id == rsub->id) return;
		r = r->next;
	}

	r = (struct acp_remote_sub *) hmalloc(sizeof(struct acp_remote_sub));
	*r = *rsub;
	r->next = remote_subs;
	remote_subs = r;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_rsub_decl

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_rsub_decl(lpn,id)
int lpn,id;
{
	struct acp_remote_sub *r;

	cps_msg m;

/* Find the subroutine */

	r = remote_subs;
	while (r != NULL) {
		if (r->id == id) break;
		r = r->next;
	}

/* Send the declaration if found */

	if (r != NULL) {
		cps_bld_sub_decl(&m,0,lpn,r);
		send_to_user(lpn,&m);
	}
	else {
/* Send a request complete message */
		cps_bld_rsub_complete(&m,0,lpn);
		send_to_user(lpn,&m);
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_queue_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_queue_info(lpn,q)
int lpn,q;
{
	cps_message m;

/* Send the declaration if found */

	if (jmq[q].active) {
		cps_bld_declare_queue(&m,0,lpn,q,jmq[q].nargs,jmq[q].argcnt);
		send_to_user(lpn,&m);
	}
	else {
/* Send a request complete message */
		cps_bld_req_complete(&m,0,lpn);
		send_to_user(lpn,&m);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_dead_shm_found

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_dead_shm_found(i,x)
int i;
int x;
{
	int *p;

	int j,k;

/* Mark the process as unconnected */

	shm_map[i].status = 0;
/* Added 4/26/93 */
	FD_CLR(shm_map[i].sock,&fdset);
	socket_close(shm_map[i].sock);

/* For each process local to that shm, if it was alive, declare it dead.
   If this routine was called as a result of the shared memory manager telling
   us he is shutting down because no more processes are alive on his CPU,
   skip the check. */
/*
	if (x == 4) goto label1;
*/
	/* can't assume that the shms user-processes are dead
	if (x == 4)
		return 0;
	if (x == 3)
		return 0;
	*/

	p = set_query(shm_map[i].proc,MAX_PROCESSES);
	j = 0;
	while (p[j] != 0) {
	k = usr_map[p[j]-1].class;
	if (set_member(p[j],class[k].alive))
		jm_dead_proc_found(p[j]);

/* Re-build set of processes in case it was corrupted 
	(bug found by Charles Chang & Lee Roberts at SSC) */

		p = set_query(shm_map[i].proc,MAX_PROCESSES);

		j++;
	}	
/*
	label1:
	FD_CLR(shm_map[i].sock,&fdset);
*/
}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_dead_proc_found

  DESCRIPTION:

	  The JM has learned that a user process is no longer alive.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_dead_proc_found(lpn)
int lpn;
{

	int i;

	int c,shm;

	char errmsg[132];

	cps_message m;

	int s1,s2;

/* Log that the process has died */

	sprintf(errmsg,"Process %d has been declared dead",lpn);
	printerr(FACILITY,WARNING_MSG,errmsg,NULL,18);

/* Register the dead process */

	c = usr_map[lpn-FIRST_USER_LPN].class;
	add_to_set(lpn,class[c].dead);
	remove_from_set(lpn,class[c].alive);

/* Check to see if the job should be stopped */

	s1 = set_size(class[c].alive,MAX_PROCESSES);
	s2 = set_size(class[c].fini,MAX_PROCESSES);
	if ((s1+s2) < class[c].min_proc)
	{
		sprintf(errmsg,"Too few processes alive in class %d to continue",c);
		printerr(FACILITY,WARNING_MSG,errmsg,NULL,19);
		abort_job(3);
	}

/* Notify all shm processes that the user process is dead */

	cps_bld_usr_proc_dead(&m,0,0,lpn,"death notice");
	shm_broadcast(&m);

/* Notify all user processes of the dead process */

	shm_usr_broadcast(&m);

/* Check to see if a sync point has been satisfied */

	proc_sync[lpn-1] = -1;

	for (i=0; i<MAX_SYNCS; i++)
	{
		if (sync_active[i])
		{
			if (set_member(lpn,need_sync[i]))
			{
				remove_from_set(lpn,need_sync[i]);
				check_sync(i);
			}
		}
	}

/* Handle queues */

	for (i=0; i<MAX_QUEUES; i++)
	{
		if (jmq[i].active)
		{
			remove_from_set(lpn,jmq[i].declared);
			remove_from_set(lpn,jmq[i].wait_full);
			remove_from_set(lpn,jmq[i].wait_empty);
			if (set_member(lpn,jmq[i].dq_set))
			{
				remove_from_set(lpn,jmq[i].dq_set);
				cps_extract_from_queue(lpn,jmq[i].dequeue,zfree,zfree);
			}
			if (set_member(lpn,jmq[i].q_set))
			{		/* Queue args lost */
				remove_from_set(lpn,jmq[i].q_set);
				cps_extract_from_queue(lpn,jmq[i].queued,jmq[i].free,zfree);
			}
			check_queue_full(i);
			check_queue_empty(i);
		}
	}

/* Update the monitors */

	mon_shm->p_state[lpn-FIRST_USER_LPN].state = PSTATE_DEAD;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  jm_proc_finished

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

jm_proc_finished(lpn)
int lpn;
{

	int i;

	int c,shm;

	char errmsg[132];

	cps_message m;

	int x;

/* Log that the process has finished */

	sprintf(errmsg,"Process %d has finished",lpn);
	printerr(FACILITY,WARNING_MSG,errmsg,NULL,20);

/* Register the finished process */

	c = usr_map[lpn-FIRST_USER_LPN].class;
	add_to_set(lpn,class[c].fini);
	remove_from_set(lpn,class[c].alive);

/* Notify all shm processes that the user process is dead */

	cps_bld_usr_proc_dead(&m,0,0,lpn,"death notice");
	x = suppress_errors;
	suppress_errors = -1;
	shm_broadcast(&m);
	suppress_errors = x;

/* Check to see if a sync point has been satisfied */

	proc_sync[lpn-1] = -1;

	for (i=0; i<MAX_SYNCS; i++)
	{
		if (sync_active[i])
		{
			if (set_member(lpn,need_sync[i]))
			{
				remove_from_set(lpn,need_sync[i]);
				check_sync(i);
			}
		}
	}

/* Handle queues */

	for (i=0; i<MAX_QUEUES; i++)
	{
		if (jmq[i].active)
		{
			remove_from_set(lpn,jmq[i].declared);
			remove_from_set(lpn,jmq[i].wait_full);
			remove_from_set(lpn,jmq[i].wait_empty);
			if (set_member(lpn,jmq[i].dq_set))
			{
				remove_from_set(lpn,jmq[i].dq_set);
				cps_extract_from_queue(lpn,jmq[i].dequeue,zfree,zfree);
			}
			if (set_member(lpn,jmq[i].q_set))
			{	/* Queue args lost */
				remove_from_set(lpn,jmq[i].q_set);
				cps_extract_from_queue(lpn,jmq[i].queued,jmq[i].free,zfree);
			}
			check_queue_full(i);
			check_queue_empty(i);
		}
	}

/* Update the monitors */

	mon_shm->p_state[lpn-FIRST_USER_LPN].state = PSTATE_FINISHED;

}

/******************************************************************/
/*
	FUNCTION NAME:

	  process_message

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int process_message(i)
int i;
{
	cps_message m;

	int type,src_lpn,dst_lpn,size;

	char *msg;

	int s;

/* Read the message from the socket */

	s = read_binary(shm_map[i].sock,(char *)&m,
		4*sizeof(long)+MAX_MESSAGE_SIZE);

	if (s == -1)
	{
/* Lost connection with the SHM process */
		jm_dead_shm_found(i,3);
		return;
	}
	else if (s == -2)
	{
		return;
	}

	type = (int) m.type;
	src_lpn = (int) m.src_lpn;
	dst_lpn = (int) m.dst_lpn;
	size = (int) m.size;
	msg = (char *) &(m.msg[0]);

/* If the destination LPN is not the JM, send the message to the appropriate
   shared memory manager. This only occurs when SHMs cannot talk directly
   with each other (usually never). */

	if (dst_lpn >= FIRST_SHM_LPN)
	{
		printf("Redirecting message to lpn %d\n",dst_lpn);
		send_to_shm(dst_lpn,&m);
		return(0);
	}
	else if (dst_lpn >= FIRST_USER_LPN)
	{
		printf("Redirecting message to lpn %d\n",dst_lpn);
		send_to_user(dst_lpn,&m);
		return(0);
	}

#ifdef DIAG
	log_cps_msg(jml,&m);
	fflush(jml);
#endif

/* store message for jmdb */
	if( jm_pause )
		jmdb_store_message( &m );
/* jm monitor message */
	if( jm_prof )
		jm_profile_log( src_lpn, type );
/* print message to upshot file */
	if( upshot )
		upshot_print_event( &m );

	switch (type) {

	case MSG_MON_UPDATE :
		{
			int opcode,lpn,datum1,datum2;
			cps_get_mon_update(&m,&opcode,&lpn,&datum1,&datum2);
			monitor_update(opcode,lpn,datum1,&datum2);
		}
		break;
	case MSG_SHM_ERROR :
		{
			cps_get_shm_error(&m,&msg);
			print_shm_log(src_lpn-FIRST_SHM_LPN,msg);
		}
		break;
	case MSG_SHM_SHUTDOWN :
		{
			jm_dead_shm_found(src_lpn-FIRST_SHM_LPN,4);
		}
		break;
	case MSG_USR_LOG :
		{
			cps_get_usr_log(&m,&msg);
			print_usr_log(src_lpn,msg);
		}
		break;
	case MSG_USR_ABORT_JOB :
		{
			cps_get_abort_job(&m,&msg);
			print_usr_log(src_lpn,msg);
			abort_job(4);
		}
		break;
	case MSG_USR_PROC_DEAD :
		{
			int lpn;
			char *msg;
			cps_get_usr_proc_dead(&m,&lpn,&msg);
			if (src_lpn > MAX_PROCESSES)
				print_shm_log(src_lpn-FIRST_SHM_LPN,msg);
			else
				print_usr_log(src_lpn,msg);
			jm_dead_proc_found(lpn);
		}
		break;
	case MSG_USR_FINISHED :
		{
			int lpn;
			char *msg;
			cps_get_usr_finished(&m,&lpn,&msg);
			print_usr_log(src_lpn,msg);
			jm_proc_finished(lpn);
		}
		break;
	case MSG_DECLARE_QUEUE :
		{
			int q,n,*b;
			cps_get_declare_queue(&m,&q,&n,&b);
			jm_declare_queue(src_lpn,q,n,b);
			mon_shm->q_declared[q] = -1;
		}
		break;
	case MSG_QUEUE_PROCESS :
		{
			int lpn,q;
			char *args;
			cps_get_queue_process(&m,&lpn,&q,&args);
			jm_queue_process(lpn,q,args);
			mon_shm->call_cnt[src_lpn-FIRST_USER_LPN][INC_QUEUE]++;
		}
		break;
	case MSG_DEQUEUE_PROCESS :
		{
			int q,wait_flag;
			cps_get_dequeue_process(&m,&q,&wait_flag);
			if (wait_flag)
			{
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
					PSTATE_WAIT_DEQUEUE;
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].attrib = q;
			}
			jm_dequeue_process(src_lpn,q,wait_flag);
			mon_shm->call_cnt[src_lpn-FIRST_USER_LPN][INC_DEQUEUE]++;
		}
		break;
	case MSG_DATA_READY :
		{
			int lpn,blockno,bcount;
			cps_get_data_ready(&m,&lpn,&blockno,&bcount);
			handle_data_ready(lpn,blockno,bcount);
		}
		break;
	case MSG_END_OF_QUEUE :
		{
			int q;
			cps_get_end_of_queue(&m,&q);
			jmq[q].eoq = -1;
			check_dequeue(q);
			check_queue_full(q);
			m.src_lpn = (long) 0;
			shm_usr_broadcast(&m);
		}
		break;
	case MSG_DECLARE_BLOCK :
		{
			shm_usr_broadcast(&m);
		}
		break;
	case MSG_WAIT_QUEUE :
		{
			int q,state;
			cps_get_wait_queue(&m,&q,&state);
			if (state == CPS_EMPTY)
			{
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
					PSTATE_WAIT_QUEUE_EMPTY;
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].attrib = q;
				add_to_set(src_lpn,jmq[q].wait_empty);
				check_queue_empty(q);
			}
			else
			{
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
					PSTATE_WAIT_QUEUE_FULL;
				mon_shm->p_state[src_lpn-FIRST_USER_LPN].attrib = q;
				add_to_set(src_lpn,jmq[q].wait_full);
				check_queue_full(q);
			}
			mon_shm->call_cnt[src_lpn-FIRST_USER_LPN][INC_WAIT_QUEUE]++;
		}
		break;
	case MSG_WAIT_SYNC :
		{
			int sync_num,nclass,*c;
			cps_get_wait_sync(&m,&sync_num,&nclass,&c);
/* JMDB BREAKPOINT */
			jm_rpc_breakpoint( sync_stop[ sync_num ], sync_num, &sync_point );
			mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = PSTATE_WAIT_SYNC;
			mon_shm->p_state[src_lpn-FIRST_USER_LPN].attrib = sync_num;
			jm_wait_sync(src_lpn,sync_num,nclass,c);
			mon_shm->call_cnt[src_lpn-FIRST_USER_LPN][INC_SYNC]++;
		}
		break;
	case MSG_USER_STATUS :
		{
			int i,len,*data;
			cps_get_user_status(&m,&len,&data);
			for (i=0; i<len; i++)
				user_status[src_lpn-FIRST_USER_LPN][i] = data[i];
			for (i=0; i<len; i++)
				mon_shm->user_status[src_lpn-FIRST_USER_LPN][i] = data[i];
		}
		break;
	case MSG_USER_HOOK :
		{
			int pt;
			cps_get_hook(&m,&pt);
/* JMDB BREAKPOINT only for valid points */
			if( pt < MAX_USER_HOOKS )
				jm_rpc_breakpoint( user_stop[ pt ], pt, &user_point );
			cps_bld_hook_complete(&m,0,src_lpn);
			send_to_user(src_lpn,&m);
		}
		break;
	case MSG_FUNC_HOOK :
		{
			int pt;
			cps_get_hook(&m,&pt);
/* JMDB BREAKPOINT only for valid points */
			if( pt < MAX_FUNC_HOOKS )
				jm_rpc_breakpoint( proc_stop[ pt ], pt, &proc_point );
			cps_bld_hook_complete(&m,0,src_lpn);
			send_to_user(src_lpn,&m);
		}
		break;
	case MSG_ASYNCH_DQ :
		{
			int q,id;
			cps_get_asynch_dq(&m,&q,&id);
			jm_asynch_dq(src_lpn,q,id);
			mon_shm->call_cnt[src_lpn-FIRST_USER_LPN][INC_DEQUEUE]++;
		}
		break;
	case MSG_SUB_DECL :
		{
			struct acp_remote_sub *rsub;
			cps_get_sub_decl(&m,&rsub);
			add_remote_sub(src_lpn,rsub);
		}
		break;
	case MSG_REQ_SUB_DECL :
		{
			int id;
			cps_get_req_rsub_decl(&m,&id);
			send_rsub_decl(src_lpn,id);
		}
		break;
	case MSG_REQ_Q_INFO :
		{
			int q;
			cps_get_req_q_info(&m,&q);
			send_queue_info(src_lpn,q);
		}
		break;
	case MSG_GET_NEXT_TAPE :
		{
			int listno;
			mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
				PSTATE_WAIT_MOUNT;
			cps_get_get_next_tape(&m,&listno);
			send_list_mount_request(src_lpn,listno);
		}
		break;
	case MSG_GET_SPEC_TAPE :
		{
			int devtyp,ring;
			char *vsn;
			mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
				PSTATE_WAIT_MOUNT;
			cps_get_get_spec_tape(&m,&devtyp,&ring,&vsn);
			send_spec_mount_request(src_lpn,devtyp,ring,vsn);
		}
		break;
	case MSG_DONE_WITH_TAPE :
		{
			char *dev;
			cps_get_done_with_tape(&m,&dev);
			handle_done_with_tape(src_lpn,dev);
		}
		break;
	case MSG_MOUNT_REQUEST :
		{
			char *devnam,*comment,*vsn;
			int write_enabled;
			mon_shm->p_state[src_lpn-FIRST_USER_LPN].state = 
				PSTATE_WAIT_MOUNT;
			cps_get_mount_request(&m,&devnam,&vsn,&write_enabled,&comment);
			send_mount_request(src_lpn,devnam,vsn,write_enabled,comment);
		}
		break;
#ifdef R3000
	case MSG_VBBC_INFO :
		{
			shm_usr_broadcast(&m);
		}
		break;
#endif
	default:
		break;

	}

	return(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  manage_job

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

manage_job()
{
	fd_set fds;
	fd_set rpcset;
	int nfds = 0;
	int nfound;
	int i;
	int s;
	int old_pm_sock;
	int pm_lost = 0;
	int status;

/* Initialize the set of file descriptors */
	FD_ZERO(&fdset);

	for (i=0; i<num_shm_procs; i++)
	{
#if defined(R3000) || defined(MIPS)
		fcntl(shm_map[i].sock,F_SETFL,O_NDELAY);  /* Mark socket non-blocking */
#else
		fcntl(shm_map[i].sock,F_SETFL,O_NONBLOCK);  /* Mark socket non-blocking */
#endif
		FD_SET(shm_map[i].sock,&fdset);
	}
	if (batch_mode || pm_connect_flag)
	{
		FD_SET(pm_sock,&fdset);
#if defined(R3000) || defined(MIPS)
		fcntl(pm_sock,F_SETFL,O_NDELAY);	/* Mark socket non-blocking */
#else
		fcntl(pm_sock,F_SETFL,O_NONBLOCK);	/* Mark socket non-blocking */
#endif
		old_pm_sock = pm_sock;
	}

/* Wait for some messages */

	sel:

	if( alarm_occured )
		handle_alarm();

	if (!pm_lost && (batch_mode || pm_connect_flag))
	{
		if (pm_sock != old_pm_sock)
		{  /* We reconnected to another PM */
			FD_CLR(old_pm_sock,&fdset);
			FD_SET(pm_sock,&fdset);
#if defined(R3000) || defined(MIPS)
			fcntl(pm_sock,F_SETFL,O_NDELAY);	/* Mark socket non-blocking */
#else
			fcntl(pm_sock,F_SETFL,O_NONBLOCK);	/* Mark socket non-blocking */
#endif
			old_pm_sock = pm_sock;
		}
	}

/*
   Create a file descriptor set based on `jm_pause'.
   If `jm_pause' is set, only select on sockets that are waiting on RPC calls,
   otherwise, select on any socket.
*/
	jm_rpc_manage( &fdset, &fds, &rpcset );

	nfound = select(FD_SETSIZE,&fds,(fd_set *) 0,(fd_set *) 0,0);
	if (nfound <= 0)
	{
#ifdef AIX
		if (nfound == 0)
			goto sel;
#endif
		if (errno == EINTR)
			goto sel;
		printerr(FACILITY,FATAL_MSG,NULL,NULL,21);
		abort_job(5);
	}

/*
   Check for messages from SHMs
   If `jm_pause' is set and we are processing messages, then we are in
   step/trace mode and we should only process 1 message at a time.
*/
	for (i=0; i<num_shm_procs; i++)
	{
		if (shm_map[i].status) {
			if (FD_ISSET(shm_map[i].sock,&fds))
			{
/* Added 4/26/93 */
				if (shm_map[i].status) {
					process_message(i);
					if( jm_pause )
						break;
				}
				else {
					FD_CLR(shm_map[i].sock,&fdset);
					socket_close(shm_map[i].sock);
				}
			}
		}
	}

/*
   Process RPC requests
   This call does nothing if an RPC call was processed while `process_message'
   was waiting on a break point.
*/
	jm_rpc_get_request( (fd_set *)set_intersect( &rpcset, &fds, FD_SETSIZE ) );

/* Check for messages from the PM (if it hasn't been permanently lost) */

	if (!pm_lost && (batch_mode || pm_connect_flag))
	{
		if (FD_ISSET(pm_sock,&fds))
		{
			if ((status = read_from_pm()) == -1)
			{
				FD_CLR(pm_sock,&fdset);
				pm_lost = -1;
			}
			else if (status == 0)
			{
				process_pm_msg();
			}
		}
	}

	goto sel;
}
