/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnfaus.fnal.gov

			Copyright (c) 1991 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

	Source File:   jm_pm.c


*/

#define DB_GLOBALS 1
#include "all.h"
#include "../inc/types.h"
#include "../inc/time.h"
#include "../inc/pm.h"

struct mountreq *active = NULL;
struct mountreq *active_tail = NULL;

extern struct database db;
extern int asynch_flag;
extern int connected_to_pm;

extern char *msg_cmd,**msg_args;

static int translate_status(s)
int s;
{
	switch (s)
	{
		case DB_MOUNT_SUCCESS :
			return(CPS_OK);
		case DB_MOUNT_EOL :
			return(CPS_EOF);
		case DB_MOUNT_NODEV :
			return(CPS_DEVICE_UNAVAILABLE);
		default :
			return(CPS_ERROR);
	}
}

/************************************************************************/
/*

	Function:  pm_init

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int pm_init()
{
	struct prod_sys *p;
	struct prod_cpu_list *pcl;

	int task;

/* Connect to the PM */

	if (batch_mode)
		set_batch_job(p_system,-1,DB_PIPE_CURRENT);
	retry:
	if (pm_connect() == -1)
	{
		printerr(FACILITY,WARNING_MSG,"Failed to connect to a PM server",
			NULL,59);
		sleep_1(60);
		goto retry;
	}
	asynch_flag = -1;
	if (!batch_mode)
		return 0;

/* Get our username and password */

	p = find_prodsys(db.ps,p_system);
	if ((p == NULL) || (p->current == NULL))
	{
		printerr(FACILITY,FATAL_MSG,
			"Failed to get production system definition",NULL,60);
		exit(0);
	}
	strcpy(login_name,p->current->sub->username);
	strcpy(password,unscramble(p->current->sub->address));

/* Get the available CPUs from the production system definition */

	pcl = p->cpus;
	while (pcl != NULL)
	{
		if (pcl->cpu->status == DB_WORKING)
		{
			if (pcl->cpu->task == DB_TASK_IO)
			{
				task = TASK_IO;
			}
			else if (pcl->cpu->task == DB_TASK_COMPUTE)
			{
				task = TASK_COMPUTE;
			}
			else
			{
				task = TASK_ANY;
			}
			create_logical_cpu(pcl->lognum - 1);
			add_sdf_cpu(pcl->cpu->cpu_name,
				db_to_jm_cpu(pcl->cpu->cpu_type),task);
		}
		pcl = pcl->next;
	}

	return 0;
}

/************************************************************************/
/*

	Function:  process_pm_msg

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int process_pm_msg()
{
	struct mountreq m,*m1;
	struct active_tapes *a;

	if (strcmp(msg_cmd,"m_reply") == 0)
	{
		decode_mount_reply(&m);
		remove_mount_request(&m);
		if ((m.rstatus == DB_MOUNT_SUCCESS) ||
			(m.rstatus == DB_MOUNT_SKIPPED))
		{
			m1 = (struct mountreq *) hmalloc(sizeof(struct mountreq));
			*m1 = m;
			add_to_mreq_list(&active,&active_tail,m1);
		}
		if (m.type == DB_LIST_MOUNT)
		{
			list_mount_reply(m.aux[0],translate_status(m.rstatus),
				m.dev,m.vsn,m.listpos,m.rstate,m.errmsg);
		}
		else if (m.type == DB_VSN_MOUNT)
		{
			spec_mount_reply(m.aux[0],translate_status(m.rstatus),
				m.dev,m.errmsg);
		}
		else if (m.type == DB_SPEC_MOUNT)
		{
			spec_mount_reply(m.aux[0],translate_status(m.rstatus),
				m.dev,m.errmsg);
		}
	}
	else if (strcmp(msg_cmd,"abort") == 0)
	{
		printerr(FACILITY,INFO_MSG,"Job aborted by PM",NULL,63);
		abort_job(6);
	}
	return 0;
}

/************************************************************************/
/*

	Function:  send_list_mount_request

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int send_list_mount_request(lpn,listnum)
int lpn,listnum;
{
	int i,j;

/* Make sure we can send a request */

	if (!batch_mode)
	{
		list_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None","None",0,0,
			"Not a batch job");
		return 0;
	}
	if (connected_to_pm == 0)
	{
		list_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None","None",0,0,
			"Job Manager not connected to a PM");
		return 0;
	}

/* Build and send the mount request */

	i = lpn-FIRST_USER_LPN;
	j = usr_map[i].cpu;
	add_list_request(sdf_cpu[j].cpu_name,usr_map[i].class,listnum,1,&lpn);

	if (send_mount_requests() == -1)
	{
		list_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None","None",0,0,
			"Failed to send mount request");
		return 0;
	}

	return 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  list_mount_reply

  DESCRIPTION:

	  Send a list mount reply to a user process.

								  */
/******************************************************************/

list_mount_reply(lpn,rstatus,dev,vsn,pos,rstate,errmsg)
int lpn,rstatus,pos,rstate;
char *dev,*vsn,*errmsg;
{
	cps_message m;

	cps_bld_list_mount_reply(&m,0,lpn,rstatus,dev,vsn,pos,rstate,errmsg);
	send_to_user(lpn,&m);
	mon_shm->p_state[lpn-FIRST_USER_LPN].state = PSTATE_RUNNING;

}

/************************************************************************/
/*

	Function:  send_spec_mount_request

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int send_spec_mount_request(lpn,devtype,io,vsn)
int lpn,devtype,io;
char *vsn;
{
	int i,j;

/* Make sure we can send a request */

	if (!batch_mode)
	{
		spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None","Not a batch job");
		return 0;
	}
	if (connected_to_pm == 0)
	{
		spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None",
			"JM not connected to a PM");
		return 0;
	}

/* Build and send the mount request */

	i = lpn-FIRST_USER_LPN;
	j = usr_map[i].cpu;

	add_vsn_request(sdf_cpu[j].cpu_name,devtype,vsn,io,1,&lpn);
	if (send_mount_requests() == -1)
	{
		spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None",
			"Failed to send mount request");
		return 0;
	}

	return 0;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  spec_mount_reply

  DESCRIPTION:

	  Send a specific mount reply to a user process.

								  */
/******************************************************************/

spec_mount_reply(lpn,rstatus,dev,errmsg)
int lpn,rstatus;
char *dev,*errmsg;
{
	cps_msg m;

	cps_bld_spec_mount_reply(&m,0,lpn,rstatus,dev,errmsg);
	send_to_user(lpn,&m);
	mon_shm->p_state[lpn-FIRST_USER_LPN].state = PSTATE_RUNNING;

}

/************************************************************************/
/*

	Function:  send_mount_request

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int send_mount_request(lpn,devnam,vsn,io,comment)
int lpn,io;
char *devnam,*vsn,*comment;
{
	int i,j;
	char line[132];

/* Make sure we can send a request */

	if (connected_to_pm == 0)
	{
		spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None",
			"JM not connected to a PM");
		return 0;
	}

/* Send the comment if it exists */

	if (strcmp(comment,"no_comment") != 0)
	{
		sprintf(line,"oper_msg|%s",comment);
		if (send_to_pm(line) == -1)
		{
			spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None",
				"JM not connected to a PM");
			return 0;
		}
	}

/* Build and send the mount request */

	i = lpn-FIRST_USER_LPN;
	j = usr_map[i].cpu;

	add_spec_request(sdf_cpu[j].cpu_name,vsn,io,devnam,1,&lpn);
	if (send_mount_requests() == -1)
	{
		spec_mount_reply(lpn,CPS_NO_MOUNT_REQUEST,"None",
			"Failed to send mount request");
		return 0;
	}

	return 0;
}

/************************************************************************/
/*

	Function:  handle_done_with_tape

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int handle_done_with_tape(lpn,dev)
int lpn;
char *dev;
{
	struct mountreq *a;

	union itemptrs d[5];

	int i,j;

	struct tape_drive_list *t;
	char errmsg[132];

	if (connected_to_pm == 0)
	{
		printerr(FACILITY,ERROR_MSG,"done_with_tape: not connected to PM",
			NULL,62);
		return 0;
	}

	i = lpn-FIRST_USER_LPN;
	j = usr_map[i].cpu;

/* Look for the device in the list of active tapes */

	a = active;
	while (a != NULL)
	{
		if (cpu_name_compare(a->cpu,sdf_cpu[j].cpu_name) == 0)
		{
			if (strcmp(a->dev,dev) == 0)
			{
				goto found_it;
			}
		}
		a = a->next;
	}
	return;

	found_it:

	t = find_tape(db.tape,dev,sdf_cpu[j].cpu_name);
	if (t == NULL)
	{
		sprintf(errmsg,"done_with_tape: Cannot find %s:%s",
			sdf_cpu[j].cpu_name,dev);
		printerr(FACILITY,ERROR_MSG,errmsg,NULL,61);
	}

/* If it was a list mount, update the tape reel status and time finished */

	if (a->type == DB_LIST_MOUNT)
	{
		d[0].p = find_prodsys(db.ps,p_system);
		if (d[0].p == NULL) goto label1;
		d[1].pjl = find_job(d[0].p,-1,DB_PIPE_CURRENT);
		if (d[1].pjl == NULL) goto label1;
		d[2].cl = find_class(d[1].pjl,a->class);
		if (d[2].cl == NULL) goto label1;
		d[3].tlh = find_tape_list(d[1].pjl,a->class,a->listnum);
		if (d[3].tlh == NULL) goto label1;
		d[4].otl = find_tape_reel(d[3].tlh,a->vsn);
		if (d[4].otl == NULL) goto label1;

		d[4].otl->status = DB_PROCESSED;
		send_to_pm(build_update_msg(DB_TAPE_REEL_STATUS,d));

		time(&(d[4].otl->time_done));
		send_to_pm(build_update_msg(DB_TAPE_TIME_DONE,d));

	}

	label1:

	remove_from_mreq_list(&active,&active_tail,a);
	hfree(a);

	if (t != NULL)
	{

/* Update the mount information (for the tape drive) */

		t->mi->mstate = DB_NOT_MOUNTED;
		d[0].t = t;
		send_to_pm(build_update_msg(DB_MOUNT_INFO,d));

/* Deallocate the tape drive */

		t->used = DB_UNALLOC;
		d[0].t = t;
		send_to_pm(build_update_msg(DB_MODIFY_TAPE,d));
	}
	return 0;
}
