/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*		start_proc.c - Start user & shared memory processes				*/
/*																		*/
/*																		*/
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>
#ifdef R3000
#include <bsd/netdb.h>
#else
#include <netdb.h>
#endif
#include "../inc/version.h"

/**************************/
/* Parameter Declarations */
/**************************/



/*********************/
/* Macro Definitions */
/*********************/



/********************/
/* Global Variables */
/********************/


/************************/
/* External Definitions */
/************************/
extern int pm_sock;

extern void jm_child_died();
FILE *fopen();
struct servent *getservbyname();
char *getpass();

/* This function sets a flag which is tested with manage_job */
/* If manage_job sees it set then handle_alarm is called */
void catch_sigalrm(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	alarm_occured = 1;
}

void handle_alarm()
{
	cps_message m;

	alarm_occured = 0;
/* send a message that the shm will just ignore just to test if the shm is up */
	cps_bld_hook_switch(&m);
	shm_broadcast(&m);

	alarm(60);
}

void catch_sigusr1(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{ }

void catch_sigusr2(sig,code,sc)
int sig,code;
#if defined(R3000)
char *sc;
#else
void *sc;
#endif
{
	int pid;
	int status;

	pid = fork();

	if( pid == 0 )
	{
	}
	else
	{
		kill( pid, SIGTRAP );
		wait( &status );
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  kill_starting_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

kill_starting_shm()
{
	int i;
	char *msg = "suicide";

	for (i=0; i<num_shm_procs; i++) {
		if (shm_map[i].status) {	/* Process has been started */
			send_counted(shm_map[i].sock,msg);
		}
	}
	kill(getppid(),SIGUSR1);
	jm_exit();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  shm_proc_continue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

shm_proc_continue() {

	char *msg = "go!";

	int i;

/* Tell all started processes to continue */

	for (i=0; i<num_shm_procs; i++) {
		if (shm_map[i].status) {	/* Process has been started */
			if (send_counted(shm_map[i].sock,msg) < 0) {
				shm_map[i].status = 0;
				kill_starting_shm();
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  get_shm_ok

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

																  */
/******************************************************************/

int get_shm_ok(i) 
int i;
{
	char msg[132];

	if (read_counted(shm_map[i].sock,msg,132) < 0) {
		shm_map[i].status = 0;
		return(-1);
	}

	if (strcmp(msg,"ok") == 0)
		return(0);
	else {
		shm_map[i].status = 0;
		return(-1);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  start_shm

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

																  */
/******************************************************************/

start_shm()
{
	int i,j;

	char msg[80];
	char *startmsg = NULL;

	char *host,**ahost;

	char *envstr,*getenv();

	char shm_file[132];

	int in_sync;

/* Start each shm process */

	for (i=0; i<num_shm_procs; i++)
	{

		j = shm_map[i].cpu;
		host = sdf_cpu[j].cpu_name;
		ahost = &host;

		sprintf(shm_file,"%s/cps_shm",sdf_cpu[j].cpsdir);

		if (batch_mode)
			shm_map[i].sock =
				rexec(ahost, rexec_port,login_name,password,shm_file,0);
		else if (netrc_exists)
			shm_map[i].sock =
				rexec(ahost, rexec_port,NULL,NULL,shm_file,0);
		else if (login_prompt)
			shm_map[i].sock =
				rexec(ahost, rexec_port,login_name,password,shm_file,0);
		else
			shm_map[i].sock =
				rexec(ahost, rexec_port,sdf_cpu[j].login_name,
					sdf_cpu[j].password, shm_file,0);

		if (shm_map[i].sock < 0) {
			printerr(FACILITY,ERROR_MSG,"Call to rexec failed",NULL,30);
			kill_starting_shm();
			jm_exit();
		}
		shm_map[i].status = 1;

/* Read & discard shm output until we hit the synchronization message */

		in_sync = 0;
		do {

			startmsg = NULL;

			if( read_message(shm_map[i].sock,&startmsg) < 0 )
			{
				sprintf(msg,
					"Failed to collect initial output from shared memory manager on %s",
					sdf_cpu[j].cpu_name);
				printerr(FACILITY,ERROR_MSG,msg,NULL,58);
				shm_map[i].status = 0;	/* Process not started */
				in_sync = -1;
			}
			else if (strncmp(startmsg,
				"Cooperative Processes Software",30) == 0)
			{
				in_sync = -1;
			}
			else
			{
				print_shm_log(i,startmsg);
			}
			if (startmsg != NULL) {
				if( strlen( startmsg ) > 0 )
					hfree( startmsg );
			}
		} while (!in_sync);
	}

	for (i=0; i<num_shm_procs; i++)
		if (shm_map[i].status == 0)
			kill_starting_shm();

/* Tell each share memory process to continue */

	shm_proc_continue();

/* Tell each shm its id and how many user & shm processes there are */

	for (i=0; i<num_shm_procs; i++) {
		sprintf(msg,"%d",i+FIRST_SHM_LPN);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			kill_starting_shm();
		}
		sprintf(msg,"%d",num_usr_procs);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			kill_starting_shm();
		}
		sprintf(msg,"%d",num_shm_procs);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			kill_starting_shm();
		}
		sprintf(msg,"%d",set_size(shm_map[i].proc,MAX_PROCESSES));
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			kill_starting_shm();
		}
	}

/* Tell each share memory process to continue */

	shm_proc_continue();

/* Get information from each shm process */

	for (i=0; i<num_shm_procs; i++) {

/* Check to see if the shared memory was created */

		if (get_shm_ok(i) == -1)
			continue;

/* Get the id of the shared memory segment */

		if (read_counted(shm_map[i].sock,msg,80) < 0) {
			shm_map[i].status = 0;
			continue;
		}

		shm_map[i].shmid = atoi(msg);

/* Get the location of the shared memory manager */

		if (read_counted(shm_map[i].sock,msg,80) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		strcpy(shm_map[i].hostname,msg);

		if (read_counted(shm_map[i].sock,msg,80) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		strcpy(shm_map[i].hostaddr,msg);

		if (read_counted(shm_map[i].sock,msg,80) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		shm_map[i].portno = atoi(msg);

	}

/* Make sure all shm processes are still alive */

	for (i=0; i<num_shm_procs; i++)
		if (shm_map[i].status == 0)
			kill_starting_shm();

/* Tell each shared memory process to continue */

	shm_proc_continue();

/* Distribute shm locations to all shared memory processes */

	for (i=0; i<num_shm_procs; i++) {

		sprintf(msg,"%s",local_hostname);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		sprintf(msg,"%s",local_hostaddr);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		sprintf(msg,"%d",assigned_port_number);
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			continue;
		}

		for (j=0; j<num_shm_procs; j++) {
			sprintf(msg,"%s",shm_map[j].hostname);
			if (send_counted(shm_map[i].sock,msg) < 0) {
				shm_map[i].status = 0;
				break;
			}
			sprintf(msg,"%s",shm_map[j].hostaddr);
			if (send_counted(shm_map[i].sock,msg) < 0) {
				shm_map[i].status = 0;
				break;
			}
			sprintf(msg,"%d",shm_map[j].portno);
			if (send_counted(shm_map[i].sock,msg) < 0) {
				shm_map[i].status = 0;
				break;
			}
		}
	}

/* Make sure all shm processes are still alive */

	for (i=0; i<num_shm_procs; i++)
	if (shm_map[i].status == 0)
		kill_starting_shm();

/* Tell each share memory process to continue */

	shm_proc_continue();

/* Send user-->shm mappings terminated by a negative one */

	for (i=0; i<num_shm_procs; i++) {
		for (j=0; j<num_usr_procs; j++) {
			sprintf(msg,"%d",usr_map[j].shm+1);
			if (send_counted(shm_map[i].sock,msg) < 0) {
				shm_map[i].status = 0;
				break;
			}
		}
		if (shm_map[i].status == 0) continue;

		sprintf(msg,"-1");
		if (send_counted(shm_map[i].sock,msg) < 0) {
			shm_map[i].status = 0;
			continue;
		}
	}

/* Check to see if the shm got the entire map */
	for (i=0; i<num_shm_procs; i++)
		if (get_shm_ok(i) == -1) shm_map[i].status = 0;

/* Make sure all shm processes are still alive */

	for (i=0; i<num_shm_procs; i++)
	if (shm_map[i].status == 0)
		kill_starting_shm();

/* Tell each share memory process to continue */

	shm_proc_continue();

/* Check to see if the semaphores were created */

	for (i=0; i<num_shm_procs; i++)
		if (get_shm_ok(i) == -1)
			shm_map[i].status = 0;

/* Get the semaphore ids */

	for (i=0; i<num_shm_procs; i++) {
		if (read_counted(shm_map[i].sock,msg,80) < 0) {
			shm_map[i].status = 0;
			continue;
		}
		shm_map[i].semid = atoi(msg);
	}

/* Make sure all shm processes are still alive */

	for (i=0; i<num_shm_procs; i++)
	if (shm_map[i].status == 0)
		kill_starting_shm();

/* Tell each share memory process to continue */

	shm_proc_continue();

	return(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  send_class_map

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

send_class_map(p)
int p;
{
	int i,j;

	char msg[20];

	int *x;

	for (i=1; i<=MAX_CLASSES; i++) {
		if (class[i].valid) {
			sprintf(msg,"%d",i);
			if (send_counted(usr_map[p].sock,msg) < 0) {
				usr_map[p].status = 0;
				return;
			}
			x = set_query(class[i].sop,MAX_PROCESSES);
			j = -1;
			do {
				j++;
				sprintf(msg,"%d",x[j]);
				if (send_counted(usr_map[p].sock,msg) < 0) {
					usr_map[p].status = 0;
					return;
				}
			} while (x[j] != 0);
		}
	}

	sprintf(msg,"%d",0);
	if (send_counted(usr_map[p].sock,msg) < 0) {
		usr_map[p].status = 0;
		return;
	}

}

/******************************************************************/
/*
	FUNCTION NAME:

	  kill_starting_proc

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

kill_starting_proc()
{
	int i;
	char *msg = "suicide";

	for (i=0; i<num_usr_procs; i++) {
		if (usr_map[i].status) {	/* Process has been started */
			if (send_counted(usr_map[i].sock,msg) < 0) {
				usr_map[i].status = 0;
			}
		}
	}
	jm_exit();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  proc_continue

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

proc_continue() {

	char *msg = "go!";

	int i;

/* Tell all started processes to continue */

	for (i=0; i<num_usr_procs; i++) {
		if (usr_map[i].status) {	/* Process has been started */
			if (send_counted(usr_map[i].sock,msg) < 0) {
				usr_map[i].status = 0;
				remove_from_set(i+1,class[usr_map[i].class].alive);
				add_to_set(i+1,class[usr_map[i].class].dead);
			}
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  sigcld_handler

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

sigcld_handler(sig,code,sc)
int sig,code;
struct sigcontext *sc;
{
	kill_starting_proc();
	jm_exit();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  start_proc

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

																  */
/******************************************************************/

int start_proc()
{
	struct servent *s;

	int i,j;

	char *host,**ahost,*pass;

	char msg[256];
	char *startmsg = NULL;
	int in_sync;

	int pid;

	char shmfil[80];

	FILE *fp;

	int c;

	char prompt[80];

/* Initialize usr and shm maps */

	for (i=0; i<MAX_PROCESSES; i++)
		usr_map[i].status = 0;
	for (i=0; i<MAX_SHM_PROCESSES; i++)
		shm_map[i].status = 0;

/* If -pw option, get name and password once */

	if (!netrc_exists && !batch_mode && login_prompt)
	{
		printf("login: ");
		if( getline( login_name, 24 ) == 0 )
			jm_exit();
		while (1)
		{
			pass = getpass("password: ");
			if( strlen( pass ) == 0 )
				jm_exit();
			if (pass != NULL) 
			{
				if (pass[0] != '\0') break;
			}
		}
		strcpy(password,pass);
	}

/* Get the rexec port number */

	s = getservbyname("exec","tcp");
	if (s == NULL)
	{
		printerr(FACILITY,FATAL_MSG,"Error getting rexec port number",NULL,45);
		jm_exit();
	}
	rexec_port = s->s_port;

/* Start each user process and read until the sync point is reached */

	for (i=0; i<num_usr_procs; i++)
	{
		int user_message = 1;

		j = usr_map[i].cpu;

		host = sdf_cpu[j].cpu_name;
		ahost = &host;

		if (!netrc_exists && !batch_mode && !login_prompt && 
			(sdf_cpu[j].login_name[0] == '\0'))
		{
			printf("%s login: ",host);
			if( getline( sdf_cpu[j].login_name, 24 ) == 0 )
				jm_exit();
			sprintf(prompt,"%s password: ",host);
			while (1)
			{
				pass = getpass(prompt);
				if( strlen( pass ) == 0 )
					jm_exit();
				if (pass != NULL)
				{
					if (pass[0] != '\0') break;
			   	}
		   	}
			strcpy(sdf_cpu[j].password,pass);
		}

		if (batch_mode)
		{
			usr_map[i].sock = rexec(ahost, rexec_port,login_name,
				password,class[usr_map[i].class].program,0);
		}
		else if (netrc_exists)
		{
			usr_map[i].sock = rexec(ahost, rexec_port,NULL,
				NULL,class[usr_map[i].class].program,0);
		}
		else if (login_prompt)
		{
			usr_map[i].sock = rexec(ahost, rexec_port,login_name,
			password,class[usr_map[i].class].program,0);
		}
		else
		{
			usr_map[i].sock = rexec(ahost, rexec_port,sdf_cpu[j].login_name,
			sdf_cpu[j].password,class[usr_map[i].class].program,0);
		}

		if (usr_map[i].sock < 0)
		{
			printerr(FACILITY,FATAL_MSG,"Call to rexec failed",
				"Make sure the account exists on each CPU",46);
			kill_starting_proc();
			jm_exit();
		}
		usr_map[i].status = 1;

/* Read user output until we hit the synchronization message */

		in_sync = 0;
		do
		{
			startmsg = NULL;

			if( read_message(usr_map[i].sock,&startmsg ) < 0 )
			{
				sprintf(msg,
					"Failed to collect initial output: class = %d, proc = %d, cpu = %s",
					usr_map[i].class,i+1,sdf_cpu[usr_map[i].cpu].cpu_name);
				printerr(FACILITY,ERROR_MSG,msg,NULL,47);
				usr_map[i].status = 0;	/* Process not started */
				in_sync = -1;
			}
			else if (strncmp(startmsg,
				"Cooperative Processes Software",30) == 0)
			{
				char lib_version[8], lib_relcode[2];
				int inconsistent = 0;

				sprintf(msg,"Cooperative Processes Software %-8s %2s\n",
					cps_version,cps_relcode);
				/* Check for Non-Version Code */
				if( strncmp( startmsg, msg, 34 ) != 0 )
					inconsistent = 1;
				/* Test for releases within version */
				else
				{
					sscanf( startmsg,
						"Cooperative Processes Software %8s %2s\n",
						lib_version, lib_relcode );
					if( strcmp( cps_relcode, lib_relcode ) != 0 )
						inconsistent = 1;
				}
				if( inconsistent )
				{
					sprintf(msg,
						"Job Manager version is inconsistent with CPS library",
						usr_map[i].class,i+1,sdf_cpu[usr_map[i].cpu].cpu_name);
					printerr(FACILITY,ERROR_MSG,msg,NULL,49);
					usr_map[i].status = 0;  /* Process not started */
				}
				in_sync = -1;
			}
			else if (strncmp(startmsg,"Process starting on",19) == 0)
			{
				print_usr_log( i + 1, startmsg );
			}
			else
			{
				if( user_message )
					print_lpn( i + 1 );
				print_log_no_cr( startmsg );
				user_message = 0;
			}

			if (startmsg != NULL) {
				if( strlen( startmsg ) > 0 )
					hfree( startmsg );
			}
		} while (!in_sync);

	}

/* Build set of alive & dead processes for each class */

	for (i=0; i<num_usr_procs; i++)
	{
		c = usr_map[i].class;
		if (usr_map[i].status)
		{
			add_to_set(i+1,class[c].alive);
		}
		else
		{
			add_to_set(i+1,class[c].dead);
		}
	}

/* Check to see if enough processes in each class have started */

	for (i=1; i<=MAX_CLASSES; i++)
	{
		if (class[i].valid)
		{
			if (set_size(class[i].alive,MAX_PROCESSES) < class[i].min_proc)
			{
				sprintf(msg,"Too few processes started in class %d",i);
				printerr(FACILITY,FATAL_MSG,msg,NULL,48);
				kill_starting_proc();
				jm_exit();
			}
		}
	}

/* Tell all started processes to continue */

	proc_continue();

/* Send the lpn */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status)
			continue;
		sprintf(msg,"%d",i+1);
		if (send_counted(usr_map[i].sock,msg) < 0)
		{
			usr_map[i].status = 0;
			remove_from_set(i+1,class[usr_map[i].class].alive);
			add_to_set(i+1,class[usr_map[i].class].dead);
		}
	}

/* Send the system hooks flag */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status)
			continue;
		sprintf(msg,"%d",jm_hooks);
		if (send_counted(usr_map[i].sock,msg) < 0)
		{
			usr_map[i].status = 0;
			remove_from_set(i+1,class[usr_map[i].class].alive);
			add_to_set(i+1,class[usr_map[i].class].dead);
		}
	}

/* Get the pid */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status) continue;
		if (read_counted(usr_map[i].sock,msg,256) < 0)
		{
			usr_map[i].status = 0;	/* Process not started */
		}
		else
		{
			usr_map[i].pid = atoi(msg);
		}
	}

/* Get the process name */

/*
	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status) continue;
		if (read_counted(usr_map[i].sock,msg,256) < 0)
		{
			usr_map[i].status = 0;
		}
		else
		{
			usr_map[i].name = (char *)hmalloc( strlen( msg ) + 1 );
			strcpy( usr_map[i].name, msg );
		}
	}
*/

/* Get the process working directory */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status) continue;
		if (read_counted(usr_map[i].sock,msg,256) < 0)
		{
			usr_map[i].status = 0;
		}
		else
		{
			usr_map[i].directory = (void *)hmalloc( strlen( msg ) + 1 );
			strcpy( usr_map[i].directory, msg );
		}
	}

/* Get the user's port number for direct data transfers */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status) continue;
		if (read_counted(usr_map[i].sock,msg,256) < 0)
		{
			usr_map[i].status = 0;	/* Process not started */
		}
		else
		{
			usr_map[i].portno = atoi(msg);
		}
	}

/* Send all port numbers to each user */

	for (i=0; i<num_usr_procs; i++)
	{
		if (!usr_map[i].status) continue;
		for (j=0; j<num_usr_procs; j++)
		{
			if (j == i) continue;
			sprintf(msg,"%d %d",j+1,usr_map[j].portno);
			if (send_counted(usr_map[i].sock,msg) < 0)
			{
				usr_map[i].status = 0;
				remove_from_set(i+1,class[usr_map[i].class].alive);
				add_to_set(i+1,class[usr_map[i].class].dead);
				break;
			}
		}
		if (!usr_map[i].status) continue;
		sprintf(msg,"0 0");
		if (send_counted(usr_map[i].sock,msg) < 0)
		{
			usr_map[i].status = 0;
			remove_from_set(i+1,class[usr_map[i].class].alive);
			add_to_set(i+1,class[usr_map[i].class].dead);
		}
	}

/* Enable parent process to catch SIGUSR1 one time */

	sig_establish_handler(SIGUSR1,catch_sigusr1);

/* Core file signal for debugging */

	sig_establish_handler(SIGUSR2,catch_sigusr2);

/* Alarm Handler to check up on SHMs */

	sig_establish_handler(SIGALRM,catch_sigalrm);

/* Fork a subprocess */

	fflush(stdout);		/* the fork */
	if ((pid = fork()) < 0)
	{
		printerr(FACILITY,FATAL_MSG,NULL,NULL,50);
		kill_starting_proc();
		jm_exit();
	}

	sprintf(shmfil,"%s.shm",job_name);
	unlink(shmfil);

	if (pid == 0)
	{		/* child */

/* Start shared memory processes */

		start_shm();

/* Create file with shmids */

		fp = fopen(shmfil,"w");
		if (fp == NULL)
		{
			printerr(FACILITY,FATAL_MSG,NULL,NULL,51);
			jm_exit();
		}
		for (i=0; i<num_shm_procs; i++)
			fprintf(fp,"%d\n",shm_map[i].shmid);
		fclose(fp);

/* Signal parent that the file is ready */

		if (kill(getppid(),SIGUSR1) < 0)
		{
			printerr(FACILITY,FATAL_MSG,"Unable to signal parent process",
				NULL,52);
			jm_exit();
		}

/* Close all rexec (term i/o) sockets */

		for (i=0; i<num_usr_procs; i++)
			socket_close(usr_map[i].sock);

/* Wait for a SIGUSR1 signal from parent */

		pause();

/* All processes are now started */

		return;

	}
	else
	{			/* parent */

/* Wait for a SIGUSR1 signal */

		pause();

/* Re-define the signal handler for SIGCLD */

		sig_establish_handler(SIGCLD,jm_child_died);

/* Ignore interrupts */

		sig_establish_handler(SIGINT,SIG_IGN);

/* Read in the file containing the shared memory ids */

		sprintf(shmfil,"%s.shm",job_name);
		fp = fopen(shmfil,"r");
		if (fp == NULL)
		{
			printerr(FACILITY,FATAL_MSG,NULL,
				"System may be out of semaphores, or shared memory may be too large",53);
			kill_starting_proc();
			jm_exit();
		}
		for (i=0; i<num_shm_procs; i++)
			fscanf(fp,"%d",&(shm_map[i].shmid));
		fclose(fp);

/* Tell all started processes to continue */

		proc_continue();

/* Tell each user process the shmid to attach to */

		for (i=0; i<num_usr_procs; i++)
		{
			if (usr_map[i].status)
			{
				sprintf(msg,"%d",shm_map[usr_map[i].shm].shmid);
				if (send_counted(usr_map[i].sock,msg) < 0)
				{
					usr_map[i].status = 0;
				}
				else
				{
					send_class_map(i);
				}
			}
		}
/* Wait for each user process to complete cps_init */

		for (i=0; i<num_usr_procs; i++)
		{
			if (!usr_map[i].status) continue;
			if (read_counted(usr_map[i].sock,msg,256) < 0)
			{
				usr_map[i].status = 0;  /* Process not started */
			}
			else
			{
				usr_map[i].portno = atoi(msg);
			}
			if (strcmp(msg,"done") != 0)
			{
			}
		}

/* Signal the JM child process so it can continue */

		if (kill(pid,SIGUSR1) < 0)
		{
			printerr(FACILITY,FATAL_MSG,"Unable to signal child process",
				NULL,53);
			jm_exit();
		}

/* Close the socket with the PM */

		if (pm_connect_flag || batch_mode) socket_close(pm_sock);

		jm_parent();	/* Collect terminal I/O */

	}
}
