/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

#include "jm_param.h"

struct cpu_info {

    int cpu_type;
    char *cpu_name;
	int task;
    int nprocs;			/* Number of procs so far assigned */
    unsigned long *procs;	/* Set of procs on this CPU */
    unsigned long *class;	/* Set of classes sharing this CPU */

    char login_name[24];
    char password[24];

    char cpsdir[80];
};


struct class_info {		/* Indexed by class number */

    int valid;			/* 0 = not in use, -1 = in use */
    unsigned long *sop;		/* Set of processes */
    unsigned long *share;	/* Set of classes allowed to share with */

    char *program;

    int cpu_type;
	int task;
    int nspec;			/* Number of specific CPUs */
    int *spec_cpu;		/* Indices into cpu_info */

    int nproc;			/* Number of processes */
    int per_cpu;		/* Processes per CPU */
    int ncpu;			/* Number of CPUs needed */
    int min_proc;		/* Minimum # processes */

    int debug;			/* Number of debug processes */
    int timeout;		/* User specified timeout */

    unsigned long *cpu;		/* Set of CPUs we can choose from */

    unsigned long *fini;	/* Set of finished processes */
    unsigned long *alive;	/* Set of alive processes */
    unsigned long *dead;	/* Set of dead processes */

	int logical_processes[ MAX_SDF_CPUS ];
	int logical_assignment;
};

struct shm_map {	/* array index = LPN - FIRST_SHM_LPN */

    int cpu;		/* index into the sdf_cpu array */
    char hostname[64];
    char hostaddr[24];
    int portno;
    int shmid;		/* shared memory identifier */
    int semid;		/* semaphore id */
    int status;		/* 0 = unconnected, 1 = connected */
    int sock;		/* socket descriptor */
    unsigned long *proc;	/* Set of local user processes (lpns) */
};

struct user_map {		/* array index = LPN - FIRST_USER_LPN */

    int class;
    int shm;			/* Index into shm map */
    int status;			/* 0 = unconnected, 1 = connected */
    int sock;			/* socket descriptor */
    int cpu;			/* CPU assigned to this process */
    int pid;			/* Pid of this process */
    int portno;			/* Port number used for accepting connections */
	char *directory;	/* Process working directory */
};

struct queue_info {	/* Array indexed by queue number - 1 */

    int active;			/* 0 = undeclared, -1 = declared */

    int nargs;			/* Number of arguments */
    int argcnt[MAX_QUEUE_ARGS];	/* Size of each argument */
    int bcount;			/* Sum of byte count of arguments */
    unsigned long *declared;	/* Set of procs declared (and alive) */
    unsigned long *wait_full;	/* Set of procs waiting for full */
    unsigned long *wait_empty;  /* Set of procs waiting for empty */

    unsigned long *q_set;	  /* Set of queued processes */
    struct cps_msg_queue *queued; /* Processes on this queue */
    struct cps_msg_queue *free;	  /* Pool of free queue elements */

    unsigned long *dq_set;	  /* Set of procs waiting to dequeue */
    struct cps_msg_queue *dequeue;/* Processes waiting to dequeue */

    int eoq;			  /* -1 if end of queue was queued */
};


struct acp_remote_sub {

    int id;
    int faddr;		/* Virtual address of the function */
    int nargs;
    int bcount[32];
    int bsum;		/* Sum of byte counts */
    int argtyp[32];	/* Argument types (read-only, write-only, read-write) */

    struct acp_remote_sub *next;
};


struct acp_data_rdy {

    int blockno;
    int bcount;
    int valid;
    struct acp_data_rdy *next;

};
