/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "jmdb.h"

int num_points;

PointInfo break_point[MAX_BREAK_POINTS];

extern ClientInfo client;
extern VariableInfo var;

/******************************************************************/
/*
	FUNCTION NAME:

		print_breakpoint

	DESCRIPTION:

		print break point info

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
print_breakpoint( fp, index, point )
FILE *fp;
int index;
PointInfo point;
{
	switch( point.type )
	{
	case NO_POINT:
		if( index == -1 )
			fprintf( fp, STR_NOT_AT_BREAKPOINT );
		break;
	case SYNC_POINT:
		if( index >= 0 )
			fprintf( fp, STR_STOP, index );
		fprintf( fp, STR_AT_SYNC, point.number );
		break;
	case QUEUE_POINT:
		if( index >= 0 )
			fprintf( fp, STR_STOP, index );
		if( point.number > 128 )
			point.number = point.number - 128;
		fprintf( fp, STR_AT_QUEUE, point.number );
		break;
	case USER_HOOK_POINT:
		if( index >= 0 )
			fprintf( fp, STR_STOP, index );
		fprintf( fp, STR_AT_USER_HOOK, point.number );
		break;
	case FUNC_HOOK_POINT:
		if( index >= 0 )
			fprintf( fp, STR_STOP, index );
		fprintf( fp, STR_AT_FUNCTION, function_name[ point.number ] );
		break;
	case START_POINT:
		fprintf( fp, STR_AT_START );
		break;
	default:
		break;
	}
	fflush( fp );
}

/******************************************************************/
/*
	FUNCTION NAME:

		breakpoint_exists

	DESCRIPTION:

		returns 1 if any breakpoint is set
		returns 0 if no breakpoints are set

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
char
breakpoint_exists()
{
	int index;
	for( index = 0; index < MAX_BREAK_POINTS; index++ )
		if( break_point[ index ].type != NO_POINT )
			return( 1 );
	return( 0 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_status

	DESCRIPTION:

		display break points

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_status()
{
	int i;
	FILE *temp;
	char tempfile[128];
	char string[128];
	int flag = 2;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	strcpy( tempfile, tempnam( "", "jmdb" ) );
	if( ( temp = fopen( tempfile, "w" ) ) == NULL )
	{
		fprintf( stdout, STR_FOPEN_ERR, tempfile, strerror(errno) );
		return;
	}

	fprintf( stdout, STR_HOOKS );
	if( *rpc_hook_switch_1( &flag, client.cl ) )
		fprintf( stdout, STR_ON );
	else
		fprintf( stdout, STR_OFF );

	for( i = 0; i < MAX_BREAK_POINTS; i++ )
		print_breakpoint( temp, i, break_point[ i ] );

	fclose( temp );

	sprintf( string, "%s %s", var.pager, tempfile );
	if( system( string ) < 0 )
		perror("status");

	unlink( tempfile );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_set_breakpoint

	DESCRIPTION:

		set a range of break points

	FORMAL PARAMETERS:

		hook1: low number of range
		hook2: high number of range

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_set_breakpoint( type, point1, point2 )
int type, point1, point2;
{
	int max_range;
	int_pair pair;
	int i;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( type == SYNC_POINT )
		max_range = MAX_SYNCS - 1;
	else if( type == QUEUE_POINT )
		max_range = MAX_QUEUES - 1;
	else if( type == USER_HOOK_POINT )
		max_range = MAX_USER_HOOKS - 1;
	else if( type == FUNC_HOOK_POINT )
		max_range = MAX_FUNC_HOOKS - 1;

	if( ( point1 == -1 ) && ( point2 == -1 ) )
	{
		fprintf( stdout, STR_UNKNOWN_FUNCTION );
		return;
	}

	/* Special case for set all */
	if( point2 == -1 ) point2 = max_range;

	/* Test for value too large */
	if( point1 > max_range )
	{
		fprintf( stdout, STR_RANGE_ERR, point1, 0, max_range );
		return;
	}
	if( point2 > max_range )
	{
		fprintf( stdout, STR_RANGE_ERR, point2, 0, max_range );
		return;
	}

	for( i = point1; i <= point2; i++ )
	{
		pair.x = type;
		pair.y = i;
		rpc_stop_at_brkpt_1( &pair, client.cl );
		break_point[ num_points ].type = type;
		break_point[ num_points ].number = i;
		print_breakpoint( stdout, num_points, break_point[ num_points ] );
		num_points++;
		if( num_points == MAX_BREAK_POINTS )
			num_points = 0;
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_delete

	DESCRIPTION:

		delete a range of break points

	FORMAL PARAMETERS:

		arg1: low number of range
		arg2: high number of range

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_delete( arg1, arg2 )
int arg1, arg2;
{
	int arg, number;
	int_pair pair;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( arg2 == -1 ) arg2 = MAX_BREAK_POINTS - 1;

	check_number_range( arg1, 0, MAX_BREAK_POINTS - 1 );
	check_number_range( arg2, 0, MAX_BREAK_POINTS - 1 );

	for( arg = arg1; arg <= arg2; arg++ )
	{
		if( break_point[ arg ].type != NO_POINT )
		{
			pair.x = break_point[ arg ].type;
			pair.y = break_point[ arg ].number;
			rpc_cont_at_brkpt_1( &pair, client.cl );
			break_point[ arg ].type = NO_POINT;
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_import

	DESCRIPTION:

		read in break points from current JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_import()
{
	int index;
	int type;
	int_array_res *result;
	int_array plist;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	num_points = 0;

	for( type = SYNC_POINT; type <= FUNC_HOOK_POINT; type++ )
	{
		result = rpc_get_breakpoints_1( &type, client.cl );
		index = 0;
		for( plist = result->int_array_res_u.plist;
		    plist != NULL; plist = plist->next )
		{
			if( plist->lpn )
			{
				break_point[ num_points ].type = type;
				break_point[ num_points ].number = index;
				print_breakpoint( stdout, num_points,
				    break_point[ num_points ] );
				num_points++;
			}
			index++;
		}
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_stop

	DESCRIPTION:

		pause the JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_stop()
{
	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( !*rpc_stop_jm_1( NULL, client.cl ) )
		fprintf( stdout, STR_ALREADY_STOPPED );
	else
		fprintf( stdout, STR_STOPPING_JM );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_cont

	DESCRIPTION:

		continue the JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_cont()
{
	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	if( !*rpc_continue_jm_1( NULL, client.cl ) )
		fprintf( stdout, STR_ALREADY_RUNNING );
	else
		fprintf( stdout, STR_STARTING_JM );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_where

	DESCRIPTION:

		return current location of JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void handle_where()
{
	int_pair result;

	if( !connected( client.cl ) ) return;

	if( !in_debug_mode() ) return;

	result = *rpc_get_location_1( NULL, client.cl );
	print_breakpoint( stdout, -1, result );
}

/******************************************************************/
/*
	FUNCTION NAME:

		init_break_points

	DESCRIPTION:

		initialize break point array

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
void init_break_points()
{
	int index;

	num_points = 0;
	for( index = 0; index < MAX_BREAK_POINTS; index++)
		break_point[ index ].type = NO_POINT;
}
