/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/
#include "jmdb.h"

extern ClientInfo client;
extern VariableInfo var;
unsigned long *AllowableFunctionSet;
extern char Interrupt;

/******************************************************************/
/*
	FUNCTION NAME:

		init_allowable_function_set

	DESCRIPTION:

		Creates a set of functions that the JM can catch
		All these functions must call CHECK_FUNC_HOOK
		Beware, for performance, some functions which call 
		CHECK_FUNC_HOOK will never generate a message
		(e.g.) FUNC_ERROR_HANDLER

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
init_allowable_function_set()
{
	int index;

	AllowableFunctionSet = (unsigned long *)
	    create_empty_set( MAX_ALLOWABLE_FUNCTION );

	for( index = 0; index < MAX_ALLOWABLE_FUNCTION; index++ )
	{
		switch( index )
		{
		case FUNC_INIT:
		case FUNC_STOP_JOB:
		case FUNC_STOP_PROCESS:
		case FUNC_DECLARE_BLOCK:
		case FUNC_GET:
		case FUNC_SEND:
		case FUNC_CALL:
		case FUNC_DECLARE_ARGUMENTS:
		case FUNC_DECLARE_SUBROUTINE:
		case FUNC_SERVICE_CALLS:
		case FUNC_CHANGE_RETURN_QUEUE:
		case FUNC_SYNC:
		case FUNC_WAIT_QUEUE:
		case FUNC_DECLARE_QUEUE:
		case FUNC_DEQUEUE_PROCESS:
		case FUNC_QUEUE_PROCESS:
		case FUNC_DEQUEUE_IF_POSSIBLE:
		case FUNC_CHECK_MESSAGE:
		case FUNC_RECEIVE_MESSAGE:
		case FUNC_TRANSMIT_MESSAGE:
		case FUNC_CLASS_INFO:
		case FUNC_DEAD_PROCESS_INFO:
		case FUNC_JOB_INFO:
		case FUNC_UPDATE_USER_STATUS:
		case FUNC_START_SEND:
		case FUNC_WAIT_FOR_DATA:
		case FUNC_DATA_READY:
		case FUNC_START_GETS:
		case FUNC_HANDLE_DATA:
		case FUNC_END_OF_DATA:
		case FUNC_BUFFER_SET_WAIT:
		case FUNC_QUEUE_XFER:
		case FUNC_DQ_GET:
		case FUNC_DQ_SEND:
		case FUNC_WAIT_FOR_GET:
		case FUNC_LOG:
		case FUNC_CHANGE_ACTION:
		case FUNC_CONVERT:
		case FUNC_HOOK:
		case FUNC_DONE_WITH_TAPE:
		case FUNC_GET_NEXT_TAPE:
		case FUNC_GET_SPEC_TAPE:
		case FUNC_MOUNT_REQUEST:
			add_to_set( index + 1, AllowableFunctionSet );
			break;
		default:
			break;
		}
	}
}
/******************************************************************/
/*
	FUNCTION NAME:

		get_procedure_number

	DESCRIPTION:

		given text string return index to function array

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
int
get_procedure_number( string )
char *string;
{
	int index;

	for( index = 0; index < MAX_ALLOWABLE_FUNCTION - 1; index++ )
	{
		if( ( strcmp( string, function_name[ index ] ) == 0 ) &&
		    set_member( index + 1, AllowableFunctionSet ) )
			return( index );
	}
	return( -1 );
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_step

	DESCRIPTION:

		step through JM a message at a time

	FORMAL PARAMETERS:

		nsteps: number of times to step
		mode:		2 for normal step mode
					1 for trace mode

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_step( nsteps, mode, fp, p_set )
int nsteps;
int mode;
FILE *fp;
unsigned long *p_set;
{
	int index;
	int type, src, dst, size;
	rpc_message *message;
	long *data;
	int process, queue;
	FILE *outfp;
	int status;

	if( !connected( client.cl ) ) return;

	if( ( mode == STEP_MODE ) && !in_debug_mode() ) return;

	if( fp == NULL )
		outfp = stdout;
	else
		outfp = fp;

	for( index = 0; index < nsteps; index++ )
	{
		status = *rpc_step_jm_1( NULL, client.cl );
		if( !status )
		{
			fprintf( stdout, STR_ALREADY_RUNNING );
			break;
		}

		message = rpc_get_jm_message_1( NULL, client.cl );

		type = message->type;

		src  = message->src_lpn;
		dst  = message->dst_lpn;
		size = message->size;

		if( !var.display )
		{
			switch( type )
			{
			case MSG_SHM_ERROR:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
					"SHM_Error" );
				break;
			case MSG_SHM_SHUTDOWN:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
					"SHM_Shutdown" );
				break;
			case MSG_USR_LOG:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
					"User_Log" );
				break;
			case MSG_USR_ABORT_JOB:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
					"Stop_Job" );
				break;
			case MSG_USR_PROC_DEAD:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
					"User_Process_Dead" );
				break;
			case MSG_DECLARE_QUEUE:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s queue=%d\n", src,
					"Declare_Queue", *data );
				break;
			case MSG_QUEUE_PROCESS:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				process = *data;
				data++;
				queue = *data;
				fprintf( outfp, "[JM:%3d] %-18s queue=%d process=%d\n", src,
					"Queue_Process", queue,process );
				break;
			case MSG_DEQUEUE_PROCESS:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s queue=%d\n", src,
				    "Dequeue_Process", *data );
				break;
			case MSG_END_OF_QUEUE:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s queue=%d\n", src,
				    "End_of_Queue", *data );
				break;
			case MSG_WAIT_QUEUE:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s queue=%d\n", src,
				    "Wait_Queue", *data );
				break;
			case MSG_WAIT_SYNC:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s sync=%d\n", src,
				    "Wait_Sync", *data );
				break;
			case MSG_USER_STATUS:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "User_Status" );
				break;
			case MSG_SUB_DECL:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s subr=%d\n", src,
				    "Declare_Subroutine", *data );
				break;
			case MSG_REQ_SUB_DECL:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Get_Subroutine_Info" );
				break;
			case MSG_REQ_Q_INFO :
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Get_Queue_Info" );
				break;
			case MSG_USR_FINISHED:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Stop_Process" );
				break;
			case MSG_DECLARE_BLOCK:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				data++;
				fprintf( outfp, "[JM:%3d] %-18s block=%d\n", src,
				    "Declare_Block", *data );
				break;
			case MSG_GET_NEXT_TAPE:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Get_Next_Tape" );
				break;
			case MSG_GET_SPEC_TAPE:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Get_Specific_Tape" );
				break;
			case MSG_DONE_WITH_TAPE:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Done_with_Tape" );
				break;
			case MSG_MON_UPDATE:
				if( !set_member( src + 1, p_set ) ) break;
				break;
			case MSG_MOUNT_REQUEST:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Mount_Request" );
				break;
			case MSG_VBBC_INFO:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "VBBC_Info" );
				break;
			case MSG_ASYNCH_DQ:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Asynch_Dequeue" );
				break;
			case MSG_DATA_READY:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s\n", src,
				    "Data_Ready" );
				break;
			case MSG_USER_HOOK:
				if( !set_member( src + 1, p_set ) ) break;
				data = (long *) message->msg;
				fprintf( outfp, "[JM:%3d] %-18s num=%d\n", src,
				    "User_Hook", *data );
				break;
			case MSG_FUNC_HOOK:
				data = (long *) message->msg;
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "   [%3d] %-18s\n", src,
				    function_name[ *data ] );
				break;
			case -1:
				break;
			default:
				if( !set_member( src + 1, p_set ) ) break;
				fprintf( outfp, "[JM:%3d] %-18s type=%d\n", src,
				    "JM Message", type );
				break;
			} /* end of switch */
			fflush( outfp );
		}
		else if( var.display )
		{
			switch( type )
			{
			case MSG_MON_UPDATE:
				qmon_update( 1 );
				break;
			case MSG_DECLARE_QUEUE:
				data = (long *) message->msg;
				qmon_newq( *data );
				break;
			default:
				break;
			} /* end of switch */
		}
		/*
		if( mode == TRACE_MODE )
			dump_info();
		*/
	} /* end of for loop */
	if( mode == STEP_MODE )
		dump_info();
}

/******************************************************************/
/*
	FUNCTION NAME:

		handle_trace

	DESCRIPTION:

		trace through the JM

	FORMAL PARAMETERS:

		none

	FUNCTION VALUE:

		none

*/
/******************************************************************/
handle_trace( mode, p_set )
int mode;
unsigned long *p_set;
{
	int_pair result;
	FILE *tracefp = NULL;

	if( !connected( client.cl ) ) return;

	if( ( mode == TRACE_MODE ) && !in_debug_mode() ) return;

	if( ( mode == TRACE_MODE ) && !*rpc_get_jm_status_1( NULL, client.cl ) )
		fprintf( stdout, STR_MUST_BE_STOPPED );
	else if( ( mode == VIEW_MODE ) && *rpc_get_jm_status_1( NULL, client.cl ) )
		fprintf( stdout, STR_MUST_BE_RUNNING );
	else
	{
		if( var.display )
			if( qmon_init() )
				return;

		if( var.output != NULL )
		{
			if( ( tracefp = fopen( var.output, "w" ) ) == NULL )
			{
				fprintf( stdout, STR_FOPEN_ERR, var.output, strerror(errno) );
				return;
			}
		}

		if( mode == VIEW_MODE )
			rpc_stop_jm_1( NULL, client.cl );

		if( mode == TRACE_MODE )
			fprintf( stdout, STR_TO_STOP_JOB );
		else
			fprintf( stdout, STR_TO_STOP_VIEW );

		Interrupt = 0;
		for(;;)
		{
			handle_step( 1, TRACE_MODE, tracefp, p_set );
			result = *rpc_get_location_1( NULL, client.cl );
			if( result.x != 0 )
				break;
			if( Interrupt == 1 )
				break;
		}

		/* print out location */
		if( ( mode == TRACE_MODE ) ||
		    ( ( mode == VIEW_MODE ) && result.x != 0 ) )
		{
			fprintf( stdout, STR_JOB_STOPPED );
			handle_where();
			dump_info();
		}
		/* restart the JM if we are in VIEW mode */
		else
			rpc_continue_jm_1( NULL, client.cl );
	}
}
