/*

 Author:	Roberto Ullfig
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::ULLFIG
			Internet: ullfig@fnrobo.fnal.gov

			Copyright (c) 1993 Universities Research Association, Inc.
			All Rights Reserved 


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

#include "eventnames.h"
#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Xos.h>
#include <X11/Xatom.h>

#include <stdio.h>
#include <setjmp.h>

Display		*display = NULL;
Window		win = 0;
int			screen;
GC			gc;
unsigned int width, height;

int channel[2];

extern int num_procs;
#if defined(MIPS) || defined(R3000)
jmp_buf env2;
#else
sigjmp_buf env2;
#endif

#define MAX_COLORS 5
int colors[ MAX_COLORS ];
static char *name[] = { 
	"black", "white", "SkyBlue4", "DarkGreen", "grey20" };

int xio_error_handler( disp, error )
Display *disp;
XErrorEvent *error;
{
	char text[64];
	XGetErrorText( display, error->error_code, text, 64 );
	printf("XIO Error %s %d\n",text, error->error_code);
	printf("          %d\n",error->type);
	printf("          %d\n",error->serial);
	printf("          %d\n",error->request_code);
	printf("          %d\n",error->minor_code);
	win = 0;
	siglongjmp( env2, 1 );
}

int x_error_handler( disp, error )
Display *disp;
XErrorEvent *error;
{
	char text[64];
	XGetErrorText( display, error->error_code, text, 64 );
	printf("X Error %s %d\n",text, error->error_code);
	win = 0;
	siglongjmp( env2, 1 );
}

int
create_window()
{
	int pid;
	XEvent report;
	char buffer[32];

	if( display == NULL )
	{
		/*
		XSetErrorHandler( x_error_handler );
		XSetIOErrorHandler( xio_error_handler );
	*/

		if( ( display = XOpenDisplay( NULL ) ) == NULL )
		{
			printf( "jmdb: cannot connect to X server %s\n",
			    XDisplayName( NULL ) );
			return( 1 );
		}
		screen = DefaultScreen( display );
		get_colors( colors );
		XSynchronize( display, True );
	}
	else
		return( 0 );


	/*
	sigsetjmp( env2, 1 );
	*/
	if( win == 0 )
	{
		width = 50*num_procs+55;
		height = 100;
		win = XCreateSimpleWindow( display, RootWindow( display, screen ),
		    0, 0, width, height, 1,
		    BlackPixel( display, screen ),
		    WhitePixel( display, screen ) );

		gc = XCreateGC( display, win, 0, 0 );

		window_attributes( display, win );

		XMapWindow( display, win );

		/*
		pipe( channel );

		pid = fork();

		if( pid == 0 )
		{
			long event_mask = ExposureMask | StructureNotifyMask;
			XSelectInput( display, win, event_mask );

			for(;;)
			{
				int i;

				XMaskEvent( display, ExposureMask | StructureNotifyMask,
					&report );
				printf("EVENT %s\n",event_names[report.type]);
				fflush(stdout);
				write( channel[1], event_names[ report.type ], 32 );
			}
		}
		else
*/
		window_display( display, win );

	}

	/*
	printf("return\n");
	fflush(stdout);
	*/
	return( 0 );
}

fill_rectangle( int x, int y, int w, int h, int color )
{
	XSetForeground( display, gc, colors[ color ] );
	XFillRectangle( display, win, gc, x, y, w, h );
}

draw_rectangle( int x, int y, int w, int h, int color )
{
	XSetForeground( display, gc, colors[ color ] );
	XDrawRectangle( display, win, gc, x, y, w, h );
}

draw_box( int x, int y, int w, int h, int color1, int color2 )
{
	XPoint pts[3];
	XSetForeground( display, gc, colors[ color1 ] );
	pts[0].x = x - 1;
	pts[0].y = y + h;
	pts[1].x = x - 1;
	pts[1].y = y - 1;
	pts[2].x = x + w;
	pts[2].y = y - 1;
	XDrawLines( display, win, gc, pts, 3, CoordModeOrigin );
	XSetForeground( display, gc, colors[ color2 ] );
	pts[0].x = x - 1;
	pts[0].y = y + h;
	pts[1].x = x + w;
	pts[1].y = y + h;
	pts[2].x = x + w;
	pts[2].y = y - 1;
	XDrawLines( display, win, gc, pts, 3, CoordModeOrigin );
}

draw_string( int x, int y, char *string, int color )
{
	XSetForeground( display, gc, colors[ color ] );
	XDrawString( display, win, gc, x, y, string, strlen( string ) );
}

update_display()
{
	/*
	printf("update_display\n");
	fflush(stdout);
	*/
	XFlush( display );
}

int get_colors( colors )
int colors[];
{
	int default_depth;
	Visual *default_visual;
	Colormap default_cmap;
	XColor exact_def;
	int ncolors = 0;
	int i = 5;
	XVisualInfo visual_info;

	default_depth = DefaultDepth( display, screen );
	default_visual = DefaultVisual( display, screen );
	default_cmap = DefaultColormap( display, screen );
	if( default_depth == 1 )
	{
		colors[ 0 ] = WhitePixel( display, screen );
		colors[ 1 ] = BlackPixel( display, screen );
		return( 0 );
	}

	while( !XMatchVisualInfo( display, screen, default_depth,
	    i--, &visual_info ) )

		if( i < StaticColor )
		{
			colors[ 0 ] = WhitePixel( display, screen );
			colors[ 1 ] = BlackPixel( display, screen );
			return( 0 );
		}

	colors[ 0 ] = BlackPixel( display, screen );
	colors[ 1 ] = WhitePixel( display, screen );
	for( i = 2; i < MAX_COLORS; i++ )
	{
		if( !XParseColor( display, default_cmap, name[i], &exact_def ) )
			return( -1 );

		if( !XAllocColor( display, default_cmap, &exact_def ) )
			return( -1 );

		colors[i] = exact_def.pixel;
		ncolors++;
	}

	return( 1 );
}

window_attributes( display, win )
Display *display;
Window win;
{
	char *window_name	= "Process Queue Monitor";
	char *icon_name		= "Process Queue Monitor";
	long event_mask		= ExposureMask | StructureNotifyMask;
	unsigned long			valuemask;
	XTextProperty			windowName, iconName;

	XStringListToTextProperty( &window_name, 1, &windowName );
	XStringListToTextProperty( &icon_name, 1, &iconName );
	XSetWMProperties( display, win, &windowName, &iconName, NULL, 0,
	    NULL, NULL, NULL );

	XSetWindowBackground( display, win, colors[ 2 ] );

	XSelectInput( display, win, event_mask );
}

window_display( display, win )
Display *display;
Window win;
{
	char buffer[32];
	fd_set fdset;
	XEvent report;

	for(;;)
	{
		XNextEvent( display, &report );
		if( report.type == Expose )
			break;
	}
	/*
	for(;;)
	{
		read( channel[0], buffer, 32 );

		printf("BUFFER1 %s.\n",buffer);
		fflush(stdout);

		if( !strncmp( buffer, "Expose", 6 ) )
			break;
	}
*/
}

window_resize( new_height )
int new_height;
{
	char buffer[32];
	XEvent report;
	XSizeHints size_hints;

	if( new_height == height )
		return;
	height = new_height;
	size_hints.flags = PMinSize | PMaxSize;
	size_hints.min_width = width;
	size_hints.max_width = width;
	size_hints.min_height = height;
	size_hints.max_height = height;
	XSetWMNormalHints( display, win, &size_hints );

	XResizeWindow( display, win, width, height );

	for(;;)
	{
		XNextEvent( display, &report );
		if( report.type == ConfigureNotify )
			break;
	}
	/*
	for(;;)
	{
		read( channel[0], buffer, 32 );
		printf("BUFFER2 %s.\n",buffer);
		fflush(stdout);
		if( !strcmp( buffer, "ConfigureNotify" ) )
			break;
	}
*/
}
