/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/*****************/
/* Include Files */
/*****************/

#include <signal.h>
#include <stdio.h>
#include "../inc/perr.h"
#include "../inc/param.h"
#include "../inc/shm.h"
#include "../inc/xfunc.h"


#define FACILITY "SHM_LOOK"

int num_user_proc;
int num_shm_proc;
int num_local_proc;
int num_semaphores;

/* Shared memory globals */

int *acp_nprocs;
int *acp_semids;
struct procmap *acp_pmap;
struct shmmap *acp_smap;
cps_message *acp_shmem;
cps_msg *acp_sshmem;
extern struct qelem *qelem;
extern int *q_head;
extern int *q_tail;
extern int *q_block;

/* Semaphore ids */

extern int q_semid;

display_error(str)
char *str;
{
	printf("%s\n",str);
}

main(argc,argv) 
int argc;
char **argv;
{
	char *base_addr,*addr;

	int i,j;

	char fname[80];

	FILE *fopen(),*fp;

	/*
	int shmid[32];
	*/
	int shmid[ MAX_SHM_PROCESSES ];
	int shmindex;

	char cmd[80];

	if (argc != 2) {
		printf("Usage: shm_look <job_name>\n");
		exit(0);
	}

	sprintf(fname,"%s.shm",argv[1]);
	fp = fopen(fname,"r");
	i = 0;
	while (fscanf(fp,"%d",&shmid[i++]) != EOF);
	fclose(fp);

/* Attach to the shared memory */

	for (j=0; j<i; j++) {
		if ((int)(base_addr = shm_attach(shmid[j])) != -1) break;
	}
	if (j == i) exit(0);
	shmindex = j;

/* Initialize pointers into shared memory */

	addr = base_addr;
	acp_nprocs = (int *) addr;
	addr += 3 * sizeof(int);
	num_user_proc = acp_nprocs[0];
	num_shm_proc = acp_nprocs[1];
	num_local_proc = acp_nprocs[2];
	num_semaphores = num_local_proc + 3;

	acp_semids = (int *) addr;
	addr += 2 * sizeof(int);
	q_semid = acp_semids[0];

	acp_pmap = (struct procmap *) addr;
	addr += num_user_proc * sizeof(struct procmap);
	acp_smap = (struct shmmap *) addr;
	addr += (num_shm_proc + 1) * sizeof(struct shmmap);  /* One for the JM */

	acp_shmem = (struct shmem *) addr;
	addr += SHM_FUNC(num_semaphores) * sizeof(cps_message);
	acp_sshmem = (struct shmem_s *) addr;
	addr += S_SHM_FUNC(num_semaphores) * sizeof(cps_msg);

	qelem = (struct qelem *) addr;
	addr += (SHM_FUNC(num_semaphores) + S_SHM_FUNC(num_semaphores))
		* sizeof(struct qelem);
	q_head = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_tail = (int *) addr;
	addr += num_semaphores * sizeof(int *);
	q_block = (int *) addr;

/* Kill the shared memory manager */

	for (i=0; i<=num_shm_proc; i++) {
		if (acp_smap[i].pid != 0) {
			sprintf(cmd,"kill -9 %d",acp_smap[i].pid);
			system(cmd);
		}
	}

/* Kill each local user */

	for (i=0; i<num_user_proc; i++) {
		if (acp_pmap[i].local) {
			sprintf(cmd,"kill -9 %d",acp_pmap[i].pid);
			system(cmd);
		}
	}

/* Remove the shared memory and semaphores */

	sprintf(cmd,"ipcrm -s %d",acp_semids[0]);
	system(cmd);

	sprintf(cmd,"ipcrm -m %d",shmid[shmindex]);
	system(cmd);

}
