/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/****************************************************************************/
/*									    */
/*									    */
/*			    Error Message Utility			    */
/*									    */
/*									    */
/*									    */
/****************************************************************************/

#include <stdio.h>
#ifdef VMS
#include errno
#include time
#else
#include <sys/types.h>
#include <time.h>
extern int errno;
extern char *sys_errlist[];
extern int sys_nerr;
#endif
#include "../inc/perr.h"

int suppress_errors = 0;

static int suppress_ptr = 0;
static int suppress_stk[12] = {
	0,0,0,0,0,0,0,0,0,0,0,0};

int set_suppress(x)
int x;
{
	suppress_stk[suppress_ptr] = suppress_errors;
	suppress_ptr++;
	suppress_errors = x;
}

int restore_suppress()
{
	if (suppress_ptr != 0) suppress_ptr--;
	suppress_errors = suppress_stk[suppress_ptr];
}

char *get_sys_msg(errnum)
int errnum;
{
	static char sys_msg[132];

#ifndef VMS

	if ((errnum > 0) && (errnum < sys_nerr))
		strcpy(sys_msg,sys_errlist[errnum]);
	else
		strcpy(sys_msg,"unknown error");

	return(sys_msg);

#else

	sprintf(sys_msg,"system error");
	return(sys_msg);

#endif
}

/******************************************************************/
/*
    FUNCTION NAME:

      printerr

  DESCRIPTION:

      Displays an error message.

  FORMAL PARAMETERS:

      char *facility - A string identifying the software module from which 
		       the error came from.

      int severity - One of four severities - INFO_MSG, WARNING_MSG, ERROR_MSG,
		     and FATAL_MSG.

      char *message - Message describing the error.  If this parameter is the
		      NULL pointer, then a system error message will be printed
		      using the value of the global variable errno.

      char *user_action - Message telling the user what can be done to correct
			  the problem.  If this is the NULL pointer, then no
			  user action message will be printed.

      int index - Unique index isolating where the error message came from
		  in code.

								  */
/******************************************************************/

void printerr(facility,severity,message,user_action,index)
char *facility,*message,*user_action;
int severity,index;
{
	static char errmsg[512];

	char *sys_msg;

	char id[20];

	int i;
	char *m;

	long tloc;
	char timbuf[32];

	time(&tloc);
	cftime(timbuf,"%b %d %T",&tloc);

	/* If error messages are being suppressed, just return */

	if (suppress_errors) return;

	/* Format the error */

	sprintf(errmsg,"%s %s",timbuf,facility);

	if (severity == INFO_MSG)
		strcat (errmsg, "/I");
	else if (severity == WARNING_MSG)
		strcat (errmsg, "/W");
	else if (severity == ERROR_MSG)
		strcat (errmsg, "/E");
	else if (severity == FATAL_MSG)
		strcat (errmsg, "/F");

	sprintf(id,"%d",index);
	strcat(errmsg,id);
	strcat(errmsg,": ");

	i = strlen(errmsg);
	m = &errmsg[i];
	if (message == NULL) {  /* Get the system message */

		sys_msg = get_sys_msg(errno);
		strcat(errmsg,sys_msg);
		errno = 0;

	}
	else
		strcat(errmsg,message);

	display_error(errmsg);

	if (user_action != NULL) {
		sprintf(errmsg,"  User Action: %s",user_action);
		display_error(errmsg);
	}

}


#if defined(VMS) || defined(AIX) || defined(SUNOS) || defined(ULTRIX) || defined(HPUX)
cftime(timbuf,str,tloc)
char *timbuf,*str;
long *tloc;
{
	char *ctime(),*strchr();
	char *c;

	strcpy(timbuf,ctime(tloc));
	c = strchr(timbuf,'\n');
	if (c != NULL) *c = '\0';
}
#endif
