/* XCPS_SYSMON.H V2.0						7 Jan, 1993

 Author:        Bill Meyer
                Fermi National Accelerator Laboratory - Mail Station 369
                P.O Box 500
                Batavia, IL 60510
                Internet: meyer@fnfaus.fnal.gov

 Copyright (c) 1993 Universities Research Association, Inc.
       All Rights Reserved


 This material resulted from work developed under a Government Contract and
 is subject to the following license:  The Government retains a paid-up,
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative
 works, perform publicly and display publicly by or for the Government,
 including the right to distribute to other Government contractors.  Neither
 the United States nor the United States Department of Energy, nor any of
 their employees, makes any warrenty, express or implied, or assumes any
 legal liability or responsibility for the accuracy, completeness, or
 usefulness of any information, apparatus, product, or process disclosed, or
 represents that its use would not infringe privately owned rights.

*/

#define _VirtKeys_h
#include <Xm/Xm.h>
#include <Xm/ArrowB.h>
#include <Xm/BulletinB.h>
#include <Xm/Form.h>
#include <Xm/CascadeB.h>
#include <Xm/DrawingA.h>
#include <Xm/Frame.h>
#include <Xm/Label.h>
#include <Xm/MainW.h>
#include <Xm/MenuShell.h>
#include <Xm/MessageB.h>
#include <Xm/PushB.h>
#include <Xm/PushBG.h>
#include <Xm/RowColumn.h>
#include <Xm/Scale.h>
#include <Xm/ScrolledW.h>
#include <Xm/SeparatoG.h>
#include <Xm/SelectioB.h>
#include <X11/Shell.h>
#include <Xm/ToggleB.h>
#include <Xm/TextF.h>

#define FACILITY "XSYSMON"
extern struct database db;
extern char *getenv();

#define P_CPU           0
#define P_PAGE          1
#define P_SWAP          2
#define P_INTR          3
#define P_DISK          4
#define P_CNTXT         5
#define P_LOAD          6
#define P_PKTS          7
#define P_IPKTS         8
#define P_OPKTS         9
#define P_COLLS         10
#define P_ERRS          11
#define MAXITEMS        12

#ifdef AIX
#define CP_USER 0
#define CP_NICE 1
#define CP_SYS  2
#define CP_IDLE 3
#endif

#define max(a, b) ((a) > (b) ? (a) : (b))

#define LABEL_WIDTH             117
#define LABEL_HEIGHT            30
#define LABEL_VSPACING          5
#define DRAWINGAREA_WIDTH       117
#define DRAWINGAREA_HEIGHT      117
#define CELL_WIDTH              LABEL_WIDTH+5
#define CELL_HEIGHT             LABEL_HEIGHT+LABEL_VSPACING+DRAWINGAREA_HEIGHT+5


#define FROM_FILE       1
#define FROM_PM         2
#define	UNDEFINED	3

int update = 5, old_update = 5, total_cpus = 0;
int total_update = 0;
int option = P_CPU, input_option = UNDEFINED;

struct prod_sys *p;

struct CPULIST {
    char name[40];
	Boolean connect; /* testando */
    CLIENT *client;
    struct sockaddr_in server_addr;
    struct statstime stats_info, prev_v;
    enum clnt_stat status;
    int sock;
    float cdata[MAXITEMS]; /* Currently read data  */
    float pdata[MAXITEMS]; /* Previously read data */
    Widget bulletinboard, drawingarea_frame, drawingarea, label_frame, label;
    Pixmap pixmap;
    GC gc;
	int active;
    struct CPULIST *next;
} *head;

struct {
    int red, green, blue;
    unsigned long pixel;
} colors[] = {  {0xff, 0xff, 0x00, NULL}, /* PEAK color */
                {0xff, 0x80, 0x80, NULL}, /* FILL color */
				/* {0xcb, 0x4a, 0x95, NULL},*/
                {0x00, 0x00, 0xff, NULL}, /* CELL background color */
                {0xff, 0x00, 0x00, NULL}, /* WARNING color */
};

#define PEAK            0
#define FILL            1
#define CELL            2
#define WARNING         3
#define NUM_COLORS      4

Widget CreatemainWindow();
void get_data();
void display_data();
void redraw();
void reset();
void get_cpulist_from_file();
void get_cpulist_from_pm();
void acpu();
void rcpu();
void truncate();
void destroy_list();
void establish_remote_connections();
void processOptions();
void help_handler();
void DestroyPopup();
void sigint_handler();
static caddr_t CONVERT();
static int compare();
void sysmon_init();

Arg args[15];
Display *display;
Screen *screen;
XGCValues gcv;

static char *menuOptions[] = {
        "CPU Utilization", "Page", "Swap", "Interrupt", "Disk",
        "Context Switching", "Load", "Total Packets", "Incoming Packets",
        "Outgoing Packets", "Collisions", "Errors", "Help", "Exit"};

float maxvalue[MAXITEMS];
XtIntervalId update_id;

/* Global Widgets */

Widget AppShell;
Widget MainWindow;
Widget axis_minimum_label;
Widget axis_maximum_label;
Widget update_inc;
Widget update_dec;
Widget PopupShell;

/* Six Widgets make up the title block.  Store pointers in an array */

Widget titleblock[6];
#define HOST    0
#define OPTION  1
#define PSNAME  2
#define JMNAME  3
#define UPDATE  4
#define TIME    5

XtAppContext context;

char HelpText[] =
 "NAME: CPS_XPSMON\
\n \
\nUSAGE: xcps_sysmon -h [-f <filename>][-p <production_system>][-u <update>] \
\n \
\nDESCRIPTION:\
\n     The xcps_sysmon command graphically displays current information about\
\nselected CPUs as they are running.  xcps_sysmon can get it's cpu list from \
\nthree sources:\n\
\n \
\n        1. A specific production system                                   \
\n        2. A file specified by the -f command-line option                 \
\n        3. A default file named .sysmon stored in the user's home directory\
\n \
\n     If the -p <production_system> option is used, the cpu list will be   \
\nobtained automatically from the PM and will monitor the cpus specifically \
\nallocated to that production system.  The -f <filename> option allows the \
\nuser to specify a file which contains a list of cpus.  A sample file looks\
\nlike this:\
\n \
\n        fnsfa\
\n        fnsfb\
\n        fnsfc\
\n \
\n        and so on...\
\n \
\n    If neither of the above two options are given, then xcps_sysmon will   \
\nsearch the user's home directory for a file named .sysmon .  The .sysmon    \
\nfile also contains a list of cpus in the above format.\
\n \
\n    Once xcps_sysmon obtains its list of cpus, it creates a connection with\
\na server running on each cpu.  If the connection, for some reason, cannot \
\nbe made on a particular cpu, it will be skipped and will not be displayed \
\non the screen list.\
\n \
\n    xcps_sysmon creates connections with the rstatd daemon which is running\
\non each cpu.  rstatd is a daemon that runs on almost any UNIX machine.    \
\nSpecial permissions are not required to connect to the server, hence      \
\nnearly any machine reachable through the internet can be monitored        \
\nindependent of machine type; SGI, IBM, SUN, HP, etc...                    \
\n \
\n    xcps_sysmon updates the screen every 5 seconds, or 20 seconds if there \
\nare more than 20 cpus in the list.  This default can be changed by using  \
\nthe -u <update> option on the command-line or during execution.           \
\n \
\nxcps_sysmon reports the following information:\
\n \
\n    CPU UTILIZATION:   Percentage of CPU being utilized                   \
\n    PAGE:              Paging activity in pages per second                \
\n    SWAP:              Jobs swapped per second                            \
\n    INTERRUPT:         Number of device interrupts per second             \
\n    DISK:              Disk traffic in transfers per second               \
\n    CONTEXT SWITCHING: Number of context switches per second              \
\n    LOAD:              Avg number of jobs running over the last minute    \
\n    TOTAL PACKETS:     Total packets on the ethernet per second.          \
\n    INCOMING PACKETS:  Incoming packet transfers per second.              \
\n    OUTGOING PACKETS:  Outgoing packet transfers per second.              \
\n    COLLISIONS:        Collisions/second detected on the ethernet         \
\n    ERRORS:            Errors per second on incoming packets              \
\n \
\n    Occasionally, a red X will mark the graph of a particular node.  This \
\nindicates that xcps_sysmon could not retrieve information from that node   \
\nand timed out.  The X will continue to be displayed as long as data cannot\
\nbe retrieved from the remote node.					    \
\n \
\n                           AUTHOR:  Bill Meyer\
\n \
\n                            Copyright(c) 1993\
\n                  Fermi National Accelerator Laboratory\
\n                           All Rights Reserved";
