/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.
			All Rights Reserved


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/********************************************************************/
/*																	*/
/*																	*/
/*				ERROR.C												*/
/*																	*/
/*																	*/
/********************************************************************/

#include "all.h"
#include <signal.h>


/******************************************************************/
/*
	FUNCTION NAME:

	  cps_stop_job

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_terminate_job(text)
char *text;
{
	struct qelem *q;
	cps_msg *m;

/* Do not get any more messages */

	cps_sighold();

/* Flush standard output */

	fflush(stdout);

/* Get a buffer, build the message, and send it to the JM */

	q = acp_q_pop1(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_usr_abort_job(m,acp_local_lpn,JM_LPN,text);
	acp_q_push1(acp_shm_qnum,q);
	acp_signal_shm();

/* Stop this process */
/*
	acp_pmap[acp_local_lpn-FIRST_USER_LPN].status = 1;

	exit(0);
*/

/* Wait until we are killed */

	/*
	while (1) sleep(100);
	*/
	while (1) sleep(1);

}

cps_stop_job()
{
	acp_fntype = "cps_";
	acp_func = "stop_job";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_STOP_JOB )
	acp_terminate_job("called cps_stop_job");
}

cps_stop_job_()
{
	cps_stop_job();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_stop_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_process_exit(err,text)
int err;
char *text;
{
	struct qelem *q;
	cps_msg *m;

	int i;

/* Do not get any more messages */

	cps_sighold();

/* Flush standard output */

	fflush(stdout);

/* Get a buffer, build the message, and send it to the JM */

	q = acp_q_pop1(S_FREE_Q);
	m = SMALL_MSG(q);
	if (err)
		cps_bld_usr_proc_dead(m,acp_local_lpn,JM_LPN,acp_local_lpn,text);
	else
		cps_bld_usr_finished(m,acp_local_lpn,JM_LPN,acp_local_lpn,text);
	acp_q_push1(acp_shm_qnum,q);
	acp_signal_shm();

/* Block our queue from receiving any more messages */

	q_block[acp_local_qnum] = -1;

/* Discard all messages on our queue and put them back on the free queue */

	while (q_head[acp_local_qnum] != -1) {
		q = acp_q_pop1(acp_local_qnum);
		acp_q_push1(S_FREE_Q,q);
	}

	if (err)
		acp_pmap[acp_local_lpn-FIRST_USER_LPN].status = 0;  /* we are dead */
	else
		acp_pmap[acp_local_lpn-FIRST_USER_LPN].status = 1;  /* we are done */

	exit(0);
}

cps_stop_process()
{
	acp_fntype = "cps_";
	acp_func = "stop_process";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_STOP_PROCESS )
	acp_process_exit(0,"Called cps_stop_process");
}

cps_stop_process_()
{
	cps_stop_process();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_sleep_process

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_sleep_process()
{
	/*
	while (1) sleep(100);
	*/
	while (1) sleep(1);
}

cps_sleep_process_()
{
	cps_sleep_process();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_handle_error

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_handle_error1(action,text)
int action;
char *text;
{
	int a;
	a = action;
	cps_handle_error(&a,text);
}

cps_handle_error(action1,text)
int *action1;
char *text;
{
	int action;
	char errmsg[256];

	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_HANDLE_ERROR )

/* Do not set acp_fntype or acp_func, these will be set to the last routine
   called */

/* If no error, then return */

	if (acp_return_status == CPS_OK) return;

	if (text[0] == '\0')
		sprintf(errmsg,"%s%s: %s",acp_fntype,acp_func,acp_error_info);
	else
		sprintf(errmsg,"%s%s: %s; %s",acp_fntype,acp_func,acp_error_info,text);

	action = *action1;

	select_action:
	switch (action) {

		case CPS_IGNORE_ERROR :

			break;

		case CPS_REPORT_TO_JOB_MANAGER :

			acp_log_msg(errmsg);
			break;

		case CPS_KILL_PROCESSES :

			acpsta_fatal_proc_error(acp_set_of_procs,errmsg);
			break;

		case CPS_KILL_AREA_PROCESSES :

			break;

		case CPS_ABORT_JOB :

			acp_terminate_job(errmsg);
			break;

		case CPS_DEFAULT_ACTION :

			switch (acp_return_status) {
				default :
					action = CPS_REPORT_TO_JOB_MANAGER;
					goto select_action;
			}
			break;

		default :

			break;
	}

	acp_clear_error_info();
	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_fatal_process_error

  DESCRIPTION:

	  First sends messages to the JM for each process other than this
	  one which needs to be stopped.  The JM will broadcast a 
	  usr_proc_dead message to each SHM for each process.
	  If this process is among the set of processes to stop, tells the
	  JM that it is committing suicide and then exits.

  FORMAL PARAMETERS:

	  int sop - set of processes

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_fatal_process_error(sop,text)
unsigned long *sop;
char *text;
{
	acp_fntype = "cps_";
	acpsta_fatal_process_error(sop,text);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

acp_fatal_process_error_(sop,text)
unsigned long *sop;
char *text;
{
	acp_fatal_process_error(sop,text);
}

int acpsta_fatal_process_error_(sop,text)
unsigned long *sop;
char *text;
{
	return(acpsta_fatal_process_error(sop,text));
}

int acpsta_fatal_process_error(sop1,text)
unsigned long *sop1;
char *text;
{
/* First two lines of every acpsta_ function */

	acp_clear_error_info();
	acp_func = "fatal_process_error";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_HANDLE_ERROR )
	return(acpsta_fatal_proc_error(sop1,text));
}

int acpsta_fatal_proc_error(sop1,text)
unsigned long *sop1;
char *text;
{
	int s;

	static unsigned long *sop = NULL;
	int this_process = 0;
	int *mem;

	struct qelem *q;
	cps_msg *m;
	int i,j;

	char errmsg[80];

/* Translate a set of classes into a set of processes */

	if (sop == NULL) sop = create_empty_set(MAX_PROCESSES);
	acp_translate_sop(*sop1,sop);

/* Remove this process from the set of processes */

	if (set_member(acp_local_lpn,sop)) {
		remove_from_set(acp_local_lpn,sop);
		this_process = -1;
	}

/* Check the size of the message */

	if (strlen(text)+1 > SMALL_MSG_SIZE) text[0] = '\0';

/* For each user process other than our own, tell the JM */
/* The JM will then broadcast the message to all SHMs */

	mem = set_query(sop,MAX_PROCESSES);
	i = 0;
	while (mem[i] != 0) {

/* Get a buffer, build the message, and send it to the JM */

		q = acp_q_pop2(S_FREE_Q);
		m = SMALL_MSG(q);
		cps_bld_usr_proc_dead(m,acp_local_lpn,JM_LPN,mem[i],text);
		acp_q_push2(acp_shm_qnum,q);
		acp_signal_shm();

		i++;
	}

	if (this_process == 0) return(CPS_OK);

	acp_process_exit(-1,text);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_fatal_job_error

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_fatal_job_error(text)
char *text;
{
	acp_fntype = "cps_";
	acp_func = "fatal_job_error";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_HANDLE_ERROR )

/* Check the size of the message */

	if (strlen(text)+1 > SMALL_MSG_SIZE) text[0] = '\0';

	acp_terminate_job(text);
}

acp_fatal_job_error_(text)
char *text;
{
	acp_fatal_job_error(text);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_fatal_area_error

  DESCRIPTION:

	  Sends a message to the SHM process to stop all processes on this
	  CPU and notify the JM.  The JM will then decide on whether or not
	  to stop the job.  The text supplied will be printed to the JM's
	  terminal and log file.

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_fatal_area_error(text)
char *text;
{

}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_clear_error_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_clear_error_info()
{
	acp_set_of_procs = CPS_THIS_PROCESS;
	acp_return_status = CPS_OK;
	acp_error_info[0] = '\0';
}

/*#**********************************************************************
*									*
* cps_log(text)								*
*									*
* Sends a log-entry message to the Job Manager				*
*									*
* Inputs:								*
*									*
*	text = character string that contains the log entry		*
*									*
*************************************************************************/

cps_log(text)
char *text;
{
	acp_fntype = "cps_";
	acpsta_log(text);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_log_(text)
char *text;
{
	cps_log(text);
}

int acpsta_log_(text)
char *text;
{
	return(acpsta_log(text));
}

int acpsta_log(text)
char *text;
{
	acp_clear_error_info();
	acp_func = "log";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_LOG )

/* Check the size of the message */

	if (strlen(text) > 255) {
		sprintf(acp_error_info,
			"Message too long; must be < 256 characters");
		acp_return_status = CPS_MESSAGE_TOO_LONG;
		return(CPS_MESSAGE_TOO_LONG);
	}

	return(acp_log_msg(text));

}

int acp_log_msg(text)
char *text;
{
	static int route = 0;

/* Send the message to the JM (through the rexec socket) */

	if (route == 0)
	if (printf("%s\n",text) < 0) 
		route = 1;
	else {
		fflush(stdout);
		return(CPS_OK);
	}

	if (route == 1)
	return(acpsta_log1(text));
}

int acpsta_log1(text)
char *text;
{
	struct qelem *q;
	cps_msg *m;

/* Get a buffer, build the message, and send it to the JM */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_usr_log(m,acp_local_lpn,JM_LPN,text);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_change_action

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acpsta_change_action(x)
int *x;
{
	acp_clear_error_info();
	acp_func = "change_action";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CHANGE_ACTION )
	return(CPS_OK);
}

int acpsta_change_action_(x)
int *x;
{
	return(acpsta_change_action(x));
}

acp_change_action(x)
int *x;
{
	acp_fntype = "cps_";
	acpsta_change_action(x);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

acp_change_action_(x)
int *x;
{
	acp_change_action(x);
}
