/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1991 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				EXPLICIT.C												*/
/*																		*/
/*																		*/
/************************************************************************/


/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>

/* #define DIAG 1 */

#define NQ 17
#define SZ 128

static struct acp_msg_queue free[NQ];

acp_init_msg_queue()
{
	int i,j;

	for (i=0; i<NQ; i++) {
		free[i].head = NULL;
		free[i].tail = NULL;
	}

	acp_xpl.head = NULL;
	acp_xpl.tail = NULL;

	for (i=0; i<21; i++) {
		acp_xpl_ready[i] = 0;
	}
}

struct acp_msg_block *acp_dequeue_block(queue)
struct acp_msg_queue *queue;
{
	struct acp_msg_block *block;

	block = queue->head;
	if (queue->head != NULL) {
		queue->head = queue->head->next;
		if (queue->head == NULL) queue->tail = NULL;
	}
	return(block);
}

acp_queue_block(block,queue)
struct acp_msg_block *block;
struct acp_msg_queue *queue;
{
	if (queue->tail == NULL) {
		queue->head = block;
		queue->tail = block;
		block->next = NULL;
	}
	else {
		queue->tail->next = block;
		queue->tail = block;
		block->next = NULL;
	}
}

struct acp_msg_block *acp_checkout_block(size)
int size;
{
	int i,j;

	struct acp_msg_block *block;

	size--;
	size /= SZ;
	i = size;
	size *= SZ;
	size += SZ;

/* Check to see if there are any blocks of this size or larger */

	for (j=i; j<NQ; j++) {
		if (free[j].head != NULL) {
			return(acp_dequeue_block(&free[j]));
		}
	}

/* No blocks are available, we must allocate one */

	block = (struct acp_msg_block *) hmalloc(sizeof(struct acp_msg_block));
	block->buf = (char *) hmalloc(size);
	block->index = i;
	block->next = NULL;
	return(block);

}

acp_return_block(block)
struct acp_msg_block *block;
{
	int i;
	i = block->index;
	acp_queue_block(block,&free[i]);
}

static int check_msg_type(n)
int n;
{
	if ((n < 1) || (n > MAX_EXPLICIT_MSG)) {
		sprintf(acp_error_info,
			"Message type (%d) out of range; must be between 1 and %d",
			n,MAX_EXPLICIT_MSG);
		acp_return_status = CPS_BAD_MESSAGE_TYPE;
		return(CPS_BAD_MESSAGE_TYPE);
	}
	return(CPS_OK);
}

static int check_byte_count(n)
int n;
{
	if ((n < 4) || (n > MAX_MESSAGE_SIZE)) {
		sprintf(acp_error_info,
			"Message size (%d) out of range; must be between 4 and %d",
			n,MAX_MESSAGE_SIZE);
		acp_return_status = CPS_BYTE_COUNT_EXCEEDED;
		return(CPS_BYTE_COUNT_EXCEEDED);
	}
	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_set_of_message_types

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

unsigned int cps_set_of_message_types (n, 
			 c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20)

int  *n, *c1, *c2, *c3, *c4, *c5, *c6, *c7, *c8, *c9, *c10,	
		 *c11,*c12,*c13,*c14,*c15,*c16,*c17,*c18,*c19,*c20;
{
	unsigned int set = 0;

	int s;

	char errmsg[80];

	acp_clear_error_info();
	acp_fntype = "cps_";
	acp_func = "set_of_message_types";

	if (*n<1)  {						/* Number out of range */
		sprintf(errmsg,
			"cps_set_of_message_types: not enough arguments"
			,*n);
		acp_process_exit(-1,errmsg);
	}
	if (*n>MAX_EXPLICIT_MSG)  {					/* Number out of range */
		sprintf(errmsg,
			"cps_set_of_message_types: too many arguments"
			,*n);
		acp_process_exit(-1,errmsg);
	}


	if ((s = check_msg_type(*c1)) != CPS_OK) return(s);
	add_to_set(*c1,&set);
	if (*n >= 2) {
		if ((s = check_msg_type(*c2)) != CPS_OK) return(s);
		add_to_set(*c2,&set);
	}
	else return(set);
	if (*n >= 3) {
		if ((s = check_msg_type(*c3)) != CPS_OK) return(s);
		add_to_set(*c3,&set);
	}
	else return(set);
	if (*n >= 4) {
		if ((s = check_msg_type(*c4)) != CPS_OK) return(s);
		add_to_set(*c4,&set);
	}
	else return(set);
	if (*n >= 5) {
		if ((s = check_msg_type(*c5)) != CPS_OK) return(s);
		add_to_set(*c5,&set);
	}
	else return(set);
	if (*n >= 6) {
		if ((s = check_msg_type(*c6)) != CPS_OK) return(s);
		add_to_set(*c6,&set);
	}
	else return(set);
	if (*n >= 7) {
		if ((s = check_msg_type(*c7)) != CPS_OK) return(s);
		add_to_set(*c7,&set);
	}
	else return(set);
	if (*n >= 8) {
		if ((s = check_msg_type(*c8)) != CPS_OK) return(s);
		add_to_set(*c8,&set);
	}
	else return(set);
	if (*n >= 9) {
		if ((s = check_msg_type(*c9)) != CPS_OK) return(s);
		add_to_set(*c9,&set);
	}
	else return(set);
	if (*n >= 10) {
		if ((s = check_msg_type(*c10)) != CPS_OK) return(s);
		add_to_set(*c10,&set);
	}
	else return(set);
	if (*n >= 11) {
		if ((s = check_msg_type(*c11)) != CPS_OK) return(s);
		add_to_set(*c11,&set);
	}
	else return(set);
	if (*n >= 12) {
		if ((s = check_msg_type(*c12)) != CPS_OK) return(s);
		add_to_set(*c12,&set);
	}
	else return(set);
	if (*n >= 13) {
		if ((s = check_msg_type(*c13)) != CPS_OK) return(s);
		add_to_set(*c13,&set);
	}
	else return(set);
	if (*n >= 14) {
		if ((s = check_msg_type(*c14)) != CPS_OK) return(s);
		add_to_set(*c14,&set);
	}
	else return(set);
	if (*n >= 15) {
		if ((s = check_msg_type(*c15)) != CPS_OK) return(s);
		add_to_set(*c15,&set);
	}
	else return(set);
	if (*n >= 16) {
		if ((s = check_msg_type(*c16)) != CPS_OK) return(s);
		add_to_set(*c16,&set);
	}
	else return(set);
	if (*n >= 17) {
		if ((s = check_msg_type(*c17)) != CPS_OK) return(s);
		add_to_set(*c17,&set);
	}
	else return(set);
	if (*n >= 18) {
		if ((s = check_msg_type(*c18)) != CPS_OK) return(s);
		add_to_set(*c18,&set);
	}
	else return(set);
	if (*n >= 19) {
		if ((s = check_msg_type(*c19)) != CPS_OK) return(s);
		add_to_set(*c19,&set);
	}
	else return(set);
	if (*n >= 20) {
		if ((s = check_msg_type(*c20)) != CPS_OK) return(s);
		add_to_set(*c20,&set);
	}
	else return(set);
	return(set);
}

unsigned int cps_set_of_message_types_(n, 
			 c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20)

int  *n, *c1, *c2, *c3, *c4, *c5, *c6, *c7, *c8, *c9, *c10,	
		 *c11,*c12,*c13,*c14,*c15,*c16,*c17,*c18,*c19,*c20;
{
	return(cps_set_of_message_types (n, 
			 c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20));
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_transmit_message

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:


  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_transmit_message(sop,mtype,mdata,mcount)
int *sop,*mtype,*mcount;
char *mdata;
{
	acp_fntype = "cps_";
	cpssta_transmit_message(sop,mtype,mdata,mcount);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_transmit_message_(sop,mtype,mdata,mcount)
int *sop,*mtype,*mcount;
char *mdata;
{
	cps_transmit_message(sop,mtype,mdata,mcount);
}

int
cpssta_transmit_message_(sop,mtype,mdata,mcount)
int *sop,*mtype,*mcount;
char *mdata;
{
	return(cpssta_transmit_message(sop,mtype,mdata,mcount));
}

int
cpssta_transmit_message(sop,mtype,mdata,mcount)
int *sop,*mtype,*mcount;
char *mdata;
{
	int som;

	static unsigned long *ssop = NULL;
	static unsigned long *rdy_set = NULL;

	int s,status;

	int *plist;

	struct qelem *q;
	cps_message *m;

	int i,j;

	int size;

	struct acp_msg_block *block;

	acp_clear_error_info();
	acp_func = "transmit_message";
	acp_set_of_procs = *sop;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_TRANSMIT_MESSAGE )

/* Translate CPS set of processes into standard set of processes */

	if (ssop == NULL) {
		ssop = create_empty_set(MAX_PROCESSES);
		rdy_set = create_empty_set(MAX_PROCESSES);
	}

	acp_translate_sop(*sop,ssop);
	set_clear(rdy_set,MAX_PROCESSES);

/* Check the message type */

	if ((s = check_msg_type(*mtype)) != CPS_OK) return(s);

/* Check to see that the message size is a multiple of four */

	if (*mcount % 4) {
		sprintf(acp_error_info,
			"Byte count must be a multiple of four (count=%d)",*mcount);
		acp_return_status = CPS_ILLEGAL_BYTE_COUNT;
		return(CPS_ILLEGAL_BYTE_COUNT);
	}
	size = 6*sizeof(long) + *mcount;

/* Hold signals */

	cps_sighold();

#ifdef DIAG
printf("cps_transmit_message: message type = %d\n",*mtype);
fflush(stdout);
#endif

/* Store explicit messages in queues (if not already requested) */

	plist = set_query(ssop,MAX_PROCESSES);

	i = 0;
	while (plist[i] != 0) {
#ifdef DIAG
printf("cps_transmit_message: ");
printf("storing message for process %d\n",plist[i]);
fflush(stdout);
#endif
		if (size <= SMALL_MSG_SIZE) {
			q = acp_q_pop1(S_FREE_Q);
			m = (cps_message *) SMALL_MSG(q);
		}
		else {
			q = acp_q_pop1(L_FREE_Q);
			m = LARGE_MSG(q);
		}
		cps_bld_explicit_msg(m,acp_local_lpn,plist[i],*mtype,mdata,*mcount);
		if (acp_send_to_user(q,plist[i],1) != CPS_OK)
			acp_log_msg(acp_error_info);
		remove_from_set(plist[i],ssop);
		add_to_set(plist[i],rdy_set);
		i++;
	}

/* Enable interrupts */

	cps_sigrelease();

	status = CPS_OK;

/* Send explicit messages to processes which have already requested them */
/* This code should never be executed */

	plist = set_query(ssop,MAX_PROCESSES);
	if (plist[0] != 0) {
		printerr(FACILITY,ERROR_MSG,
			"Protocol error in explicit message passing",NULL,23);
	}
	return(status);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_check_message

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:


  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int cps_check_message_(som)
int *som;
{
	return(cps_check_message(som));
}

int cps_check_message(som)
int *som;
{
	int *mlist;
	int i;

	acp_fntype = "cps_";
	acp_clear_error_info();
	acp_func = "check_message";
	acp_set_of_procs = CPS_THIS_PROCESS;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CHECK_MESSAGE )

	mlist = set_query(som,8*sizeof(int));
	i = 0;
	while (mlist[i] != 0) {
		if (acp_xpl_ready[mlist[i]] > 0) return(CPS_OK);
		i++;
	}
	return(CPS_NO_MESSAGE_AVAILABLE);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_receive_message

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:


  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

static int acp_xpl_mtype,acp_xpl_src_lpn;

int acp_message_ready(som)
int *som;
{
	cps_message *m;
	struct acp_msg_block *block,*prev_block;
	int mtype;
	int size;
	long *x;

	cps_sighold();
	block = acp_xpl.head;
	prev_block = NULL;
	while( block != NULL )
	{
		m = (cps_message *) block->buf;
		x = (long *) m->msg;
		mtype = *x;
		if( *som & ( 1 << ( mtype - 1 ) ) )
		{
			acp_xpl_mtype = mtype;
			acp_xpl_src_lpn = m->src_lpn;

			size = 4*sizeof(long) + m->size;
			memcpy(acp_xpl_msg, (char *) m, size);

			if( block == acp_xpl.head )
			{
				acp_xpl.head = acp_xpl.head->next;
				if( acp_xpl.head == NULL )
					acp_xpl.tail = NULL;
			}
			else
			{
				if( block == acp_xpl.tail )
					acp_xpl.tail = prev_block;
				prev_block->next = block->next;
			}
			acp_return_block(block);

			cps_sigrelease();
			return(-1);
		}
		prev_block = block;
		block = block->next;
	}
	cps_sigrelease();
	return(0);
}

cps_receive_message(som,p,mtype,mdata,mcount)
int *som,*p,*mtype,*mcount;
char *mdata;
{
	acp_fntype = "cps_";
	cpssta_receive_message(som,p,mtype,mdata,mcount);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

cps_receive_message_(som,p,mtype,mdata,mcount)
int *som,*p,*mtype,*mcount;
char *mdata;
{
	cps_receive_message(som,p,mtype,mdata,mcount);
}

cpssta_receive_message_(som,p,mtype,mdata,mcount)
int *som,*p,*mtype,*mcount;
char *mdata;
{
	return(cpssta_receive_message(som,p,mtype,mdata,mcount));
}

cpssta_receive_message(som,p,mtype,mdata,mcount)
int *som,*p,*mtype,*mcount;
char *mdata;
{
	struct qelem *q;
	cps_msg *m1;
	cps_message *m;
	int som1;

	acp_clear_error_info();
	acp_func = "receive_message";
	acp_set_of_procs = CPS_THIS_PROCESS;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_RECEIVE_MESSAGE )

/* Tell the JM we are waiting for an explicit message */

	q = acp_q_pop2(S_FREE_Q);
	m1 = SMALL_MSG(q);
	cps_bld_mon_update(m1,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
		PSTATE_WAIT_XPL,*som);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait until a message is ready to be read */

	acp_wait_till_true(acp_message_ready,som);
	*p = acp_xpl_src_lpn;
	som1 = 0;
	add_to_set(acp_xpl_mtype,&som1);

/* Tell the JM we have received the explicit message */

	q = acp_q_pop2(S_FREE_Q);
	m1 = SMALL_MSG(q);
	cps_bld_mon_update(m1,acp_local_lpn,JM_LPN,MOP_STATE_CHANGE,acp_local_lpn,
		PSTATE_RUNNING,*som);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Copy message to arguments */

	m = (struct  shmem *) acp_xpl_msg;
	cps_get_explicit_msg(m,mtype,mdata,mcount);
	*p = m->src_lpn;

/* Decrement number of messages waiting to be read */

	acp_xpl_ready[*mtype]--;
#ifdef DIAG
printf("cps_receive_message: ");
printf("treq=$%x, trec=%d, #rdy = %d, lpn=%d\n",*som,*mtype,acp_xpl_ready[*mtype],*p);
fflush(stdout);
#endif
	return(CPS_OK);
}
