/****************************************************************************
 *																			*
 *  Routines for locking a block of data in memory on a acpr3000 node		*
 *																			*
 *  obs : this file must be linked with /usr/lib/libm.a and					*
 *		/usr/lib/libbsd.a													*
 *																			*
 *  Author : Carmem Silva													*
 *  02/11/92																*
 ****************************************************************************/

#include <stdio.h>

#ifdef R3000

#include <sys/sysmips.h>
#include <sys/types.h>
#include <sys/immu.h>
#include <malloc.h>
#include <math.h>
#include <errno.h>
#include <ctype.h>
#include <bsd/netdb.h>
#include <bsd/netinet/in.h>
#include "user_struct.h"
#include "../inc/perr.h"


#define COMPL			255
#define FACILITY		"LIBCPS"
#define EXA_NUMBER	  0xfff
#define BUFFER_SIZE	 4096
#define CPS_ALLOCPHYS	12

/****************************************************************************
 *																			*
 * acp_set_vbbc_info - Routine that locks a block of data in memory and		*	
 *		  and fills the structure acp_vbbc_info								*
 *																			*
 * Sintax:																	*
 *	  acp_set_vbbc_info - (vbbc_info , data_buffer , nbytes)				*
 *		where:																*
 *		 struct acp_vbbc_info *vbbc_info - a pointer to the structure		*
 *						that will be filled									*
 *		 char *data_buffer - Virtual address from the block that			*
 *						will be locked in memory							*
 *		 int nbyte - Number of bytes of the block							*
 *																			*
 * Author: Carmem Silva														*
 * 02/11/92																	*
 *																			*
 ****************************************************************************/


acp_set_vbbc_info(vbbc_info,data_buffer,nbytes)
struct acp_vbbc_info *vbbc_info;
int nbytes;
char *data_buffer;
{
	caddr_t physical_address;
	char *buffer,access;
	int status,i,page_offset;
	int rsize ,buffer_number;
 
	buffer = data_buffer;

/* finds out how many pages this block will occupied */	

	if ((buffer_number = nbytes/BUFFER_SIZE) > 0)
	{
		rsize = drem(nbytes,BUFFER_SIZE);
		if ( rsize > 0)
			buffer_number++;
		else if (rsize < 0)
		{
			rsize = BUFFER_SIZE + rsize;
			buffer_number++;
		} 
	} 
	else
		buffer_number = 1;

	vbbc_info->paddr = (ulong*)malloc((buffer_number+2) * sizeof(unsigned long));
	
/* locks the first page */
 
	access = *buffer;
	status = sysmips(CPS_ALLOCPHYS, &physical_address,buffer);
	if (status != 0)
	{
		printerr(FACILITY,FATAL_MSG,"Error locking page",NULL,32);
		exit();
	}


/* finds out where the blocks begins from the first address of the page */
 
	vbbc_info->paddr[0] = (ulong)buffer & EXA_NUMBER |(ulong)physical_address;	

	page_offset = abs(vbbc_info->paddr[0] - (ulong)physical_address);	 
	vbbc_info->offset = page_offset;
	buffer += BUFFER_SIZE - page_offset;

/* finds out if we have to include another page depending on the first page
	offset */

	if (rsize > 0)
	{
		if ( (rsize += page_offset) > BUFFER_SIZE)
			buffer_number++;
	}		
	else if (page_offset > 0)
		buffer_number++;	
	
/* locks the rest of the pages */

	for ( i=1; i < buffer_number; i++)
	{ 
		access = *buffer;
		status = sysmips(CPS_ALLOCPHYS, &physical_address,buffer);
		if (status != 0)
		{
			printerr (FACILITY,FATAL_MSG,"Error locking pages",NULL,33);
			exit();
		}

		vbbc_info->paddr[i]=(ulong)buffer & EXA_NUMBER|(ulong)physical_address;	
		buffer += BUFFER_SIZE;
	}

	vbbc_info->npaddr = buffer_number;

}

getcratevmeaddress(vbbc_info)
struct acp_vbbc_info *vbbc_info;
{
	static struct hostent *hp;
	ulong local_addr;
	struct in_addr addr;
	char acpcom_name[64],local_hostname[64],local_hostaddr[24];
	int cratevme;


	get_acpcom_address(acpcom_name);
 
/* Get our internet address */

	if ((hp = gethostbyname (acpcom_name))== NULL)
	{
		perror( "can't get hostname ");
		exit();		
	}

	bcopy(hp->h_addr,&addr,hp->h_length);
	strcpy(local_hostaddr,inet_ntoa(addr)); 
	local_addr = inet_network(local_hostaddr);

	cratevme = local_addr & COMPL;
	vbbc_info->crate = cratevme >> 5; 
	vbbc_info->vme_addr = (cratevme & 31) << 3;
	return(0);
}


get_acpcom_address(acpcom_name)
char acpcom_name[64];
{
	FILE *fd;
	int aux,i = 0; 


	if ((fd = fopen ("/specific/enp_address","r")) == NULL)
	{
		retry1:
		if (gethostname(acpcom_name,64) < 0) {
			if (errno == EINTR) goto retry1;
				exit();
		}
	}
	else 
	{
		close(fd);
		if ((fd = fopen ("/specific/acpcom_address","r")) == NULL)
		{
			perror ("can't open file /specific/acpcom_address");
			exit();
		}
		do
		{
			acpcom_name[i] = getc(fd);
			aux = acpcom_name[i];
			i = i + 1;
		} while ( ((isalnum(aux) != 0) || (aux == '.')) && !feof(fd) );
		acpcom_name[i-1] = NULL;		
		close(fd);	
	}
}

#endif

