/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/******************************************************************************/
/* Global variables used to pass information between CPS function calls       */
/******************************************************************************/

FILE *acp_fp;

int acp_init_called = 0;

int acp_local_lpn = 0;
int acp_local_class = 0;
int acp_local_shm = -1;
int acp_shm_lpn = 0;

char acp_local_host[64];

int acp_set_of_procs = 0;		/* CPS-style set of classes */

int acp_return_status = 0;		/* status returned by the last
					   call to an acpsta_ function */

char acp_error_info[132];		/* error message of last acpsta_ call
					   returning an error */

char *acp_fntype = "acpsta_";
char *acp_func = "init";

int acp_local_qnum;			/* Queue number for this process */

int acp_shm_qnum;			/* Number of the queue to use to send
					   messages through the shm */

int num_user_proc;
int num_shm_proc;
int num_local_proc;
int num_semaphores;

/* Shared memory globals */

int *acp_nprocs;
int *acp_semids;
struct procmap *acp_pmap;
struct shmmap *acp_smap;
cps_message *acp_shmem;
cps_msg *acp_sshmem;
extern struct qelem *qelem;
extern int *q_head;
extern int *q_tail;
extern int *q_block;

/* Semaphore ids */

extern int q_semid;

/* Linked list of declared memory blocks */

struct acp_data_block *acp_blocks = NULL;	/* Our process */

struct acp_data_block *acp_remote_blocks[MAX_PROCESSES];

/* Linked list of declared subroutines */

struct acp_remote_sub *acp_subs = NULL;

/* (Dynamic) Class --> Process mappings */

unsigned long *acp_class_proc[MAX_CLASSES+1];	/* Indexed by class number,
						   sets may change when 
						   processes die. */

/* (Static) class --> Process mappings. Never changes after acp_init. */

unsigned long *acp_static_cp[MAX_CLASSES+1];	/* v2.9 */

/* Flag to wait until data request is complete */

int acp_request_complete = 0;

int acp_sync_complete = 0;	/* Sync point complete message */
int acp_wait_complete = 0;	/* acp_wait_queue complete */
int acp_qreq_complete = 0;	/* Dequeue operation complete flag */
int acp_treq_complete = 0;	/* Tape request completion flag */
int acp_rsub_complete = 0;	/* Remote sub declaration has arrived */
int acp_send_complete = 0;
int acp_get_complete = 0;
int acp_send_aborts = 0;	/* Number of aborted sends (new in v2.9) */
int acp_get_aborts = 0;		/* Number of aborted gets (new in v2.9) */

/* Tape mount request globals */

int mount_reply_status;
int mount_reply_pos;
int mount_reply_protection;
char mount_reply_dev[64];
char mount_reply_vsn[32];
char mount_reply_msg[132];

/* Pointer to current remote subroutine server executing */

int acp_return_queue;

/* Declared queues (number of arguments, -1 for undeclared,
				     	 -2 = end-of-queue) */

int acp_queue_declared[MAX_QUEUES];
int *acp_queue_bcount[MAX_QUEUES];	/* Byte count for each arg */
int acp_queue_bsum[MAX_QUEUES];		/* Total byte count */

/* Queue reply information */

int acp_queue_number;				/* Queue we are waiting on */
int acp_process_dequeued;			/* Process dequeued */
char acp_queue_data[MAX_QUEUE_BYTES];		/* Data */

/* Remote subroutine server pending calls */

struct qelem *acp_rms_head = NULL;
struct qelem *acp_rms_tail = NULL;

struct acp_msg_queue acp_rms;

/* Explicit Messages */
struct acp_msg_queue acp_xpl; /* Incoming explicit messages */
char acp_xpl_msg[MAX_MESSAGE_SIZE+16];
int acp_xpl_ready[21]; /* Number of messages waiting to be read of each type */

/* Queued asynchronous data transfers */

struct queued_xfer *acp_qxfer_h = NULL;		/* Queued data transfers (head) */
struct queued_xfer *acp_qxfer_t = NULL;		/* Queued data transfers (tail) */

struct active_qxfer *acp_aqx = NULL;		/* Active queued data xfers */
int acp_next_qx_id = 2;				/* Next active xfer id (must
						   be greater than 1) */

/* High-level asynchronous data transfers */

struct buffer_set *acp_bufset = NULL;


/* Asynchronous data transfer failure count */

int acp_num_aborted_xfers = 0;			/* v2.9 */

/* Internal function calls */

struct acp_msg_block *acp_dequeue_block();
struct acp_msg_block *acp_checkout_block();

/* Port numbers for direct data transfers, accessed by lpn */

int acp_port[MAX_PROCESSES+1];		/* -1 means never available */

/* Data sockets */

int acp_dsock;				/* Socket for accepting connections */
int acp_data_sock[MAX_PROCESSES+1];	/* -1 means no connection */

/* Remotely initiated direct transfers */

struct direct_xfer acp_r_dxfer[MAX_PROCESSES+1];

/* Locally initiated direct transfers */

#define MAX_L_XFERS 32

struct direct_xfer acp_l_dxfer[MAX_PROCESSES+1][MAX_L_XFERS];
int acp_dxfer_ptr[MAX_PROCESSES+1];	/* Current data transfer */
int acp_dxfer_nleft[MAX_PROCESSES+1];	/* Number data transfers left */

char proc_hooks;
int acp_hook_complete;	/* Hook point complete message */
char profile;
