/*

 Author:	Matt Fausey
			Fermi National Accelerator Laboratory - Mail Station 369
			P.O Box 500
			Batavia, IL 60510
			DECnet: FNAL::FAUSEY
			Internet: fausey@fnsgia.fnal.gov

			Copyright (c) 1990 Universities Research Association, Inc.		
			All Rights Reserved													


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,	  
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,	   
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of	 
 their employees, makes any warrenty, express or implied, or assumes any	  
 legal liability or responsibility for the accuracy, completeness, or		 
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*																		*/
/*																		*/
/*				util.c													*/
/*																		*/
/*																		*/
/************************************************************************/

/*****************/
/* Include Files */
/*****************/

#include "all.h"
#include <signal.h>
#include <errno.h>

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_dump_shm

  DESCRIPTION:

	  [@tbs@]

								  */
/******************************************************************/

acp_dump_shm()
{
	int i;

	struct qelem *q;
/*
	fprintf(acp_fp,"Number of user processes = %d\n",acp_nprocs[0]);
	fprintf(acp_fp,"Number of shm processes = %d\n",acp_nprocs[1]);
	fprintf(acp_fp,"Number of local processes = %d\n",acp_nprocs[2]);
	fprintf(acp_fp,"Queue semaphore id = %d\n",acp_semids[0]);
	fprintf(acp_fp,"\nShared Memory Processes\n");
	fprintf(acp_fp,  "-----------------------\n");
	for (i=0; i<=num_shm_proc; i++) {
		fprintf(acp_fp,"(%s,%s,%d) pid=%d semnum=%d sock=%d status=%d\n",
			acp_smap[i].hostname,acp_smap[i].hostaddr,acp_smap[i].portno,
			acp_smap[i].pid,acp_smap[i].semnum,acp_smap[i].sock,
			acp_smap[i].status);
	}

	fprintf(acp_fp,"\nUser Processes\n");
	fprintf(acp_fp,  "---------------\n");
	for (i=0; i<num_user_proc; i++) {
		fprintf(acp_fp,"local=%d pid=%d shmid=%d semnum=%d\n",
			acp_pmap[i].local,acp_pmap[i].pid,acp_pmap[i].shmid,
			acp_pmap[i].semnum);
	}

	fflush(acp_fp);

	fprintf(acp_fp,"\nQueues (%d)\n",num_semaphores);
	fprintf(acp_fp,  "------\n");
	for (i=0; i<num_semaphores; i++) {
		if (q_head[i] == -1)
			q = NULL;
		else
			q = &qelem[q_head[i]];
		fprintf(acp_fp,"%d: ",i);
		while (q != NULL) {
			fprintf(acp_fp,"%d",q->buf);
			if (q->size == 0)
			fprintf(acp_fp,"s ");
			else
			fprintf(acp_fp,"l ");
			if (q->next == -1)
			q = NULL;
			else
			q = &(qelem[q->next]);
		}
		fprintf(acp_fp,"\n");
	}

	fflush(acp_fp);
*/
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_wait_condition

								  */
/******************************************************************/

int acp_wait_condition(x,n)
int *x,n;
{
	sig_wait_ge(SIGUSR1,x,n);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_wait_till_true

								  */
/******************************************************************/

int acp_wait_till_true(func,x)
int (*func)();
int *x;
{
	sig_wait_true(SIGUSR1,func,x);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_signal_shm

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_signal_shm()
{
	if (kill(acp_smap[acp_local_shm].pid,SIGUSR1) == -1) exit(0);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_q_pop2

								  */
/******************************************************************/

struct qelem *acp_q_pop2(i)
int i;
{
	struct qelem *q;

	start:

/* Disable interrupts */

	cps_sighold();

/* Pop a message from the queue */

	q = q_pop_no_wait(i);

/* Enable interrupts */

	cps_sigrelease();

	if (q == NULL) goto start;

	return(q);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_q_pop2_no_wait

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

struct qelem *acp_q_pop2_no_wait(i)
int i;
{
	struct qelem *q;

/* Disable interrupts */

	cps_sighold();

/* Pop a message from the queue */

	q = q_pop_no_wait(i);

/* Enable interrupts */

	cps_sigrelease();

	return(q);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_q_push2

  DESCRIPTION:

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_q_push2(i,q)
int i;
struct qelem *q;
{
	int s;

	cps_sighold();

	s = q_push(i,q);

	cps_sigrelease();

	if (s == -1) {
		sprintf(acp_error_info,"%s%s: Failed to push message buffer",
			acp_fntype,acp_func);
		acp_process_exit(-1,acp_error_info);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_q_pop1

  DESCRIPTION:


  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

struct qelem *acp_q_pop1(i)
int i;
{
	struct qelem *q;

	q = q_pop(i);

	if (q == NULL) {
		sprintf(acp_error_info,"%s%s: Failed to pop message buffer",
			acp_fntype,acp_func);
		acp_process_exit(-1,acp_error_info);
	}

	return(q);
}

struct qelem *acp_q_pop3(i)
int i;
{
	struct qelem *q;
	q = q_pop_no_wait(i);
	return(q);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_q_push1

  DESCRIPTION:

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_q_push1(i,q)
int i;
struct qelem *q;
{
	if (q_push(i,q) == -1) {
		sprintf(acp_error_info,"%s%s: Failed to push message buffer",
			acp_fntype,acp_func);
		acp_process_exit(-1,acp_error_info);
	}
}

acp_q_push3(i,q)
int i;
struct qelem *q;
{
	if (q_push_urgent(i,q) == -1) {
		sprintf(acp_error_info,"%s%s: Failed to push message buffer",
			acp_fntype,acp_func);
		acp_process_exit(-1,acp_error_info);
	}
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_death_notice

  DESCRIPTION:

	  Notify the JM of a dead process

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_death_notice(lpn,msg,flag)
int lpn;
char *msg;
int flag;
{
	struct qelem *q;
	cps_msg *m;

	if (flag == 1) {
		q = acp_q_pop1(S_FREE_Q);
	}
	else {
		q = acp_q_pop2(S_FREE_Q);
	}
	m = SMALL_MSG(q);
	cps_bld_usr_proc_dead(m,acp_local_lpn,JM_LPN,lpn,msg);
	if (flag == 1) {
		acp_q_push1(acp_shm_qnum,q);
	}
	else {
		acp_q_push2(acp_shm_qnum,q);
	}
	acp_signal_shm();
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_send_to_user

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_send_to_user(q,lpn,flag)
struct qelem *q;
int lpn;
int flag;
{
	int i;

	i = lpn-FIRST_USER_LPN;

	if (acp_pmap[i].status != -1) {
		sprintf(acp_error_info,
			"Process %d is dead, cannot send message",lpn);
		acp_return_status = CPS_PARTNER_DEAD;
		return(CPS_PARTNER_DEAD);
	}

	if (acp_pmap[i].local) {
		if (flag == 1) {
			acp_q_push1(acp_pmap[i].semnum,q);
		}
		else {
			acp_q_push2(acp_pmap[i].semnum,q);
		}
		if (kill(acp_pmap[i].pid,SIGUSR1) == -1) {
			acp_death_notice(lpn,"Failed to signal local process",flag);
			sprintf(acp_error_info,
				"Process %d is dead, cannot send message",lpn);
			acp_return_status = CPS_PARTNER_DEAD;
			return(CPS_PARTNER_DEAD);
		}
	}
	else {
		if (flag == 1) {
			acp_q_push1(acp_shm_qnum,q);
		}
		else {
			acp_q_push2(acp_shm_qnum,q);
		}
		acp_signal_shm();
	}
	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_translate_sop

  DESCRIPTION:

	  Receives a set of classes (in CPS-style representation) and
	  returns a set of processes (standard representation).

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_translate_sop(sop,set)
unsigned int sop;
unsigned long *set;
{

	int *mem;
	unsigned int cset;

	int i,j;

	set_clear(set,MAX_PROCESSES);

	if ((sop & 0x80000000) == 0) {
		if (sop == CPS_THIS_PROCESS)
			add_to_set(acp_local_lpn,set);
		else
			add_to_set(sop,set);
		return;
	}

	sop &= 0x7FFFFFFE;
	cset = sop >> 1;

	mem = set_query(&cset,8*sizeof(unsigned int));
	i = 0;
	while (mem[i] != 0) {
		if (acp_class_proc[mem[i]] != NULL) {
			set_union(set,acp_class_proc[mem[i]],MAX_PROCESSES);
		}
		i++;
	}
/*	fflush(stdout); */
	return;
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_set_of_processes

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

unsigned int cps_set_of_processes (n, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20,	
			 c21,c22,c23,c24,c25,c26,c27,c28,c29,c30)	

int  *n, *c1, *c2, *c3, *c4, *c5, *c6, *c7, *c8, *c9, *c10,	
		 *c11,*c12,*c13,*c14,*c15,*c16,*c17,*c18,*c19,*c20,	
		 *c21,*c22,*c23,*c24,*c25,*c26,*c27,*c28,*c29,*c30;	
{
	unsigned int set = 0;

	char errmsg[80];

	static unsigned int class_bit_pattern[MAX_CLASSES+1] =
	 {		 0, 0x80000002, 0x80000004, 0x80000008,
	  0x80000010, 0x80000020, 0x80000040, 0x80000080,
	  0x80000100, 0x80000200, 0x80000400, 0x80000800,
	  0x80001000, 0x80002000, 0x80004000, 0x80008000,
	  0x80010000, 0x80020000, 0x80040000, 0x80080000,
	  0x80100000, 0x80200000, 0x80400000, 0x80800000,
	  0x81000000, 0x82000000, 0x84000000, 0x88000000,
	  0x90000000, 0xa0000000, 0xc0000000};

	acp_clear_error_info();
	acp_fntype = "cps_";
	acp_func = "set_of_processes";
/*
	CHECK_CPS_INIT
*/
	if ((*n<1) || (*n>MAX_CLASSES))  {		/* Number out of range */
		sprintf(errmsg,
			"cps_set_of_processes: class number %d out of range"
			,*n);
		acp_process_exit(-1,errmsg);
	}

	set |= class_bit_pattern[*c1];
	if (*n >= 2) set |= class_bit_pattern[*c2];
	else return(set);
	if (*n >= 3) set |= class_bit_pattern[*c3];
	else return(set);
	if (*n >= 4) set |= class_bit_pattern[*c4];
	else return(set);
	if (*n >= 5) set |= class_bit_pattern[*c5];
	else return(set);
	if (*n >= 6) set |= class_bit_pattern[*c6];
	else return(set);
	if (*n >= 7) set |= class_bit_pattern[*c7];
	else return(set);
	if (*n >= 8) set |= class_bit_pattern[*c8];
	else return(set);
	if (*n >= 9) set |= class_bit_pattern[*c9];
	else return(set);
	if (*n >= 10) set |= class_bit_pattern[*c10];
	else return(set);
	if (*n >= 11) set |= class_bit_pattern[*c11];
	else return(set);
	if (*n >= 12) set |= class_bit_pattern[*c12];
	else return(set);
	if (*n >= 13) set |= class_bit_pattern[*c13];
	else return(set);
	if (*n >= 14) set |= class_bit_pattern[*c14];
	else return(set);
	if (*n >= 15) set |= class_bit_pattern[*c15];
	else return(set);
	if (*n >= 16) set |= class_bit_pattern[*c16];
	else return(set);
	if (*n >= 17) set |= class_bit_pattern[*c17];
	else return(set);
	if (*n >= 18) set |= class_bit_pattern[*c18];
	else return(set);
	if (*n >= 19) set |= class_bit_pattern[*c19];
	else return(set);
	if (*n >= 20) set |= class_bit_pattern[*c20];
	else return(set);
	if (*n >= 21) set |= class_bit_pattern[*c21];
	else return(set);
	if (*n >= 22) set |= class_bit_pattern[*c22];
	else return(set);
	if (*n >= 23) set |= class_bit_pattern[*c23];
	else return(set);
	if (*n >= 24) set |= class_bit_pattern[*c24];
	else return(set);
	if (*n >= 25) set |= class_bit_pattern[*c25];
	else return(set);
	if (*n >= 26) set |= class_bit_pattern[*c26];
	else return(set);
	if (*n >= 27) set |= class_bit_pattern[*c27];
	else return(set);
	if (*n >= 28) set |= class_bit_pattern[*c28];
	else return(set);
	if (*n >= 29) set |= class_bit_pattern[*c29];
	else return(set);
	if (*n >= 30) set |= class_bit_pattern[*c30];

	return(set);
}

unsigned int cps_set_of_processes_(n, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20,	
			 c21,c22,c23,c24,c25,c26,c27,c28,c29,c30)	

int  *n, *c1, *c2, *c3, *c4, *c5, *c6, *c7, *c8, *c9, *c10,	
		 *c11,*c12,*c13,*c14,*c15,*c16,*c17,*c18,*c19,*c20,	
		 *c21,*c22,*c23,*c24,*c25,*c26,*c27,*c28,*c29,*c30;	
{
	return(cps_set_of_processes (n, c1, c2, c3, c4, c5, c6, c7, c8, c9, c10,	
			 c11,c12,c13,c14,c15,c16,c17,c18,c19,c20,	
			 c21,c22,c23,c24,c25,c26,c27,c28,c29,c30));
}


/******************************************************************/
/*
	FUNCTION NAME:

	  cps_bytes

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int cps_bytes(f,l)
unsigned int *f,*l;
{
	int nbytes;
	acp_clear_error_info();

	nbytes = ((int)l - (int)f) + sizeof(int);
	return(nbytes);

}

int cps_bytes_(f,l)
unsigned int *f,*l;
{
	return(cps_bytes(f,l));
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_checksum

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_checksum(f,l)
unsigned int *f,*l;
{
	unsigned int checksum = 0;

	acp_clear_error_info();

	while (f <= l) {
		checksum = checksum ^ *f;
		f++;
	}

	return(checksum);
}

int acp_checksum_(f,l)
unsigned int *f,*l;
{
	return(acp_checksum(f,l));
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_process_number

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int cps_process_number(lpn)
int *lpn;
{
	acp_clear_error_info();
	*lpn = acp_local_lpn;
	return(acp_local_lpn);
}

int cps_process_number_(lpn)
int *lpn;
{
	acp_clear_error_info();
	*lpn = acp_local_lpn;
	return(acp_local_lpn);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_class_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int cps_class_info(c,f,n)
int *c,*f,*n;
{
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CLASS_INFO )
	*c = acp_local_class;

/* Changed to used acp_static_cp in v2.9 */

	*f = first_in_set(acp_static_cp[acp_local_class],MAX_PROCESSES);
	*n = set_size(acp_static_cp[acp_local_class],MAX_PROCESSES);
	return(acp_local_class);
}
int cps_class_info_(c,f,n)
int *c,*f,*n;
{
	return(cps_class_info(c,f,n));
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_job_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int cps_job_info(n,clist,flist,nlist)
int *n,*clist,*flist,*nlist;
{
	int i,c,m;
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_JOB_INFO )

	m = *n;
	if (m > MAX_CLASSES) {
		for (i=MAX_CLASSES; i<m; i++) {
			flist[i] = 0;
			nlist[i] = 0;
		}
		m = MAX_CLASSES;
	}

	for (i=0; i<m; i++) {
		c = clist[i];
		if (acp_class_proc[c] != NULL) {
			flist[i] = first_in_set(acp_class_proc[c],MAX_PROCESSES);
			nlist[i] = set_size(acp_class_proc[c],MAX_PROCESSES);
		}
		else {
			flist[i] = 0;
			nlist[i] = 0;
		}
	}
	return(CPS_OK);
}

int cps_job_info_(n,clist,flist,nlist)
int *n,*clist,*flist,*nlist;
{
	cps_job_info(n,clist,flist,nlist);
	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_user_status_length

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int acp_user_status_length()
{
	return(4);
}
int acp_user_status_length_()
{
	return(4);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_update_user_status

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

cps_update_user_status(len,array)
int *len,*array;
{
	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_fntype = "cps_";
	acp_func = "update_user_status";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_UPDATE_USER_STATUS )

/* Check length */

	if ((*len < 0) || (*len > MAX_USER_STATUS)) {
		sprintf(acp_error_info,
			"%s%s: Number of user status values (%d) out of range",
			acp_fntype,acp_func,*len);
		acp_return_status = CPS_BAD_NUM_OF_ARGS;
		cps_handle_error1(CPS_DEFAULT_ACTION,"");
		acp_fntype = "cpssta_";
		return;
	}

/* Send a user-status message to the Job Manager */

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_user_status(m,acp_local_lpn,JM_LPN,*len,array);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

}

cps_update_user_status_(len,array)
int *len,*array;
{
	cps_update_user_status(len,array);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  acp_convert

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

acp_convert_(udata,cdata,bcount,type,src_class,dst_class)
char *udata,*cdata;
int *bcount,*type,*src_class,*dst_class;
{
	acp_convert(udata,cdata,bcount,type,src_class,dst_class);
}

acp_convert(udata,cdata,bcount,type,src_class,dst_class)
char *udata,*cdata;
int *bcount,*type,*src_class,*dst_class;
{
	acp_fntype = "cps_";
	acpsta_convert(udata,cdata,bcount,type,src_class,dst_class);
	cps_handle_error1(CPS_DEFAULT_ACTION,"");
	acp_fntype = "cpssta_";
}

int acpsta_convert_(udata,cdata,bcount,data_type,src_class,dst_class)
char *udata,*cdata;
int *bcount,*data_type,*src_class,*dst_class;
{
	return(acpsta_convert(udata,cdata,bcount,data_type,src_class,dst_class));
}

int acpsta_convert(udata,cdata,bcount,data_type,src_class,dst_class)
char *udata,*cdata;
int *bcount,*data_type,*src_class,*dst_class;
{
	int bc,type;

	acp_clear_error_info();
	acp_func = "convert";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_CONVERT )

	bc = *bcount;
	type = *data_type;

/* For now, all machines are big endian, ieee floating */

	bcopy(udata,cdata,bc);

	return(CPS_OK);
}

/******************************************************************/
/*
	FUNCTION NAME:

	  cps_dead_process_info

  DESCRIPTION:

	  [@tbs@]

  FORMAL PARAMETERS:

	  [@description or none@]

  FUNCTION VALUE:

	  [@description or none@]

								  */
/******************************************************************/

int
cps_dead_process_info(class,lsize,dlist,ndead)
int *class,*lsize,*dlist,*ndead;
{
	int *plist;

	int i,j,k;

	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_DEAD_PROCESS_INFO )

	if (*lsize <= 0) {
		*ndead = 0;
		return(CPS_OK);
	}

	plist = set_query(acp_class_proc[*class],MAX_PROCESSES);
	i = 0;
	j = 0;
	while (plist[i] != 0) {
		k = plist[i] - FIRST_USER_LPN;
		if (acp_pmap[k].status != -1) {
			*dlist = plist[i];
			dlist++;
			j++;
			if (j >= *lsize) {
				*ndead = j;
				return(CPS_OK);
			}
		}
		i++;
	}
	*ndead = j;
	return(CPS_OK);
}

int
cps_dead_process_info_(class,lsize,dlist,ndead)
int *class,*lsize,*dlist,*ndead;
{
	return( cps_dead_process_info(class,lsize,dlist,ndead) );
}

/************************************************************************/
/*

	Function: cps_sighold

	Description: Prevents SIGUSR1 from interrupting this process.

*/
/************************************************************************/

int cps_sighold_()
{
	return(cps_sighold());
}

int cps_sighold()
{
	sig_block_one(SIGUSR1);
	return(CPS_OK);
}

/************************************************************************/
/*

	Function: cps_sigrelease

	Description: Releases the hold on SIGUSR1.


*/
/************************************************************************/

int cps_sigrelease_()
{
	return(cps_sigrelease());
}

int cps_sigrelease()
{
	sig_pop_mask();
	return(CPS_OK);
}

/************************************************************************/
/*

	Function: cps_sleep

	Description: Calls UNIX sleep system call.


*/
/************************************************************************/

int cps_sleep_(x)
int *x;
{
	return(cps_sleep(x));
}

int cps_sleep(x)
int *x;
{
	int unslept;

	unslept = *x;
	while (unslept != 0) {
		unslept = sleep(unslept);
	}
	return(CPS_OK);
}

/************************************************************************/
/*

	Function:  acp_flush_stdout

	Description:

	Arguments:

	Function Value:

*/
/************************************************************************/

int acp_flush_stdout_()
{
	fflush(stdout);
}
int acp_flush_stdout()
{
	fflush(stdout);
}

cps_hook_( point )
int *point;
{
	cps_hook( point );
}

cps_hook( point )
int *point;
{
	struct qelem *q;
	cps_msg *m;

	acp_clear_error_info();
	acp_fntype = "cps_";
	acp_func = "hook";
	CHECK_CPS_INIT
	CHECK_CPS_HOOK( FUNC_HOOK )

	acp_hook_complete = 0;

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_user_hook(m,acp_local_lpn,JM_LPN,*point);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for a response from the JM */

	acp_wait_condition(&acp_hook_complete,1);
}

cps_proc_hook( point )
int point;
{
	struct qelem *q;
	cps_msg *m;

	switch( point )
	{
		case FUNC_HANDLE_ERROR:
			return;
			break;
		default:
			break;
	}

	acp_hook_complete = 0;

	q = acp_q_pop2(S_FREE_Q);
	m = SMALL_MSG(q);
	cps_bld_proc_hook(m,acp_local_lpn,JM_LPN,point);
	acp_q_push2(acp_shm_qnum,q);
	acp_signal_shm();

/* Wait for a response from the JM */

	acp_wait_condition(&acp_hook_complete,1);
}
