/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnsgia.fnal.gov

 Copyright (c) 1990 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/


/************************************************************************/
/*                                                                      */
/*                                                                      */
/*		    Routines to build CPS messages			*/
/*                                                                      */
/*                                                                      */
/************************************************************************/



/*****************/
/* Include Files */
/*****************/

#include <stdio.h>
#include "../inc/param.h"
#include "../inc/shm.h"
#include "../inc/cps_msgs.h"
#include "../inc/cps_user.h"
#include "../jm/jm_structs.h"

/************************/
/* External Definitions */
/************************/

#define SET_LPN(m,x,y) m->src_lpn = (long) x; \
		       m->dst_lpn = (long) y;

#define STRCPY(s1,s2) if (strlen(s2) < 256)	\
			  strcpy(s1,s2);	\
		      else {			\
			  bcopy(s2,s1,255);	\
			  s1[255] = '\0'; 	\
		      }

cps_bld_end_of_queue(m,s,d,q)
cps_message *m;
int s,d,q;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_END_OF_QUEUE;
	x = (long *) m->msg;
	*x = (long) q;
	m->size = (long) sizeof(long);
}

cps_bld_req_rsub_decl(m,s,d,id)
cps_message *m;
int s,d,id;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_REQ_SUB_DECL;
	x = (long *) m->msg;
	*x = (long) id;
	m->size = (long) sizeof(long);
}

cps_bld_req_q_info(m,s,d,q)
cps_message *m;
int s,d,q;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_REQ_Q_INFO;
	x = (long *) m->msg;
	*x = (long) q;
	m->size = (long) sizeof(long);
}

cps_bld_rem_call(m,s,d,id,inst,action,nargs,baddr,bcount,argtyp)
cps_message *m;
int s,d,id,inst,action,nargs,**baddr,*bcount,*argtyp;
{

	/* This message must be the same size as the MSG_QUEUE_PROCESS message */

	int i,sum;
	long *x;
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_REMOTE_CALL;
	x = (long *) m->msg;
	*x = (long) id;
	x++;
	*x = (long) inst;	    /* Instance number */
	x++;
	*x = (long) action;
	x++;
	c = (char *) x;
	sum = 0;
	for (i=0; i<nargs; i++) {
		if (argtyp[i] != CPS_WRITE_ONLY)
			bcopy(baddr[i],c,bcount[i]);
		else
			bzero(c,bcount[i]);
		c += bcount[i];
		sum += bcount[i];
	}
	m->size = (long) (3*sizeof(long) + sum);
}

cps_bld_sub_decl(m,s,d,rsub)
cps_message *m;
int s,d;
struct acp_remote_sub *rsub;
{
	struct acp_remote_sub *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_SUB_DECL;
	m->size = (long) sizeof(struct acp_remote_sub);
	x = (struct acp_remote_sub *) m->msg;
	*x = *rsub;

}

cps_bld_user_status(m,s,d,len,data)
cps_message *m;
int s,d,len,*data;
{
	long *x;
	int i;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_USER_STATUS;
	x = (long *) m->msg;
	*x = (long) len;
	x++;
	for (i=0; i<len; i++) {
		*x = (long) data[i];
		x++;
	}
	m->size = (long) (len + 1)*sizeof(long);
}

cps_bld_wait_sync(m,s,d,sync_num,nclass,c)
cps_message *m;
int s,d,sync_num,nclass,*c;
{
	int i;
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_WAIT_SYNC;
	x = (long *) m->msg;
	*x = (long) sync_num;
	x++;
	*x = (long) nclass;
	x++;
	for (i=0; i<nclass; i++) {
		*x = (long) c[i];
		x++;
	}
	m->size = (long) (2+nclass)*sizeof(long);
}

cps_bld_wait_queue(m,s,d,q,state)
cps_message *m;
int s,d,q,state;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_WAIT_QUEUE;
	x = (long *) m->msg;
	*x = (long) q;
	x++;
	*x = (long) state;
	m->size = (long) 2*sizeof(long);
}

cps_bld_dequeue_process(m,s,d,q,w)
cps_message *m;
int s,d,q,w;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DEQUEUE_PROCESS;
	x = (long *) m->msg;
	*x = (long) q;
	x++;
	*x = (long) w;
	m->size = (long) 2*sizeof(long);
}

cps_bld_queue_process(m,s,d,lpn,q,n,bcnt,baddr)
cps_message *m;
int s,d,lpn,q,n,*bcnt,**baddr;
{
	/* This message must be the same size as the MSG_REMOTE_CALL message */

	long *x;
	char *c;
	int i,sum;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_QUEUE_PROCESS;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) q;
	x++;
	x++;
	c = (char *) x;
	sum = 0;
	for (i=0; i<n; i++) {
		bcopy(baddr[i],c,bcnt[i]);
		c += bcnt[i];
		sum += bcnt[i];
	}
	m->size = (long) (3*sizeof(long) + sum);
}

cps_bld_dequeue_reply(m,s,d,lpn,q,nbytes,baddr)
cps_message *m;
int s,d,lpn,q,nbytes;
char *baddr;
{
	/* This message must be the same size as the MSG_REMOTE_CALL message */

	long *x;
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DEQUEUE_REPLY;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) q;
	x++;
	x++;
	c = (char *) x;
	if (nbytes > 0) bcopy(baddr,c,nbytes);
	m->size = (long) (3*sizeof(long) + nbytes);
}

cps_bld_declare_queue(m,s,d,q,n,b)
cps_message *m;
int s,d,q,n,*b;
{
	long *x;
	int i;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DECLARE_QUEUE;
	x = (long *) m->msg;
	*x = (long) q;
	x++;
	*x = (long) n;
	x++;
	for (i=0; i<n; i++) x[i] = (long) b[i];
	m->size = (long) (2 + n)*sizeof(long);
}

cps_bld_req_complete(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_REQ_COMPLETE;
	m->size = (long) 0;
}

cps_bld_sync_complete(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_SYNC_COMPLETE;
	m->size = (long) 0;
}

cps_bld_hook_complete(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_HOOK_COMPLETE;
	m->size = (long) 0;
}

cps_bld_wait_complete(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_WAIT_COMPLETE;
	m->size = (long) 0;
}

cps_bld_rsub_complete(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_RSUB_COMPLETE;
	m->size = (long) 0;
}

cps_bld_usr_proc_dead(m,s,d,lpn,logmsg)
cps_msg *m;
int s,d,lpn;
char *logmsg;
{
	long *l;
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_USR_PROC_DEAD;
	l = (long *) m->msg;
	*l = (long) lpn;
	l++;
	c = (char *) l;
	STRCPY(c,logmsg);
	m->size = (long) (sizeof(long) + strlen(logmsg)+1);
}

cps_bld_usr_finished(m,s,d,lpn,logmsg)
cps_msg *m;
int s,d,lpn;
char *logmsg;
{
	long *l;
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_USR_FINISHED;
	l = (long *) m->msg;
	*l = (long) lpn;
	l++;
	c = (char *) l;
	STRCPY(c,logmsg);
	m->size = (long) (sizeof(long) + strlen(logmsg)+1);
}

cps_bld_usr_abort_job(m,s,d,logmsg)
cps_msg *m;
int s,d;
char *logmsg;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_USR_ABORT_JOB;
	STRCPY(m->msg,logmsg);
	m->size = (long) strlen(logmsg)+1;
}

cps_bld_usr_log(m,s,d,logmsg)
cps_msg *m;
int s,d;
char *logmsg;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_USR_LOG;
	STRCPY(m->msg,logmsg);
	m->size = (long) strlen(logmsg)+1;
}

cps_bld_shm_error(m,s,d,errmsg)
cps_msg *m;
int s,d;
char *errmsg;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_SHM_ERROR;
	STRCPY(m->msg,errmsg);
	m->size = (long) strlen(errmsg)+1;
}

cps_bld_shm_shutdown(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_SHM_SHUTDOWN;
	m->size = (long) 0;
}

cps_bld_empty_msg(m,s,d)
cps_message *m;
int s,d;
{
	SET_LPN(m,s,d)
	    m->type = (long) MSG_EMPTY_MSG;
	m->size = (long) 0;
}

cps_bld_declare_block(m,s,d,lpn,blockno)
cps_message *m;
int s,d,lpn,blockno;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DECLARE_BLOCK;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) blockno;
	m->size = (long) 2*sizeof(long);
}

cps_bld_data_ready(m,s,d,lpn,blockno,bcount)
cps_message *m;
int s,d,lpn,blockno,bcount;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DATA_READY;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) blockno;
	x++;
	*x = (long) bcount;
	m->size = (long) 3*sizeof(long);
}

cps_bld_get_next_tape(m,s,d,listno)
cps_message *m;
int s,d,listno;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_GET_NEXT_TAPE;
	x = (long *) m->msg;
	*x = (long) listno;
	m->size = (long) sizeof(long);
}

cps_bld_list_mount_reply(m,s,d,rstatus,dev,vsn,pos,rstate,errmsg)
cps_message *m;
int s,d,rstatus,pos,rstate;
char *dev,*vsn,*errmsg;
{
	long *x;
	char *c;
	int len = 0;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_LIST_MOUNT_REPLY;

	x = (long *) m->msg;
	*x = (long) rstatus;
	x++;
	*x = (long) pos;
	x++;
	*x = (long) rstate;
	x++;
	c = (char *) x;
	strcpy(c,dev);
	c += strlen(dev)+1;
	len += strlen(dev)+1;
	strcpy(c,vsn);
	c += strlen(vsn)+1;
	len += strlen(vsn)+1;
	strcpy(c,errmsg);
	len += strlen(errmsg)+1;
	m->size = (long) (3*sizeof(long) + len);

}

cps_bld_spec_mount_reply(m,s,d,rstatus,dev,errmsg)
cps_message *m;
int s,d,rstatus;
char *dev,*errmsg;
{
	long *x;
	char *c;
	int len = 0;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_SPEC_MOUNT_REPLY;

	x = (long *) m->msg;
	*x = (long) rstatus;
	x++;
	c = (char *) x;
	strcpy(c,dev);
	c += strlen(dev)+1;
	len += strlen(dev)+1;
	strcpy(c,errmsg);
	len += strlen(errmsg)+1;
	m->size = (long) (sizeof(long) + len);

}

cps_bld_mount_request(m,s,d,dev,vsn,write_enabled,comment)
cps_message *m;
int s,d,write_enabled;
char *dev,*comment,*vsn;
{
	long *x;
	char *c;
	int len = 0;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_MOUNT_REQUEST;

	x = (long *) m->msg;
	*x = (long) write_enabled;
	x++;
	c = (char *) x;
	strcpy(c,dev);
	c += strlen(dev)+1;
	len += strlen(dev)+1;
	strcpy(c,vsn);
	c += strlen(vsn)+1;
	len += strlen(vsn)+1;
	strcpy(c,comment);
	len += strlen(comment)+1;
	m->size = (long) (sizeof(long) + len);
}

cps_bld_done_with_tape(m,s,d,dev)
cps_message *m;
int s,d;
char *dev;
{
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DONE_WITH_TAPE;
	c = (char *) m->msg;
	;
	strcpy(c,dev);
	m->size = (long) (strlen(dev)+1);
}

cps_bld_get_spec_tape(m,s,d,devtyp,ring,vsn)
cps_message *m;
int s,d,devtyp,ring;
char *vsn;
{
	long *x;
	char *c;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_GET_SPEC_TAPE;

	x = (long *) m->msg;
	*x = (long) devtyp;
	x++;
	*x = (long) ring;
	x++;
	c = (char *) x;
	strcpy(c,vsn);
	m->size = (long) (2*sizeof(long) + strlen(c)+1);
}

cps_bld_mon_update(m,s,d,opcode,lpn,datum1,datum2)
cps_message *m;
int s,d,opcode,lpn,datum1,datum2;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_MON_UPDATE;
	x = (long *) m->msg;
	*x = (long) opcode;
	x++;
	*x = (long) lpn;
	x++;
	*x = (long) datum1;
	x++;
	*x = (long) datum2;
	m->size = (long) 4*sizeof(long);
}

cps_bld_explicit_msg(m,s,d,mtype,addr,size)
cps_message *m;
int s,d,mtype,size;
char *addr;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_EXPLICIT_MSG;
	x = (long *) m->msg;
	*x = (long) mtype;
	x++;
	*x = (long) size;	    /* in bytes */
	x++;
	bcopy(addr,x,size);
	m->size = (long) (2*sizeof(long) + size);
}

cps_bld_fill_buffer(m,s,d,index,xfer_type,vaddr,blockno,offset,size)
cps_message *m;
int s,d,blockno,offset,size,index,xfer_type,vaddr;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_FILL_BUFFER;
	x = (long *) m->msg;
	*x = (long) index;
	x++;
	*x = (long) xfer_type;
	x++;
	*x = (long) vaddr;			/* Destination virtual address */
	x++;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) size;	    		/* in bytes */
	m->size = (long) 6*sizeof(long);
}

cps_bld_xfer_complete(m,s,d,xfer_type,blockno)
cps_message *m;
int s,d,xfer_type,blockno;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_XFER_COMPLETE;
	x = (long *) m->msg;
	*x = (long) xfer_type;
	x++;
	*x = (long) blockno;
	m->size = (long) (2*sizeof(long));
}

cps_bld_data_ack(m,s,d,index)
cps_message *m;
int s,d,index;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DATA_ACK;
	x = (long *) m->msg;
	*x = (long) index;
	m->size = (long) sizeof(long);
}

cps_bld_xfer_request(m,s,d,xfer_type,src_lpn,dst_lpn,vaddr,blockno,offset,size)
cps_message *m;
int s,d,blockno,offset,size,xfer_type,src_lpn,dst_lpn,vaddr;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_XFER_REQUEST;
	x = (long *) m->msg;
	*x = (long) xfer_type;
	x++;
	*x = (long) src_lpn;
	x++;
	*x = (long) dst_lpn;
	x++;
	*x = (long) vaddr;			/* Destination virtual address */
	x++;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) size;	    		/* in bytes */
	m->size = (long) 7*sizeof(long);
}

cps_bld_data_xfer(m,s,d,index,xfer_type,blockno,offset,size,buf)
cps_message *m;
int s,d,offset,size,index,xfer_type,blockno;
char *buf;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DATA_XFER;
	x = (long *) m->msg;
	*x = (long) index;
	x++;
	*x = (long) xfer_type;
	x++;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) size;	    		/* in bytes */
	x++;
	bcopy(buf,x,size);
	m->size = (long) 5*sizeof(long) + size;
}

cps_bld_explicit_req(m,s,d,som)
cps_message *m;
int s,d,som;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_EXPLICIT_REQ;
	x = (long *) m->msg;
	*x = (long) som;
	m->size = (long) sizeof(long);
}

cps_bld_explicit_rdy(m,s,d,som)
cps_message *m;
int s,d,som;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_EXPLICIT_RDY;
	x = (long *) m->msg;
	*x = (long) som;
	m->size = (long) sizeof(long);
}

cps_bld_vbbc_info(m,s,d,lpn,blockno,crate,vme_addr,offset,n,paddr)
cps_message *m;
int s,d,lpn,blockno,n,crate,vme_addr,offset;
unsigned long *paddr;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_VBBC_INFO;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) blockno;
	x++;
	*x = (long) crate;
	x++;
	*x = (long) vme_addr;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) n;
	x++;
	bcopy(paddr,x,n*sizeof(long));
	m->size = (long) (6+n)*sizeof(long);
}

cps_bld_asynch_dq(m,s,d,q,id)
cps_message *m;
int s,d,q,id;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_ASYNCH_DQ;
	x = (long *) m->msg;
	*x = (long) q;
	x++;
	*x = (long) id;
	m->size = (long) (2*sizeof(long));
}

cps_bld_asynch_dq_reply(m,s,d,lpn,q,id)
cps_message *m;
int s,d,lpn,q,id;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_ASYNCH_DQ_REPLY;
	x = (long *) m->msg;
	*x = (long) lpn;
	x++;
	*x = (long) q;
	x++;
	*x = (long) id;
	m->size = (long) (3*sizeof(long));
}

cps_bld_block_arrived(m,s,d,blockno)
cps_message *m;
int s,d,blockno;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_BLOCK_ARRIVED;
	x = (long *) m->msg;
	*x = (long) blockno;
	m->size = (long) sizeof(long);
}

cps_bld_direct_send(m,s,d,blockno,offset,nbytes)
cps_message *m;
int s,d,blockno,offset,nbytes;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DIRECT_SEND;
	x = (long *) m->msg;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) nbytes;
	m->size = (long) (3*sizeof(long));
}

cps_bld_direct_get(m,s,d,blockno,offset,nbytes)
cps_message *m;
int s,d,blockno,offset,nbytes;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DIRECT_GET;
	x = (long *) m->msg;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) nbytes;
	m->size = (long) (3*sizeof(long));
}

cps_bld_direct_xfer(m,s,d,lpn)
cps_message *m;
int s,d,lpn;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_DIRECT_XFER;
	x = (long *) m->msg;
	*x = (long) lpn;
	m->size = (long) sizeof(long);
}

cps_bld_small_msg(m,s,d,blockno,offset,size,buf)
cps_message *m;
int s,d,offset,size,blockno;
char *buf;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_SMALL_MSG;
	x = (long *) m->msg;
	*x = (long) blockno;
	x++;
	*x = (long) offset;
	x++;
	*x = (long) size;	    		/* in bytes */
	x++;
	bcopy(buf,x,size);
	m->size = (long) 3*sizeof(long) + size;
}

cps_bld_user_hook(m,s,d,pt)
cps_message *m;
int s,d,pt;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_USER_HOOK;
	x = (long *) m->msg;
	*x = (long) pt;
	m->size = (long) sizeof(long);
}

cps_bld_proc_hook(m,s,d,pt)
cps_message *m;
int s,d,pt;
{
	long *x;

	SET_LPN(m,s,d)
	    m->type = (long) MSG_FUNC_HOOK;
	x = (long *) m->msg;
	*x = (long) pt;
	m->size = (long) sizeof(long);
}

cps_bld_hook_switch(m)
cps_message *m;
{
	m->type = (long) MSG_HOOK_SWITCH;
	m->size = (long) 0;
}
