/*

 Author:	Matt Fausey
		Fermi National Accelerator Laboratory - Mail Station 369
		P.O Box 500
		Batavia, IL 60510
		DECnet: FNAL::FAUSEY
		Internet: fausey@fnfaus.fnal.gov

 Copyright (c) 1991 Universities Research Association, Inc.        
       All Rights Reserved                                                    


 This material resulted from work developed under a Government Contract and   
 is subject to the following license:  The Government retains a paid-up,      
 nonexclusive, irrevocable worldwide license to reproduce, prepare derivative 
 works, perform publicly and display publicly by or for the Government,       
 including the right to distribute to other Government contractors.  Neither  
 the United States nor the United States Department of Energy, nor any of     
 their employees, makes any warrenty, express or implied, or assumes any      
 legal liability or responsibility for the accuracy, completeness, or         
 usefulness of any information, apparatus, product, or process disclosed, or  
 represents that its use would not infringe privately owned rights.

*/

/*

    Source File: mreq.c

    Explanation: Mount request utility routines.

    Exported Routines:  remove_from_mreq_list(mrq,mrq_tail,m)
		       	add_to_mreq_list(mrq,mrq_tail,m)
		        find_mount_request(m,id)
			find_m_request_by_dev(m)
		        build_mount_request(m)
		        decode_mount_request()
		        build_mount_reply(m)
		        decode_mount_reply(m)
			build_mount_waiting(m)
			decode_mount_waiting()
*/

#include <stdio.h>
#include "../inc/pm.h"

#define FACILITY "MREQ"

extern char **msg_args;

/************************************************************************/
/*

    Function:  find_mount_request

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

struct mountreq *find_mount_request(m,id)
struct mountreq *m;
int id;
{
	while (m != NULL) {
		if (m->id == id) return m;
		m = m->next;
	}
	return m;
}

/************************************************************************/
/*

    Function:  find_m_request_by_dev

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

struct mountreq *find_m_request_by_dev(mrql,m)
struct mountreq *mrql,*m;
{
	while (mrql != NULL) {
		if (cpu_name_compare(mrql->cpu,m->cpu) == 0) {
			if (strcmp(mrql->dev,m->dev) == 0) {
				return mrql;
			}
		}
		mrql = mrql->next;
	}
	return NULL;
}

/************************************************************************/
/*

    Function:  add_to_mreq_list

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

int add_to_mreq_list(mrq,mrq_tail,m)
struct mountreq **mrq,**mrq_tail,*m;
{
	m->next = *mrq;
	if (*mrq == NULL) {
		*mrq_tail = m;
	}
	else {
		m->next->prev = m;
	}
	*mrq = m;
	m->prev = NULL;

	return 0;
}

/************************************************************************/
/*

    Function:  remove_from_mreq_list

*/
/************************************************************************/

int remove_from_mreq_list(mrq,mrq_tail,m)
struct mountreq **mrq,**mrq_tail,*m;
{
	if (m->prev == NULL) {
		*mrq = m->next;
	}
	else {
		m->prev->next = m->next;
	}

	if (m->next == NULL) {
		*mrq_tail = m->prev;
	}
	else {
		m->next->prev = m->prev;
	}
	return 0;
}

/************************************************************************/
/*

    Function:  build_mount_request

    Description:

    Arguments:

    Function Value:

*/
/************************************************************************/

char *build_mount_request(m)
struct mountreq *m;
{
	static char mreq[256];

	sprintf(mreq,"m_request|%d|%d|%d|%d|%d|%d|%d|%d|%s|%d|%d|%s|%s|%s",
	    m->id,
	    m->type,
	    m->class,m->listnum,
	    m->aux[0],m->aux[1],m->aux[2],m->aux[3],
	    m->vsn,m->rstate,m->devtype,m->dev,m->cpu,m->user);

	return mreq;
}

struct mountreq *decode_mount_request()
{
	struct mountreq *m;

	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->id = atoi(msg_args[0]);
	m->type = atoi(msg_args[1]);
	m->class = atoi(msg_args[2]);
	m->listnum = atoi(msg_args[3]);
	m->aux[0] = atoi(msg_args[4]);
	m->aux[1] = atoi(msg_args[5]);
	m->aux[2] = atoi(msg_args[6]);
	m->aux[3] = atoi(msg_args[7]);
	strncpy(m->vsn,msg_args[8],15);
	m->rstate = atoi(msg_args[9]);
	m->devtype = atoi(msg_args[10]);
	strcpy(m->dev,msg_args[11]);
	strcpy(m->cpu,msg_args[12]);
	if (msg_args[13] != NULL) strcpy(m->user,msg_args[13]);
	return m;
}

char *build_mount_reply(m)
struct mountreq *m;
{
	static char mrep[256];

	sprintf(mrep,"m_reply|%d|%d|%d|%d|%d|%d|%d|%d|%s|%d|%s|%s|%d|%d|%s",
	    m->id,
	    m->type,
	    m->class,m->listnum,
	    m->aux[0],m->aux[1],m->aux[2],m->aux[3],
	    m->vsn,m->rstate,m->dev,m->cpu,
	    m->listpos,m->rstatus,m->errmsg);

	return mrep;
}

int decode_mount_reply(m)
struct mountreq *m;
{
	bzero(m,sizeof(struct mountreq));
	m->id = atoi(msg_args[0]);
	m->type = atoi(msg_args[1]);
	m->class = atoi(msg_args[2]);
	m->listnum = atoi(msg_args[3]);
	m->aux[0] = atoi(msg_args[4]);
	m->aux[1] = atoi(msg_args[5]);
	m->aux[2] = atoi(msg_args[6]);
	m->aux[3] = atoi(msg_args[7]);
	strncpy(m->vsn,msg_args[8],15);
	m->rstate = atoi(msg_args[9]);
	strcpy(m->dev,msg_args[10]);
	strcpy(m->cpu,msg_args[11]);
	m->listpos = atoi(msg_args[12]);
	m->rstatus = atoi(msg_args[13]);
	strcpy(m->errmsg,msg_args[14]);
	return 0;
}

char *build_mount_waiting(m)
struct mountreq *m;
{
	static char mreq[40];
	sprintf(mreq,"m_waiting|%d",m->id);
	return mreq;
}

struct mountreq *decode_mount_waiting()
{
	struct mountreq *m;
	m = (struct mountreq *) hmalloc(sizeof(struct mountreq));
	bzero(m,sizeof(struct mountreq));
	m->id = atoi(msg_args[0]);
	return m;
}
