/*
**
** Test of LieOperator inversion algorithm.
** 
** --- Leo Michelotti
** --- September 15, 1993
**
*/

#include <stdlib.h>
#include <stdio.h>
#include "mxyzptlk.rsc"

main( int argc, char** argv ) {
 
 if( argc != 2 ) {
  printf( "\n" );
  printf( "Usage: %s  limit                          \n", argv[0] );
  printf( "\n" );
  exit(0);
 }

 int limit = atoi( argv[1] );

 Jet::setup( 3, 4, 3 );

 coord        x(0.0), y(0.0), z(0.0);
 LieOperator  u, v, w;
 LieOperator  id( "ident" );


 w.setVariable( x + 3.0*y + x*y - y*z,   0 );
 w.setVariable( y - x + z + x*z + y*y,   1 );
 w.setVariable( z + 2.0*x + y*z - x*y*z, 2 );

 MatrixD M( 3, 3 );
 M = w.Jacobian();
 M = M.inverse();

 int i = 0;
 u = M*id;
 v = ( w(u) - id );
 while ( ( i++ < limit ) && ( v != 0.0 ) ) {
  u = u - M*v;
  v = ( w(u) - id );
 }

 printf( "\n====== u.printCoeffs(); =================\n" );
 u.printCoeffs();

 printf( "\n====== w.printCoeffs(); =================\n" );
 w.printCoeffs();

 printf( "\n====== w(u).printCoeffs(); =================\n" );
 w(u).printCoeffs();

 printf( "\n====== u(w).printCoeffs(); =================\n" );
 u(w).printCoeffs();

 if( i >= limit ) printf( "++++ Limit exceeded! ++++\n" );
}
