#ifndef JLC_HXX
#define JLC_HXX


// *******************************************************************
struct JLCtermData;
struct JLCterm {    

  int* index;	  // An integer array giving the derivatives associated
                  //      with the JLCterm.  For example, ( 1, 1, 0, 2 )
                  //            would correspond to D_1^1 D_2^1 D_4^2 .
  int weight;     // The sum of the values in index.  For the above example,
                  //                                        this would be 4.
  complex value;  // The value associated with the JLCterm.

  // Constructors and destructors
  JLCterm();   
  JLCterm(JLterm*); 
  JLCterm( int*, complex );
  JLCterm( JLCterm* );
  JLCterm( JLCterm& );
  JLCterm( JLCtermData& );
  ~JLCterm();

  // Operators
  void operator=( const JLCterm& );
  complex coeff(){ return value; }
} ;


// *******************************************************************
struct JLC : public dlist {  

  int count;        // The number of JLC terms in the variable
  int weight;       // The maximum weight of the terms
  int accuWgt;      // Highest weight computed accurately
  complex* refPoint; // Point about which derivatives are to be expanded.
  int rc;           // Reference counting for garbage collection

  void insert( JLCterm* );
  void append( JLCterm* );
  void peekAt( int );
  void printCoeffs( int );
  JLCterm* remove( dlink* );

  // Constructors and destructors_____________________________________
  JLC();
  JLC( const complex );
  JLC( const JLC& );
  JLC( const JL& );
  JLC( JLC* );
  ~JLC();

  JLCterm* get();            // Pops the top term, which should be the 
                             // one of lowest weight.
  void addTerm( JLCterm* );
  void fixReference( const complex* );
  void fixReference();
  void fixReference( JLC& );
  void fixReferenceAtEnd( const CLieOperator& );
  void fixReferenceAtStart( const CLieOperator& );

  char isNilpotent();
  void peekAt();
  void printCoeffs();
  dlist image();
  void writeToFile( char*   /* Name of unopened file */ );
  void writeToFile( FILE* );

  void getReference( complex* );
  void scaleBy( complex );
  void setVariable( complex, int );
  void setVariable( int );
  complex standardPart();
  void clear();
  complex weightedDerivative( int* );
  complex derivative( int* );
  complex operator()( complex* );	   // Performs a multinomial evaluation of 
				   // the JLC variable.  Essentially acts as a 
				   // power series expansion.

  JLC& operator=( const JLC& );
  JLC& operator=( const complex& );
  JLC& operator+=( const complex& );
} ;

struct JLCtermData{
  int  numvar;
  int* index;
  int weight;
  complex value;
  JLCtermData();
  JLCtermData( JLCterm& );
  ~JLCtermData();
};

void operator+=( JLC&, const JLC& );   // ??? Why are these void operators?
void operator-=( JLC&, const JLC& );
void operator*=( JLC&, const JLC& );
void operator/=( JLC&, const JLC& );
char operator!=( const JLC&, const JLC& );
char operator!=( const JLC&, const complex& );
char operator!=( const complex&, const JLC& );


// *******************************************************************
//
//     Setup routines, etc.
//

extern void JLCFixReference( complex* );
extern void JLCFixReference( JLC& );
extern void JLCFixReferenceAtStart( const CLieOperator& );
extern void JLCFixReferenceAtEnd( CLieOperator& );

extern ostream& operator<<(ostream&, JLC&);
extern istream& operator>>(istream&, JLC&);
extern char operator==( const JLCterm&, const JLCterm& );
extern char operator==( const JLC&,     const JLC& );
extern char operator==( const JLC&,     const complex& );
extern char operator==( const complex&, const JLC& );
extern char operator<=( JLCterm&, JLCterm& );
extern char operator%=( JLCterm&, JLCterm& );   // Acts like == but compares
                                                // indices only.
extern JLCterm operator*( JLCterm&, JLCterm& );

#endif // JLC_HXX
