#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by JLC::addTerm to decide 
                            //   removal of a JLCterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLCFixReference( complex* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) JetC::oneEnv.refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLCFixReferenceAtEnd( CLieOperator& x ) {
 int i;
 complex* u = new complex [ JetC::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) JetC::oneEnv.refPoint[i] = u[i];
 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLCFixReferenceAtStart( const CLieOperator& x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) JetC::oneEnv.refPoint[i] = x.refPoint[i];
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class JLC

//    Constructors and destructors    |||||||||||||||||||||||||||

JLC::JLC() {
 int i;
 
 rc = 1;
 weight = -1;
 count = 0;
 accuWgt = JetC::oneEnv.MaxWeight;
 
 if( JetC::oneEnv.allSet == 137 ) {
   refPoint = new complex[ JetC::oneEnv.NumVar ];
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];
 }
 else {
   refPoint = NULL;
 }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC::JLC( const complex x ) {
 int* ndx;
 int i;
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( complex ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 clear();
 rc = 1;
 weight = 0;
 count = 0;
 accuWgt = JetC::oneEnv.MaxWeight;
 
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];
 
                             // ??? There is some danger here!  This constructor
                             //     governs the conversion of complex into JLC.
                             //     In an expression like 3.0 * z, where z is a JLC,
                             //     the multiplication may not go through if the
                             //     current reference point has changed since z
                             //     was created.  One way around this is to overload
                             //     all arithmetic operators explicitly so conversion
                             //     is never necessary.
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLCterm( ndx, x ) );
 delete [] ndx;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC::JLC( const JLC& x ) {
 dlist_iterator getNext( (dlist&) x );
 JLCterm* p;
 JLCterm* q;
 int i;
 
 static char firstCall = 1;
 if( firstCall ) {
  firstCall = 0;
  cout << "\n*** WARNING *** The JLC copy constructor was called for \n"
       <<   "*** WARNING *** some inexplicable reason.\n"
       <<   "*** WARNING *** Go figure!!\n\n";
  cout.flush();
 }

 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JLC& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JLC& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 while( p = (JLCterm*) getNext() ) {
   q = new JLCterm( p );
   append( q );
   }

 rc       = 1;
 count    = x.count;
 weight   = x.weight;
 accuWgt  = x.accuWgt;
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC::JLC( const JL& x ) {
 dlist_iterator getNext( (dlist&) x );
 JLterm* p;
 JLCterm* q;
 int i;
 

 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JL& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JL& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 while( p = (JLterm*) getNext() ) {
   q = new JLCterm( p );
   append( q );
   }

 rc       = 1;
 count    = x.count;
 weight   = x.weight;
 accuWgt  = x.accuWgt;
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = complex(x.refPoint[i],0.0);
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC::JLC( JLC* x ) {
 dlist_iterator getNext( *(dlist*) x );
 JLCterm* p;
 JLCterm* q;
 int i;
 
 static char firstCall = 1;
 if( firstCall ) {
  firstCall = 0;
  cout << "\n*** WARNING *** The JLC pointer constructor was called for \n"
       <<   "*** WARNING *** some inexplicable reason.\n"
       <<   "*** WARNING *** Go figure!!\n\n";
  cout.flush();
 }

 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JLC& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC::JLC( JLC& ) ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 while( p = (JLCterm*) getNext() ) {
   q = new JLCterm( p );
   append( q );
   }

 rc       = 1;
 count    = x->count;
 weight   = x->weight;
 accuWgt  = x->accuWgt;
 refPoint = new complex[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x->refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC::~JLC() {
 clear();
 delete [] refPoint;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

//     Member functions(public)  |||||||||||||||||||||||||||||||

void JLC::addTerm( JLCterm* a) {
 dlist_traversor getNext( *(dlist*) this );
 dlink* p;
 
 // If the value of *a is 0, don't bother with it unless
 // the weight is also 0.
 if( ( a -> value == complex_0 ) && ( a -> weight != 0 ) ) {
   delete a;
   return;
 }
 
 //
 // In the event that the candidate is lighter than the first
 // element in the JLC dlist ... or the list is empty
 //
 p = getNext();
 
 if( p == 0 ) {
   insert( a );
   return;
   }
 
 if(     *a   <=   ( *(JLCterm*) (p->info()) )   ) {
   if(   *a   %=   ( *(JLCterm*) (p->info()) )   ) {
 
     (( (JLCterm*) (p->info()) ) -> value ) += ( a -> value );
 
     if(
         (
           abs( ( (JLCterm*) (p->info()) ) -> value ) < MX_SMALL*abs( a -> value )
         )
         // REMOVE ??? &&
         // REMOVE ??? (
         // REMOVE ???   p != ( dlist::lastPtr() ) -> nextPtr()
         // REMOVE ??? )
       )
 
         delete remove( p );
 
     delete a;
     return;
     }
 
   else {
     insert( a );
     return;
     }
 
   }
 
 //
 // Otherwise, search through the list
 //
 while( p = getNext() ){
 
   if(     *a   <=   ( *(JLCterm*) (p->info()) )   ) {
     if(   *a   %=   ( *(JLCterm*) (p->info()) )   ) {
 
       (( (JLCterm*) (p->info()) ) -> value ) += ( a -> value );
 
       if(
	   (
	     abs( ( (JLCterm*) (p->info()) ) -> value ) < MX_SMALL*abs( a -> value )
	   )
	   // REMOVE ??? &&
	   // REMOVE ??? (
	   // REMOVE ???   p != ( dlist::lastPtr() ) -> nextPtr()
	   // REMOVE ??? )
	 )
 
         delete remove( p );
 
       delete a;
       return;
       }
 
     else {
       p -> putAbove( new dlink( a, 0, 0 ) );
       (this->count)++;
       if( (this->weight) < (a->weight) ) this->weight = a->weight;
       return;
       }
 
     }
   }
 
 //
 //  The new entry is heavier than existing ones.
 //
 append( a );
 return;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::fixReference( const complex* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::fixReference() {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::fixReference( JLC& x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::fixReferenceAtEnd( const CLieOperator& x ) {
 int i;
 complex* u = new complex [ JetC::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = u[i];
 delete [] u;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::fixReferenceAtStart( const CLieOperator& x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::getReference( complex* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) r[i] = refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char JLC::isNilpotent() {
 dlist_iterator getNext( *(dlist*) this );
 JLCterm* p = (JLCterm*) getNext();
 if( p->weight == 0 && p->value != complex_0 ) return 0;
 else                                    return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::peekAt() {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 cout << "\n Count = " << count << " Weight = " << weight;
 cout << " Max accurate weight = " << accuWgt << endl; 
 cout <<"Reference point:  " << endl;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   cout << " " << refPoint[i];
 cout << endl;
 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   cout << "Weight: " << p->weight << " Value: " << p->value <<" || ";
   cout << " Addresses: " << q->prevPtr() << " " << q << " ";
   cout << q->nextPtr() << " : " << p << endl;
   cout <<"Index:  ";
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     cout  << " " << (p -> index)[i];
   cout << endl;

   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::printCoeffs() {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 cout << "\n Count = " << count << " Weight = " << weight;
 cout << " Max accurate weight = " << accuWgt << endl; 
 cout <<"Reference point:  " << endl;

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   cout << " " << refPoint[i];
 cout << endl;

 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   cout <<"Index:  ";
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     cout << ::form( "%3d", (p -> index)[i] );
   cout << "    Value :  " << p->value << endl;
   }
 cout << "\n" << endl;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::peekAt( int  ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 fprintf( JetC::scratchFile,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( JetC::scratchFile, "Reference point:  \n" );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   fprintf( JetC::scratchFile, "%e  ", refPoint[i] );
 fprintf( JetC::scratchFile, "\n" );
 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   fprintf( JetC::scratchFile, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( JetC::scratchFile, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( JetC::scratchFile, "Index:  ");
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     fprintf( JetC::scratchFile, "%d  ", (p -> index)[i] );
   fprintf( JetC::scratchFile, "\n\n");
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::printCoeffs( int  ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 fprintf( JetC::scratchFile, "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( JetC::scratchFile, "Reference point:  " );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) fprintf( JetC::scratchFile, "%e  ", refPoint[i] );
 fprintf( JetC::scratchFile, "\n\n" );
 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   fprintf( JetC::scratchFile,"Index:  ");
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     fprintf( JetC::scratchFile, "%3d ", (p -> index)[i] );
   fprintf( JetC::scratchFile, "   Value: %14.6e  ", p -> value );
   fprintf( JetC::scratchFile,"\n");
   }
 fprintf( JetC::scratchFile, "\n\n" );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist JLC::image() {       // ??? Problem: nothing is recorded about the
                          // ??? reference point.

 dlist_iterator getNext( *(dlist*) this );
 JLCterm* q;
 dlist ret;

 // ??? for( i = 0; i < JetC::oneEnv.NumVar; i++ ) fprintf( JetC::scratchFile, "%e  ", refPoint[i] );

 while( q = (JLCterm*) getNext() )
   ret.append( new JLCtermData( *q ) );

 return ret;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::writeToFile( char* fileName ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 
 JetC::scratchFile = fopen( fileName, "w" );
 
 fprintf( JetC::scratchFile,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( JetC::scratchFile, "Reference point:  \n" );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   fprintf( JetC::scratchFile, "%e  ", refPoint[i] );
 fprintf( JetC::scratchFile, "\n" );
 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   fprintf( JetC::scratchFile, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( JetC::scratchFile, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( JetC::scratchFile, "Index:  ");
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     fprintf( JetC::scratchFile, "%d  ", (p -> index)[i] );
   fprintf( JetC::scratchFile, "\n\n");
   }
 
 fclose( JetC::scratchFile );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::writeToFile( FILE* filePtr ) {
 dlist_traversor getNext( *(dlist*) this );
 int i;
 JLCterm* p;
 dlink* q;
 
 fprintf( filePtr,
          "\nCount  = %d, Weight = %d, Max accurate weight = %d\n",
          count, weight, accuWgt );
 fprintf( filePtr, "Reference point:  \n" );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   fprintf( filePtr, "%e  ", refPoint[i] );
 fprintf( filePtr, "\n" );
 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   fprintf( filePtr, "Weight: %d   Value: %e  || ",
           p -> weight, p -> value );
   fprintf( filePtr, "Addresses: %d %d %d : %d\n",
                  q->prevPtr(), q, q->nextPtr(), p );
   fprintf( filePtr, "Index:  ");
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     fprintf( filePtr, "%d  ", (p -> index)[i] );
   fprintf( filePtr, "\n\n");
   }
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::scaleBy( complex y ) {
 JLCterm* p;
 dlist_iterator getNext( *(dlist*) this );
 
 if ( this->count < 1 ) {
   printf( "\n\n*** WARNING::JLC::scaleBy: function invoked " );
   printf(          "by null JLC variable with address %d.\n",   this );
   return;
   }
 
 while( p = (JLCterm*) getNext() )   p->value *= y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::setVariable( complex x, int j ) {
 int* ndx;
 int i;
 
 clear();
 if( ( j < 0 ) || ( j >= JetC::oneEnv.NumVar ) ) return;
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 insert( new JLCterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLCterm( ndx, x ) );
 delete [] ndx;
 
 JetC::oneEnv.refPoint[j] = x;
 refPoint[j] = x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::setVariable( int j ) {
 int* ndx;
 int i;
 complex x;
 
 clear();
 if( ( j < 0 ) || ( j >= JetC::oneEnv.NumVar ) ) return;
 
 x = JetC::oneEnv.refPoint[j];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 ndx[j] = 1;
 insert( new JLCterm( ndx, 1.0 ) );
 delete [] ndx;
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLCterm( ndx, x ) );
 delete [] ndx;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::insert( JLCterm* a) {
 dlist::insert(a);
 count++;
 if( weight < (a->weight) ) weight = a->weight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm* JLC::get() {
 JLCterm* p;
 if( count <= 0 ) return 0;
 p = ((JLCterm*) dlist::get());
 if( --count == 0 ) weight = -1;
 return p;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::append( JLCterm* a) {
 dlist::append(a);
 count++;
 if( weight < (a->weight) ) weight = a->weight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm* JLC::remove( dlink* w ) {
 JLCterm* a;
 dlink* predecessor;
 dlink* successor;
 int predWeight;
 int succWeight;
 
 predecessor = w->prevPtr();
 successor   = w->nextPtr();
 
 a = (JLCterm*) dlist::remove( w );
 count--;
 
 if( count == 0 ) {
   weight = -1;
   return a;
   }
 
 if ( weight > a->weight ) return a;
 
 predWeight = ( (JLCterm*) ( predecessor->info() ) ) -> weight;
 succWeight = ( (JLCterm*) ( successor->info() ) ) -> weight;
 weight = predWeight;
 if( succWeight > weight ) weight = succWeight;
 return a;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JLC::standardPart() {
 dlist_iterator g( (dlist&) *this );
 JLCterm* p; 
 if( count < 1 )       return 0.0;
 p = (JLCterm*) g();
 if( p->weight  == 0 ) return p->value;
                       return 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JLC::clear() {
 JLCterm* p;
 while( p = get() )  
   delete p;
 dlist::clear();
 count   = 0;
 weight  = -1;
 accuWgt = JetC::oneEnv.MaxWeight;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JLC::weightedDerivative( int* ind ) {
 int i;
 JLCterm* p;
 char theOne;
 
 if( count < 1 ) return 0.0;
 
 dlist_iterator getNext( (dlist&) *this );

 while( p = (JLCterm*) getNext() ) {
   theOne = 1;
   for ( i = 0; i < JetC::oneEnv.NumVar; i++ )
     theOne = theOne && ( (p->index)[i] == ind[i] );
   if( theOne ) return p->value;
 }
 
 return 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JLC::derivative( int* ind ) {
 double n;
 complex multiplier;
 complex d;
 int i;
 
 d = weightedDerivative( ind );
 if ( d == complex_0 ) return complex_0;
 
 multiplier = 1.0;
 for ( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   n = ind[i];
   while(n)  multiplier *= n--;
   }
 
 return multiplier*d;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JLC::operator()( complex* x ) {
 JLCterm* p;
 complex v;
 int w;
 int i;
 complex term;
 complex* u;
 dlist_iterator getNext( *(dlist*) this );
 
 u = new complex[ JetC::oneEnv.NumVar ];   // ??? Check other places in the source
                                 // code where this was not done correctly!!
 
 // Subtract off the reference point.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) u[i] = x[i] - refPoint[i];
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 JetC::oneEnv.monomial[0] = 1.0;
 
 // For all higher weights ...
 for( w = 1; w <= weight; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, JetC::oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = JetC::oneEnv.monomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     JetC::oneEnv.monomial[ Jet::monoRank() ] = term * u[i];
 
     } // End while loop.
 
 // Monomials have been now stored at this point.
 // Now traverse the JL variable and evaluate.
 
 v = 0.0;
 
 while( p = (JLCterm*) getNext() ) {
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) Jet::oneEnv.exponent[i] = (p->index)[i];
   Jet::monoCode();
   v += ( p->value ) * ( JetC::oneEnv.monomial[ Jet::monoRank() ] );
   }
 
 delete [] u;
 
 return v;
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||



//***************************************************************
//***************************************************************
//***************************************************************
//
//      The overloaded operators for class JLC
//
istream& operator>>(istream& is,  JLC& x) {
  char buf[100];
  int i,j;
  int count;
  complex value;
  JLCterm* q;
  int* index = new int[JetC::oneEnv.NumVar];

  x.clear();

  is >> buf;
  is >> buf;
  is >> count;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.weight;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> buf;
  is >> x.accuWgt;
  is >> buf;
  is >> buf;
  for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   is >> x.refPoint[i];
  for( j = 0; j < count; j++) {
    is >> buf;
    for( i = 0; i < JetC::oneEnv.NumVar; i++ )  {
     is >> index[i];
    }
    is >> buf;
    is >> value;
    q = new JLCterm(index,value);
    x.append( q );
  }

 delete [] index;
 return is;
}

ostream& operator<<(ostream& os, JLC& x) {
 dlist_traversor getNext( (dlist&) x );
 int i;
 JLCterm* p;
 dlink* q;

 os << "\n Count = " << x.count << " , Weight = " << x.weight;
 os << " , Max accurate weight = " << x.accuWgt << endl;
 os << "Reference point: ";
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   os << setprecision(10) << x.refPoint[i] << " ";
 os << "\n\n";

 while( q = getNext() ) {
   p = (JLCterm*) q->info();
   os << "Index: ";
   for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     os << p -> index[i] << " ";
  os << "   Value: " << setprecision(10) << p -> value << endl;
 }
 return os << "\n" << endl;
}

char operator==( const JLC& x, const JLC& y ) {
 dlist_iterator getNextX( (dlist&) x );
 dlist_iterator getNextY( (dlist&) y );
 JLCterm* p;
 JLCterm* q;
 int i;
 
 if( ( x.count   != y.count )   ||
     ( x.weight  != y.weight )  ||
     ( x.accuWgt != y.accuWgt )
   ) return 0;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   if( x.refPoint[i] != y.refPoint[i] ) return 0;
 
 while( p = (JLCterm*) getNextX() ) {
   if( !( q = (JLCterm*) getNextY() ) ) return 0;
   if( !( *p == *q ) ) return 0;
   }
 
 if( (JLCterm*) getNextY() ) {
   printf( "\n*** ERROR: Inconsistency in JLC operator==\n" );
     // This point should never be reached, since that would imply that
     // x.count != y.count
   return 0;
   }
 
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const JLC& x, const complex& y ) {
 char result = 1;
 dlist_iterator getNext( (dlist&) x );
 JLCterm* p;

 if( x.count < 1 ) {
  if( y == complex_0 ) return 1;
  else           return 0;
 }

 while( p = (JLCterm*) getNext() ) 
  result = result && ( p->weight == 0 ? p->value == y : 
                                        p->value == complex_0 );

  // REMOVE ??? result = result && ( ( ( p->weight == 0 ) && ( p->value == y   ) )
  // REMOVE ???                      ||
  // REMOVE ???                      ( ( p->weight != 0 ) && ( p->value == 0.0 ) )
  // REMOVE ???                    )
  // REMOVE ??? 
 return result;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const complex& y, const JLC& x ){
 return x == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JLC& x, const JLC& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JLC& x, const complex& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const complex& x, const JLC& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC& JLC::operator=( const complex& x ) {
 int* ndx;
 int i;
 
 if( JetC::oneEnv.allSet != 137 ) {
   printf( "\n*** JLC& JLC::operator=( const complex& x ) { ERROR *** " );
   printf( "Prior call to JLCSetup required.\n" );
   exit(0);
   }
 
 clear();
 weight = 0;   // rc is purposely left untouched!!
 count = 0;
 accuWgt = JetC::oneEnv.MaxWeight;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = JetC::oneEnv.refPoint[i];

                       // ??? There is some dacnger here!  This operator
                       //     governs the assignment of a complex to JLC.
 
 ndx = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) ndx[i] = 0;
 insert( new JLCterm( ndx, x ) );
 delete [] ndx;
 return *this;
}
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC& JLC::operator=( const JLC& x ) {
 dlist_iterator getNext( (dlist&) x );
 JLCterm* p;
 JLCterm* q;
 int i;
 
 if( this == &x ) return *this;
 
 clear();
 while( p = (JLCterm*) getNext() ) {
   q = new JLCterm( p );
   append( q );
   }

 count   = x.count;     // rc is purposely left untouched
 weight  = x.weight;
 accuWgt = x.accuWgt;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) refPoint[i] = x.refPoint[i];
 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLC& JLC::operator+=( const complex& x ) {   // ??? Untested!!
 static char firstNull = 1;
 int i;
 dlist_iterator getNext( *(dlist*) this );
 JLCterm* p;
 JLCterm* q;

 if( p = (JLCterm*) getNext() )
 {
   if ( p->weight == 0 ) { 
    p->value += x;
    if( abs( p->value ) <  MX_SMALL*abs(x) ) p->value = 0.0;
   }
  
   else {
     q = new JLCterm;
     for( i = 0; i < JetC::oneEnv.NumVar; i++ ) q->index[i] = 0;
     q->weight = 0;
     q->value = x;
     insert( q );
   } 
 }
 else
 {
   if( firstNull ) {
     cerr << "*** WARNING ***                                 \n"
             "*** WARNING *** JLC& JLC::operator+=( const complex& )\n"
             "*** WARNING *** Operator invoked by a null      \n"
             "*** WARNING *** JLC.  This warning will not be   \n"
             "*** WARNING *** repeated.                       \n"
             "*** WARNING ***                                 \n"
          << endl;
     firstNull = 0;
   }
   operator=( x );
 }

 return *this;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm operator*( JLCterm& x, JLCterm& y ) {  
 JLCterm z;
 int i;
 z.weight = x.weight + y.weight;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   z.index[i] = x.index[i] + y.index[i];
 z.value = x.value * y.value;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class JLCterm

JLCterm::JLCterm() {
 weight = 0;
 index = new int[ JetC::oneEnv.NumVar ];
 value = 0.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::JLCterm( int* l, complex x ) {
 int i, dpt;
 
 dpt = 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( (l[i] < 0) ) {
     printf( "\n*** JLC ERROR *** Bad index in JLCTerm.\n" );
     exit(0);
     }
   dpt += l[i];
   }
 if( dpt > JetC::oneEnv.MaxWeight ) {
   printf("\n*** JLC ERROR *** Attempt to load a JLCTerm with too large a weight.\n");
   exit(0);
   }
 
 index = new int[ JetC::oneEnv.NumVar ];
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = l[i];
 weight = dpt;
 value = x;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::JLCterm( JLCterm* x ) {
 int i;
 
 weight = x->weight;
 value = x->value;
 index = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = (x->index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::JLCterm( JLCterm& x ) {
 int i;
 
 weight = x.weight;
 value = x.value;
 index = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::JLCterm( JLterm* x ) {
 int i;
 
 weight = x->weight;
 value = complex(x->value,0.0);
 index = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = (x->index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::JLCterm( JLCtermData& x ) {
 int i;
 
 weight = x.weight;
 value = x.value;
 index = new int[ JetC::oneEnv.NumVar ];
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm::~JLCterm() {
 delete [] index;
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of struct JLCtermData

JLCtermData::JLCtermData() {
 printf( "\n" );
 printf( "*** ERROR ***                                   \n" );
 printf( "*** ERROR *** JLCtermData::JLCtermData()          \n" );
 printf( "*** ERROR *** Cannot declare a JLCtermData       \n" );
 printf( "*** ERROR *** in this way.                      \n" );
 printf( "*** ERROR ***                                   \n" );
 exit(0);
}

JLCtermData::JLCtermData( JLCterm& x ) {
 numvar = JetC::oneEnv.NumVar;   // ??? This has got to be changed!!
 index = new int[ numvar ];
 for( int i = 0; i < numvar; i++ ) index[i] = x.index[i];
 weight = x.weight;
 value = x.value;
}

JLCtermData::~JLCtermData() {
 if( index != 0 ) delete [] index;
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Overloaded operators for class JLCterm

void JLCterm::operator=( const JLCterm& x ) {
 int i;
 weight = x.weight;
 value = x.value;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) index[i] = (x.index)[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const JLCterm& a, const JLCterm& b ) {
 int i;
 if( ( a.weight != b.weight) || ( a.value != b.value)  ) return 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   if( ( a.index[i] != b.index[i]) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JLCterm& a, const JLCterm& b ) {
 return !( a == b );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator<=( JLCterm& a, JLCterm& b ) {
 int i;
 if( a.weight != b.weight ) return ( a.weight < b.weight );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( a.index[i] == b.index[i] ) continue;
   return ( a.index[i] < b.index[i] );
   }
 return 1;  // when all indices are the same.
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator%=( JLCterm& a, JLCterm& b ) {
 int i;
 if( a.weight != b.weight ) return 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
   if( ( a.index[i] != b.index[i]) ) return 0;
 return 1;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
