#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================
//      The environment
//

Jet__environment::Jet__environment() {
  monomial           = 0;
  JLmonomial         = 0;
  exponent           = 0;
  expCode            = 0;
  numPaths           = 0;
  NumVar             = 0;
  MaxWeight          = 0;
  allSet             = 0;
  weightLimit        = 0;    
  SpaceDim           = -1;
  dof                = 0;            
  refPoint           = 0;       
  scale              = 0;
  AllZeroes          = 0;      
}
 
 
Jet__environment::~Jet__environment() {
  if( allSet == 137 ) {
   delete [] monomial;       
   if( JLmonomial ) delete [] JLmonomial;     
   delete [] exponent;       
   delete [] expCode;        
   delete    numPaths;
   delete [] refPoint;       
   delete [] scale;
   delete [] AllZeroes;      
  
   // ??? JL* p;
   // ??? while( p = (JL*) predefJL.get() ) delete p;
   // ??? LieOperator* q;
   // ??? while( q = (LieOperator*) predefLieOp.get() ) delete q;
  }
}

//***************************************************************
//***************************************************************
//***************************************************************
//  
//  JetLetter::JetLetter() : rc(1) {
//   theJet = new JL;
//  }
//  
//  JetLetter::JetLetter( const char* ) : rc(0) {
//   theJet = NULL;
//  }
//  
//  JetLetter::JetLetter( const double x ) : rc(1) {
//   theJet = new JL( x );
//  }
//  
//  JetLetter::JetLetter( const JetLetter& x ) {
//   static char firstCall = 1;
//   if( firstCall ) {
//    firstCall = 0;
//    cout << "\n*** WARNING *** The JetLetter copy constructor was called for \n"
//         <<   "*** WARNING *** some inexplicable reason.\n"
//         <<   "*** WARNING *** Go figure!!\n\n";
//    cout.flush();
//   }
//   theJet = new JL( x.theJet );
//   rc = 1;
//  }
//  
//  JetLetter::~JetLetter() {
//   delete theJet;
//  }
//  
//  JetLetter& JetLetter::operator=( const JetLetter& x ) {
//   *theJet = *(x.theJet);
//   rc = 1;
//   return *this;
//  }
//  
//***************************************************************
//***************************************************************
//***************************************************************

Jet::Jet() : stacked(0) {
 jl = new JL;
 if( jl->refPoint == NULL ) 
  Jet::oneEnv.predefJL.insert( this );
}

Jet::Jet( const double x ) : stacked(0) {
 jl = new JL( x );
}

Jet::Jet( const Jet& x ) {
 stacked = x.stacked;
 jl = x.jl;
 ( jl->rc )++;
}

Jet::~Jet() {
 if( --(jl->rc) == 0 ) delete jl;
}

void Jet::fixReferenceAtStart( const LieOperator& x ) {
 jl->fixReferenceAtStart( x );
}

void Jet::fixReferenceAtEnd( const LieOperator& x ) {
 jl->fixReferenceAtEnd( x );
}

void Jet::getReference( double* r ) {
 jl->getReference( r );
}

void Jet::setVariable( const double x, const int j ) {
 if( jl == NULL ) jl = new JL;
 jl->setVariable( x, j );
}

void Jet::setVariable( int j ) {
 if( jl == NULL ) jl = new JL;
 jl->setVariable( j );
}

Jet& Jet::operator=( const Jet& x ) {
 if( this != &x ) {
  if( --(jl->rc) == 0 ) delete jl;
  if( x.stacked ) {
   jl = x.jl;
   (jl->rc)++;
  }
  else { 
   jl = new JL;
   *jl = *(x.jl);
  }
 }
 stacked = 0;   // ??? This line is probably unnecessary.
 return *this; 
}

Jet& Jet::operator=( const double& x ) {
 if( --(jl->rc) == 0 ) delete jl;
 jl = new JL;
 *(jl) = x;
 stacked = 0;
 return *this; 
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Initialization routines.
//      Must be called first by any program in order to establish the
//      number of (scalar) variables, n,
//      maximum weight of derivatives, w,
//      dimension of phase space, s,
//      default reference point, r.

void Jet::setup( const int n, const int w ) {
  Jet::Setup( n, w );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::setup( const int n, const int w, const double* r ) {
  Jet::Setup( n, w, n, r );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::setup( const int n, const int w, const int s ) {
  Jet::Setup( n, w, s );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::setup( const int n, const int w, const int s, const double* r ) {
  Jet::Setup( n, w, s, r );
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::Setup( const int     n, 
                 const int     w, 
                 const int     s, 
                 const double* r,
                 const double* scl 
               ) 
{
 int i, j;
 Jet* p;
 LieOperator* q;
 
 if( Jet::oneEnv.allSet == 137 )
   cout << "\n*** JL ERROR: Attempt to call Jet::setup twice.\n" << endl;
 
 else {
 
   // Some error checking ..
   if( n < 1 ) {
     cout << "\n*** JL ERROR: Jet::Setup requires at least one variable." << endl;
     exit(0);
     }
 
   if( w < 1 ) {
     cout << "\n*** JL ERROR: Jet::Setup requires minimal maximum weight of 1." << endl;
     exit(0);
     }
 
   if ( n < s ) {
     cout <<  "\n*** JL ERROR: Jet::Setup called with incompatible parameters." 
              "\n**********************************************************\n"  << endl;
     exit(0);
     }
 
   Jet::oneEnv.allSet = 137;
 
   // The default reference point is set.
   Jet::oneEnv.refPoint = new double[ n ];
   if( r )  for( i = 0; i < n; i++ ) Jet::oneEnv.refPoint[i] = r[i];
   else     for( i = 0; i < n; i++ ) Jet::oneEnv.refPoint[i] = 0.0;
 
   // Set global variables.
   Jet::oneEnv.NumVar = n;
   Jet::oneEnv.MaxWeight = w;
   Jet::oneEnv.exponent = new int[ n ];
   Jet::oneEnv.expCode = new char[ w + n ];
   Jet::oneEnv.monomial = new double[ bcfRec( w + n, n ) ];
   Jet::oneEnv.JLmonomial = new Jet[ bcfRec( w + n, n ) ];
   Jet::oneEnv.SpaceDim = s;  if( Jet::oneEnv.SpaceDim == 0 )
                                  Jet::oneEnv.SpaceDim  = n;
   Jet::oneEnv.dof = Jet::oneEnv.SpaceDim / 2;
   Jet::oneEnv.PBOK = ( 2*Jet::oneEnv.dof == Jet::oneEnv.SpaceDim );
 
   Jet::oneEnv.AllZeroes = new int[ n ];
   for( i = 0; i < n; i++ ) Jet::oneEnv.AllZeroes[i] = 0;
 
   // Go through all predefined JL variables: set the reference
   // and weight accuracy.
   while( p = (Jet*) Jet::oneEnv.predefJL.get() ) {
     p->jl->refPoint = new double[ Jet::oneEnv.NumVar ];
     for( i = 0; i < Jet::oneEnv.NumVar; i++ )
       (p->jl->refPoint)[i] = Jet::oneEnv.refPoint[i];
     p->jl->accuWgt = Jet::oneEnv.MaxWeight;
     }
   while( q = (LieOperator*) Jet::oneEnv.predefLieOp.get() ) {
     q->v = new Jet[ Jet::oneEnv.NumVar ];
     q->refPoint = new double[ Jet::oneEnv.NumVar ];
     for( i = 0; i < Jet::oneEnv.NumVar; i++ ) (q->refPoint)[i] = Jet::oneEnv.refPoint[i];
     }
 
   // Load the Jet::oneEnv.numPaths array with binomial coefficients;
   // required by Wilf's algorithm for ranking monomials.
   Jet::oneEnv.numPaths = new MatrixI( w+1, n );
   for( i = 0; i <= w; i++ )
     for( j = 1; j <= n; j++ )
       (*Jet::oneEnv.numPaths)( i, j-1 ) = bcfRec( i + j - 1, i );
   // ??? REMOVE: Jet::oneEnv.numPaths = imatrix( 0, w, 1, n );
   // ??? REMOVE: for( i = 0; i <= w; i++ )
   // ??? REMOVE:   for( j = 1; j <= n; j++ )
   // ??? REMOVE:     Jet::oneEnv.numPaths[i][j] = bcfRec( i + j - 1, i );
 
   // Initialize the Jet::coordPool dlist
   int* counter;
   for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
     counter = new int(i);
     Jet::coordPool.append( counter );
     }
 
   }
 
   // Set the scale array
   Jet::oneEnv.scale = new double[ Jet::oneEnv.NumVar ];
   if( scl )  for( i = 0; i < Jet::oneEnv.NumVar; i++ )  Jet::oneEnv.scale[i] = fabs(scl[i]);
   else       for( i = 0; i < Jet::oneEnv.NumVar; i++ )  Jet::oneEnv.scale[i] = 1.0;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::FixReference( const double* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  Jet::oneEnv.refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::FixReference( const int* r ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  Jet::oneEnv.refPoint[i] = (double) (r[i]);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::FixReference( const Jet& x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  Jet::oneEnv.refPoint[i] = x->refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::FixReferenceAtStart( const LieOperator& x ) {
 int i;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  Jet::oneEnv.refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::FixReferenceAtEnd( const LieOperator& x ) {
 int i;
 double* u = new double [ Jet::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) 
  Jet::oneEnv.refPoint[i] = u[i];
 delete [] u;
}

//***************************************************************
//***************************************************************
//***************************************************************
//
//    Implementation of class coord
//
 
coord::coord( const double x ) : Jet() {
 int* n;
 coord* y;
 
 if( oneEnv.allSet != 137 ) {
   printf( "\n*** ERROR ***                                       \n" );
   printf(   "*** ERROR *** coord::coord( double )                \n" );
   printf(   "*** ERROR *** Jet::setup must be invoked first.     \n" );
   printf(   "*** ERROR ***                                       \n" );
   exit(1);
   }
 
 if( ( n = (int*) Jet::coordPool.get() ) == 0 ) {
   printf( "\n*** ERROR ***                                       \n" );
   printf(   "*** ERROR *** coord::coord( double )                \n" );
   printf(   "*** ERROR *** Attempt made to declare too many      \n" );
   printf(   "*** ERROR *** coordinates, according to what was    \n" );
   printf(   "*** ERROR *** declared in Jet::setup.                  \n" );
   printf(   "*** ERROR ***                                       \n" );
   exit(1);
   }
 
 index = *n;
 this->setVariable( x, index );   // ??? Why use this-> here?
 delete n;
 
 Jet::coordPtrs.append( this );
 dlist_iterator getNext( Jet::coordPtrs );
 while( y = (coord*) getNext() ) y->fixReference();

 /* DGN  cout << "DGN> ";*/
 /* DGN   for( int j = 0; j < Jet::oneEnv.NumVar; j++ )*/
 /* DGN     cout << Jet::oneEnv.refPoint[j] << "  ";*/
 /* DGN   cout << endl;*/
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coord::~coord() {
 int* n;
 char c;
 n = new int( index );
 Jet::coordPool.insert( n );
 c = Jet::coordPtrs.remove( this );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coord::coord( const coord&  ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coord::coord( coord& )                \n" );
 printf(   "*** ERROR *** Coordinate copy constructor called.   \n" );
 printf(   "*** ERROR ***                                       \n" );
 exit(1);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void coord::set( const double x ) {
 coord* y;
 this->setVariable( x, index );
 dlist_iterator getNext( Jet::coordPtrs );
 while( y = (coord*) getNext() ) y->fixReference();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coord& coord::operator=( const coord& ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coord::operator=                      \n" );
 printf(   "*** ERROR *** Attempt made to change the value      \n" );
 printf(   "*** ERROR *** of a coordinate.                      \n" );
 printf(   "*** ERROR ***                                       \n" );
 if( 3 + 8 == 11 ) exit(1);  // To avoid compiler warnings.
 return *this;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coord& coord::operator=( const Jet& ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coord::operator=                      \n" );
 printf(   "*** ERROR *** Attempt made to change the value      \n" );
 printf(   "*** ERROR *** of a coordinate.                      \n" );
 printf(   "*** ERROR ***                                       \n" );
 if( 3 + 8 == 11 ) exit(1);  // To avoid compiler warnings.
 return *this;
}


//***************************************************************
//***************************************************************
//***************************************************************
//
//      Functions which rank and unrank monomials.  Developed from
//      algorithms provided by Nijenhuis and Wilf, Combinatorial
//      Algorithms. (2nd ed.)
 
void Jet::monoCode() {
 int c;
 int i, j;
 j = 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   for( c = 0; c < Jet::oneEnv.exponent[i]; c++ ) Jet::oneEnv.expCode[j++] = 1;
   Jet::oneEnv.expCode[j++] = 0;
   }
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void Jet::monoDecode() {
 int c;
 int i, j;
 j = 0;
 for( i = 0; i < Jet::oneEnv.NumVar; i++ ) {
   c = 0;
   while( Jet::oneEnv.expCode[j++] ) ++c;
   Jet::oneEnv.exponent[i] = c;
   }
}

//    |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

int Jet::monoRank() {
 //  This function assumes Jet::monoCode has been run,
 //  so that the path is appropriately coded into Jet::oneEnv.expCode.
 char edge;
 int w, n;
 int rank;
 int zeroCount;
 int i;
 
 i = 0;
 rank = 0;
 w = 0;
 n = 1;
 
 if( edge = Jet::oneEnv.expCode[i++] ) zeroCount = 0;
 else                          zeroCount = 1;
 
 while( zeroCount < Jet::oneEnv.NumVar ) {
 
   if( edge ) {
     w++;
     if( ( w > 0 ) && ( n > 1 ) )  rank += (*Jet::oneEnv.numPaths)(w, n-2);
     // ??? REMOVE: if( ( w > 0 ) && ( n > 1 ) )  rank += Jet::oneEnv.numPaths[w][n-1];
     }
   else  n++;
 
   if( !( edge = Jet::oneEnv.expCode[i++] ) ) zeroCount++;
 
   }
 
 rank += ( w * (*Jet::oneEnv.numPaths)(w, n-1) ) / n;
 // ??? REMOVE: rank += ( w * Jet::oneEnv.numPaths[w][n] ) / n;
 
 return rank;
}

