#include "mxyzptlk.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}


// ================================================================
//      The environment
//

JetC__environment::JetC__environment() {
  monomial           = 0;
  JLmonomial         = 0;
  exponent           = 0;
  expCode            = 0;
  numPaths           = 0;  // ??? Is JetC::numPaths actually used anywhere???
  NumVar             = 0;  // ??? I think not.  Why not get rid of it???
  MaxWeight          = 0;
  allSet             = 0;
  weightLimit        = 0;    
  SpaceDim           = -1;
  dof                = 0;            
  refPoint           = 0;       
  AllZeroes          = 0;      
}
 
 
JetC__environment::~JetC__environment() {
  if( allSet == 137 ) {
   delete [] monomial;       
   if( JLmonomial ) delete [] JLmonomial;     
   delete [] exponent;       
   delete [] expCode;        
   delete    numPaths;
   // ??? REMOVE: free_imatrix( numPaths, 0, MaxWeight, 1, NumVar ); // ???
   // ??? REMOVE:                               // ??? Carrying this around is stupid!!
   delete [] refPoint;       
   delete [] AllZeroes;      
  
   // ??? JLC* p;
   // ??? while( p = (JLC*) predefJL.get() ) delete p;
   // ??? LieOperator* q;
   // ??? while( q = (LieOperator*) predefLieOp.get() ) delete q;
  }
}

JetC::JetC() : stacked(0) {
 jl = new JLC;
 if( jl->refPoint == NULL ) 
  JetC::oneEnv.predefJL.insert( this );
}

JetC::JetC( const complex x ) : stacked(0) {
 jl = new JLC( x );
}

JetC::JetC(const Jet&x) : stacked(0) {
 jl = new JLC(*x.jl);
 if( jl->refPoint == NULL ) 
  JetC::oneEnv.predefJL.insert( this );
}


JetC::JetC( const JetC& x ) {
 stacked = x.stacked;
 jl = x.jl;
 ( jl->rc )++;
}

JetC::~JetC() {
 if( --(jl->rc) == 0 ) delete jl;
}

void JetC::fixReferenceAtStart( const CLieOperator& x ) {
 jl->fixReferenceAtStart( x );
}

void JetC::fixReferenceAtEnd( const CLieOperator& x ) {
 jl->fixReferenceAtEnd( x );
}

void JetC::getReference( complex* r ) {
 jl->getReference( r );
}

void JetC::setVariable( const complex x, const int j ) {
 if( jl == NULL ) jl = new JLC;
 jl->setVariable( x, j );
}

void JetC::setVariable( int j ) {
 if( jl == NULL ) jl = new JLC;
 jl->setVariable( j );
}

JetC& JetC::operator=( const JetC& x ) {
 if( this != &x ) {
  if( --(jl->rc) == 0 ) delete jl;
  if( x.stacked ) {
   jl = x.jl;
   (jl->rc)++;
  }
  else { 
   jl = new JLC;
   *jl = *(x.jl);
  }
 }
 stacked = 0;   // ??? This line is probably unnecessary.
 return *this; 
}

JetC& JetC::operator=( const complex& x ) {
 if( --(jl->rc) == 0 ) delete jl;
 jl = new JLC;
 *(jl) = x;
 stacked = 0;
 return *this; 
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Initialization routines.
//      Must be called first by any program in order to establish the
//      number of (scalar) variables, n,
//      maximum weight of derivatives, w,
//      dimension of phase space, s,
//      default reference point, r.

void JetC::setup( const int n, const int w ) {
  JetC::Setup( n, w );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::setup( const int n, const int w, const complex* r ) {
  JetC::Setup( n, w, n, r );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::setup( const int n, const int w, const int s ) {
  JetC::Setup( n, w, s );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::setup( const int n, const int w, const int s, const complex* r ) {
  JetC::Setup( n, w, s, r );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::Setup( const int      n, 
                  const int      w, 
                  const int      s, 
                  const complex* r
                ) 
{
 int i, j;
 JetC* p;
 CLieOperator* q;
 
 if( JetC::oneEnv.allSet == 137 )
   cout <<  "\n*** JLC ERROR: Attempt to call JetC::setup twice.\n" << endl;
 
 else {
 
   // Some error checking ..
   if( n < 1 ) {
     cout << "\n*** JLC ERROR: JetC::setup requires at least one variable."<< endl;
     exit(0);
     }
 
   if( w < 1 ) {
     cout << "\n*** JLC ERROR: JetC::setup requires minimal maximum weight of 1."<< endl;
     exit(0);
     }
 
   if ( n < s ) {
     cout <<  "\n*** JLC ERROR: JetC::setup called with incompatible parameters." 
              "\n**********************************************************\n" << endl;
     exit(0);
     }
 
   JetC::oneEnv.allSet = 137;
 
   // The default reference point is set.
   JetC::oneEnv.refPoint = new complex[ n ];
   if( r )  for( i = 0; i < n; i++ ) JetC::oneEnv.refPoint[i] = r[i];
   else     for( i = 0; i < n; i++ ) JetC::oneEnv.refPoint[i] = 0.0;
 
   // Set global variables.
   JetC::oneEnv.NumVar = n;
   JetC::oneEnv.MaxWeight = w;
   JetC::oneEnv.exponent = new int[ n ];
   JetC::oneEnv.expCode = new char[ w + n ];
   JetC::oneEnv.monomial = new complex[ bcfRec( w + n, n ) ];
   JetC::oneEnv.JLmonomial = new JetC[ bcfRec( w + n, n ) ];
   JetC::oneEnv.SpaceDim = s;  if( JetC::oneEnv.SpaceDim == 0 )
                                   JetC::oneEnv.SpaceDim  = n;
   JetC::oneEnv.dof = JetC::oneEnv.SpaceDim / 2;
   JetC::oneEnv.PBOK = ( 2*JetC::oneEnv.dof == JetC::oneEnv.SpaceDim );
 
   JetC::oneEnv.AllZeroes = new int[ n ];
   for( i = 0; i < n; i++ ) JetC::oneEnv.AllZeroes[i] = 0;
 
   // Go through all predefined JLC variables: set the reference
   // and weight accuracy.
   while( p = (JetC*) JetC::oneEnv.predefJL.get() ) {
     p->jl->refPoint = new complex[ JetC::oneEnv.NumVar ];
     for( i = 0; i < JetC::oneEnv.NumVar; i++ )
       (p->jl->refPoint)[i] = JetC::oneEnv.refPoint[i];
     p->jl->accuWgt = JetC::oneEnv.MaxWeight;
     }
   while( q = (CLieOperator*) JetC::oneEnv.predefLieOp.get() ) {
     q->v = new JetC[ JetC::oneEnv.NumVar ];
     q->refPoint = new complex[ JetC::oneEnv.NumVar ];
     for( i = 0; i < JetC::oneEnv.NumVar; i++ ) (q->refPoint)[i] = JetC::oneEnv.refPoint[i];
     }
 
   // Load the JetC::oneEnv.numPaths array with binomial coefficients;
   // required by Wilf's algorithm for ranking monomials.
   JetC::oneEnv.numPaths = new MatrixI( w+1, n );
   for( i = 0; i <= w; i++ )
     for( j = 1; j <= n; j++ )
       (*JetC::oneEnv.numPaths)( i, j-1 ) = bcfRec( i + j - 1, i );
   // ??? REMOVE: JetC::oneEnv.numPaths = imatrix( 0, w, 1, n );
   // ??? REMOVE: for( i = 0; i <= w; i++ )
   // ??? REMOVE:   for( j = 1; j <= n; j++ )
   // ??? REMOVE:     JetC::oneEnv.numPaths[i][j] = bcfRec( i + j - 1, i );
 
   // Initialize the JetC::coordPool dlist
   int* counter;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
     counter = new int(i);
     JetC::coordPool.append( counter );
     }
 
   }
 
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::FixReference( const complex* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  JetC::oneEnv.refPoint[i] = r[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::FixReference( const int* r ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  JetC::oneEnv.refPoint[i] = (complex) (r[i]);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::FixReference( const JetC& x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  JetC::oneEnv.refPoint[i] = x->refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::FixReferenceAtStart( const CLieOperator& x ) {
 int i;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  JetC::oneEnv.refPoint[i] = x.refPoint[i];
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::FixReferenceAtEnd( const CLieOperator& x ) {
 int i;
 complex* u = new complex [ JetC::oneEnv.NumVar ];
 x.standardPart( u );
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  JetC::oneEnv.refPoint[i] = u[i];
 delete [] u;
}

//***************************************************************
//***************************************************************
//***************************************************************
//
//    Implementation of class coordC
//
 
coordC::coordC( const complex x ) : JetC() {
 int* n;
 coordC* y;
 
 if( oneEnv.allSet != 137 ) {
   printf( "\n*** ERROR ***                                       \n" );
   printf(   "*** ERROR *** coordC::coordC( complex )                \n" );
   printf(   "*** ERROR *** JetC::setup must be invoked first.        \n" );
   printf(   "*** ERROR ***                                       \n" );
   exit(1);
   }
 
 if( ( n = (int*) JetC::coordPool.get() ) == 0 ) {
   printf( "\n*** ERROR ***                                       \n" );
   printf(   "*** ERROR *** coordC::coordC( complex )                \n" );
   printf(   "*** ERROR *** Attempt made to declare too many      \n" );
   printf(   "*** ERROR *** coordCinates, according to what was    \n" );
   printf(   "*** ERROR *** declared in JetC::setup.                  \n" );
   printf(   "*** ERROR ***                                       \n" );
   exit(1);
   }
 
 index = *n;
 this->setVariable( x, index );   // ??? Why use this-> here?
 delete n;
 
 JetC::coordPtrs.append( this );
 dlist_iterator getNext( JetC::coordPtrs );
 while( y = (coordC*) getNext() ) y->fixReference();

}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coordC::~coordC() {
 int* n;
 char c;
 n = new int( index );
 JetC::coordPool.insert( n );
 c = JetC::coordPtrs.remove( this );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coordC::coordC( const coordC&  ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coordC::coordC( coordC& )                \n" );
 printf(   "*** ERROR *** Coordinate copy constructor called.   \n" );
 printf(   "*** ERROR ***                                       \n" );
 exit(1);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void coordC::set( const complex x ) {
 coordC* y;
 this->setVariable( x, index );
 dlist_iterator getNext( JetC::coordPtrs );
 while( y = (coordC*) getNext() ) y->fixReference();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coordC& coordC::operator=( const coordC& ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coordC::operator=                      \n" );
 printf(   "*** ERROR *** Attempt made to change the value      \n" );
 printf(   "*** ERROR *** of a coordinate.                      \n" );
 printf(   "*** ERROR ***                                       \n" );
 if( 3 + 8 == 11 ) exit(1);  // To avoid compiler warnings.
 return *this;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

coordC& coordC::operator=( const JetC& ) {
 printf( "\n*** ERROR ***                                       \n" );
 printf(   "*** ERROR *** coordC::operator=                      \n" );
 printf(   "*** ERROR *** Attempt made to change the value      \n" );
 printf(   "*** ERROR *** of a coordinate.                      \n" );
 printf(   "*** ERROR ***                                       \n" );
 if( 3 + 8 == 11 ) exit(1);  // To avoid compiler warnings.
 return *this;
}
