#include "mxyzptlk.hxx"
#include "MathConstants.hxx"

// ================================================================
//      External routines
//

extern char nexcom( int, int, int* );  
                                // Computes the next composition
                                //  of an integer into a number of parts.
                                //  Algorithm devised by Herbert Wilf.

extern "C" {
 int bcfRec( int, int );        // Recursive evaluation of binomial
                                //  coefficient.
 int nearestInteger( double );  // Returns the integer nearest to its
                                //  double argument.
}

// ================================================================

#ifndef MX_SMALL
#define MX_SMALL    1.0e-12 // Used by JLC::addTerm to decide 
                            //   removal of a JLCterm.
#endif  // MX_SMALL

#ifndef MX_MAXITER
#define MX_MAXITER  100     // Maximum number of iterations allowed
                            //   in iterative routines
#endif  // MX_MAXITER

static const complex complex_one( 1.0,0.0 );   // ??? This should not be here!!!

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void operator+=( JetC& x, const JetC& y ) {
 x = x + y;  // ??? Come up with a more efficient implementation SOON.
}

void operator-=( JetC& x, const JetC& y ) {
 x = x - y;
}

void operator*=( JetC& x, const JetC& y ) {
 x = x * y;
}

void operator/=( JetC& x, const JetC& y ) {
 x = x / y;
}

void operator/=( JetC& x, const complex y ) {
 x = x / y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

//     Member functions(public)  |||||||||||||||||||||||||||||||

void JetC::addTerm( JLCterm* a) {
 jl->addTerm( a );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::fixReference( const complex* r ) {
 jl->fixReference( r );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::fixReference() {
 jl->fixReference();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||


//***************************************************************
//***************************************************************
//***************************************************************
//
//      The overloaded operators for class JetC
//
istream& operator>>(istream& is,  JetC& x) {
 return operator>>( is, *(x.jl) );
}

ostream& operator<<(ostream& os, JetC& x) {
 return operator<<( os, *(x.jl) );
}

char operator==( const JetC& x, const JetC& y ) {
 return *(x.jl) == *(y.jl);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const JetC& x, const complex& y ) {
 return *(x.jl) == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator==( const complex& y, const JetC& x ){
 return *(x.jl) == y;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JetC& x, const JetC& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const JetC& x, const complex& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char operator!=( const complex& x, const JetC& y ) {
 return !( x == y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::operator+=( const complex& x ) {   // ??? Untested!!
 if( jl == NULL ) jl = new JLC;              // ??? JLNGER: what is the 
                                            // ??? reference point?
 jl->operator+=( x );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator+( const JetC& x, const JetC& y ) { 
                                       // ??? This should be modified so that
 JetC z;                                // terms beyond the accurate weight of
 JLCterm* p;                            // x or y are not computed and carried
 JLCterm* q;                            // into the answer.
 JLC* xPtr = x.jl;
 JLC* yPtr = y.jl;
 JLC* zPtr = z.jl;

 dlist_iterator gx( *(dlist*) xPtr );
 dlist_iterator gy( *(dlist*) yPtr );
 int i;
 
 // Check for consistency and set reference point of the sum.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( abs( xPtr->refPoint[i] - yPtr->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << "\n*** JLC ADDITION ERROR: Inconsistent reference points." << endl;
     exit(0);
   }
   zPtr->refPoint[i] = xPtr->refPoint[i];
 }
 
 // If one of the arguments is void, then return the other ..
 if( xPtr->count < 1 ) return y;   // ??? This will not work.  What's 
 if( yPtr->count < 1 ) return x;   // ??? returned owns the same data
                                   // ??? as x or y!!  Change throughout!!
 // .. otherwise, continue normal operations.
 if( xPtr->count > yPtr->count ) {
   *zPtr = *xPtr;
   while( p = (JLCterm*) gy() ) {
     q = new JLCterm( p );
     zPtr->addTerm( q );
   }
 }
 else {
   *zPtr = *yPtr;
   while( p = (JLCterm*) gx() ) {
     q = new JLCterm( p );
     zPtr->addTerm( q );
   }
 }
 
 // Determine the maximum weight computed accuratelyPtr->
 if( xPtr->accuWgt < yPtr->accuWgt ) zPtr->accuWgt = xPtr->accuWgt;
 else                                zPtr->accuWgt = yPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator+( const JetC& x, const complex& y ) {
 JetC z;
 z = x;
 z += y;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator+( const complex& y, const JetC& x ) {
 JetC z;
 z = x;
 z += y;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator-( const JetC& x, const complex& y ) {
  return x + (-y);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator-( const complex& y, const JetC& x ) {
  return y + (-x);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator-( const JetC& y ) {  
 JetC z;
 JLC* yPtr = y.jl;
 JLC* zPtr = z.jl;   // ??? This is a dangerous strategy. 
 dlist_iterator getNext( *(dlist*) yPtr );  // ??? It will fail if
 JLCterm* p;                 // ??? The statement "z = ...: appears.         
 JLCterm* q;
 int i;
 
 // If the argument is void, then return it ..
 if( y->count < 1 ) return y;

 // .. otherwise continue normal operations.
 zPtr->clear();
 
 // Insert the first one
 p = (JLCterm*) getNext();
 q = new JLCterm;
 *q = *p;
 q->value = - ( p->value );
 zPtr->insert( q );
 
 // and append the rest
 while( p = (JLCterm*) getNext() ){
   q = new JLCterm;
   *q = *p;
   q->value = - ( p->value );
   zPtr->append( q );
   }
 
 zPtr->weight  = y->weight;
 zPtr->accuWgt = y->accuWgt;
 zPtr->count   = y->count;
 
 // Set the reference point.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = yPtr->refPoint[i];
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator-( const JetC& x, const JetC& y ) {  
 JetC z;
 JLCterm* p;
 JLCterm* q;
 dlist_iterator gx( *(dlist*) x.jl );
 dlist_iterator gy( *(dlist*) y.jl );
 int i;
 
 // Check for consistency and set reference point of the difference.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( abs( x->refPoint[i] - y->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << "\n*** JLC SUBTRACTION ERROR: Inconsistent reference points." << endl;
     exit(0);
     }
   z->refPoint[i] = x->refPoint[i];
   }
 
 // If one of the arguments is void, then return the other ..
 if( x->count < 1 ) return -y;
 if( y->count < 1 ) return x;

 // .. otherwise, continue normal operations.
 if( x->count > y->count ) {
   z = x;
   while( p = (JLCterm*) gy() ) {
     q = new JLCterm( p );
     ( q->value ) = - ( q->value );
     z->addTerm( q );
     }
   }
 
 else {
   z = - y;
   while( p = (JLCterm*) gx() ) {
     q = new JLCterm( p );
     z->addTerm( q );
     }
   }
 
 // Determine the maximum weight computed accurately.
 if( x->accuWgt < y->accuWgt ) z->accuWgt = x->accuWgt;  // ??? Change this too, as in
 else                          z->accuWgt = y->accuWgt;  // ??? operaotr+
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const JetC& x, const JetC& y ) {        
                                       // ??? This should be modified so that
 JetC z;                                // terms beyond the accurate weight of
 JLCterm* p;                            // x or y are not computed and carried
 JLCterm* q;                            // into the answer.
 JLCterm* r;

 JLC* xPtr = x.jl;
 JLC* yPtr = y.jl;
 JLC* zPtr = z.jl;

 dlist_looper gx( *(dlist*) xPtr );
 dlist_looper gy( *(dlist*) yPtr );
 int i;
 
 // Check for consistency and set reference point of the product.
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( abs( xPtr->refPoint[i] - yPtr->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << "\n*** JLC MULTIPLICATION ERROR: Inconsistent reference points." << endl;
     exit(0);
     }
   zPtr->refPoint[i] = xPtr->refPoint[i];
   }
 
 // If one of the arguments is void, return it ..
 if ( xPtr->count < 1 ) return x;
 if ( yPtr->count < 1 ) return y;

 // .. otherwise continue normal operations.
 while( p = (JLCterm*) gy() ) {
 while( q = (JLCterm*) gx() ) {
 
   if( ( p->weight + q->weight ) > JetC::oneEnv.MaxWeight ) continue;
                                   // ??? Change this: compute zPtr->accuWgt earlier 
   r = new JLCterm;                 // ??? and use it.
   *r = (*p)*(*q);
   zPtr->addTerm( r );
   }}
 
 // Determine the maximum weight computed accurately.
 if( xPtr->accuWgt < yPtr->accuWgt ) zPtr->accuWgt = xPtr->accuWgt;
 else                                zPtr->accuWgt = yPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const JetC& x, const complex y ) {  
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == complex_0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];  // Since zPtr is no longer
   z.stacked = 1;                        // correct, use z.
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const JetC& x, const double y ) {  
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == 0.0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];  // Since zPtr is no longer
   z.stacked = 1;                        // correct, use z.
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const JetC& x, const int j ) { 
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 if ( xPtr->count < 1 ) return x;
 if( ( y = (double) j ) == 0.0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const complex y, const JetC& x ) {  
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == complex_0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const double y, const JetC& x ) {  
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 if ( xPtr->count < 1 ) return x;
 if( y == 0.0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator*( const int j, const JetC& x ) { 
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 if ( xPtr->count < 1 ) return x;
 if( ( y = (double) j ) == 0.0 ) {
   z = complex_0;
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
    z->refPoint[i] = xPtr->refPoint[i];
   z.stacked = 1;
   return z;
   }
 
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value *= y;
   zPtr->addTerm( q );
   }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( const JetC& x, const complex& y ) { 
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 // Check for void operators ..
 if ( xPtr->count < 1 ) return x;

 // Check for division by zero ..
 if( y == complex_0 ) {
   cout << ::form("\n*** JLC ERROR *** ");
   cout << ::form("Attempt to divide by a scalar zero.\n") << endl;
   exit(0);
 }
 
 // .. and continue.
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value /= y;
   zPtr->addTerm( q );
 }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 z.stacked = 1;
 return z;
}
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( const JetC& x, const double& y ) { 
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 
 // Check for void operators ..
 if ( xPtr->count < 1 ) return x;

 // Check for division by zero ..
 if( y == 0.0 ) {
   cout << ::form("\n*** JLC ERROR *** ");
   cout << ::form("Attempt to divide by a scalar zero.\n") << endl;
   exit(0);
 }
 
 // .. and continue.
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value /= y;
   zPtr->addTerm( q );
 }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( const JetC& x, const int& j ) { 
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLC* xPtr = x.jl;
 JLC* zPtr = z.jl;
 dlist_iterator gx( *(dlist*) xPtr );
 int i;
 double y;
 
 // Check for void operators ..
 if ( xPtr->count < 1 ) return x;

 // Check for division by zero ..
 if( ( y = (double) j ) == 0.0 ) {
   cout << ::form("\n*** JLC ERROR *** ");
   cout << ::form("Attempt to divide by a scalar zero.\n") << endl;
   exit(0);
 }
 
 // .. and continue.
 while( p = (JLCterm*) gx() ) {
   q = new JLCterm( p );
   q->value /= y;
   zPtr->addTerm( q );
 }
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
  zPtr->refPoint[i] = xPtr->refPoint[i];
 
 zPtr->accuWgt = xPtr->accuWgt;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( complex a, const JetC& b ) {
  return JetC(a)/b;
} 

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( double a, const JetC& b ) {
  return JetC(complex(a,0.0))/b;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator|( const JetC& u, const JetC& v ) { 
 /*
 Truncated multiplication: used only by the division operator.
 No checks are made on the reference point.
 */
 JetC z;
 JLCterm* p;
 JLCterm* q;
 JLCterm* r;
 dlist_looper gu( *(dlist*) u.jl );
 dlist_looper gv( *(dlist*) v.jl );
 
 // If one of the arguments is void, return it ..
 if ( u->count < 1 ) return u;
 if ( v->count < 1 ) return v;

 // .. otherwise continue normal operations.
 while( p = (JLCterm*) gv() ) {
 while( q = (JLCterm*) gu() ) {
   if( ( p->weight + q->weight ) > JetC::oneEnv.weightLimit ) continue;
   r = new JLCterm;
   *r = (*p)*(*q);
   z.addTerm( r );
 }}
 
 // Determine the maximum weight computed accurately.
 if( u->accuWgt < v->accuWgt ) z->accuWgt = u->accuWgt;
 else                          z->accuWgt = v->accuWgt;
 
 z.stacked = 1;   // ??? Should the reference point be copied???
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator/( const JetC& wArg, const JetC& uArg ) { 
 dlist_looper gu( *(dlist*) uArg.jl );
 dlist_looper gw( *(dlist*) wArg.jl );
 JetC v;
 JetC vn;
 JetC w, u;
 JLCterm* qu;
 JLCterm* qw;
 complex u0, w0;
 int wgt;
 int i;
 
 // Check for void operators ..
 if ( wArg->count < 1 ) return wArg;
 if ( uArg->count < 1 ) {
   cout << ::form("\n*** JLC ERROR *** ");
   cout << ::form("Attempt to divide by a void JLC variable.\n") << endl;
   exit(0);
   }
 
 // Check for consistency and set reference point of the quotient.
 static char divCheck;
 divCheck = 0;
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
   if( abs( wArg->refPoint[i] - uArg->refPoint[i] ) > MX_SMALL*Jet::oneEnv.scale[i] ) {
     divCheck = 1;
     cout << "\n*** JLC DIVISION ERROR: Inconsistent reference points.\n" 
          << "wArg->refPoint[ " << i << " ] = " << wArg->refPoint[i] << "\n"
          << "uArg->refPoint[ " << i << " ] = " << uArg->refPoint[i] << "\n"
          << endl;
   }
   v->refPoint[i] = wArg->refPoint[i];
 }
 if( divCheck ) exit(0);
 
 // Normalize the denominator
 if( ( qu = (JLCterm*) gu() ) == 0 ) {
   cout << ::form( "\n*** JLC ERROR ***  Division algorithm called with uninitialized JLC.\n" )
        << endl;
   exit(0);
   }
 if( ( wgt = qu->weight ) != 0 ) {
   cout << ::form( "\n*** JLC ERROR ***  Attempt to divide by nilpotent element.\n" )
        << endl;
   exit(0);
   }
 if( ( u0 = qu->value ) == complex_0 ) {
   cout << ::form( "\n*** JLC ERROR ***  Attempt to divide by zero.\n" )
        << endl;
   exit(0);
   }
 u = uArg / u0;
 
 // Recursive algorithm
 qw = (JLCterm*) gw();
 v = w0 = ( qw->value );
 
 JetC::oneEnv.weightLimit = 0;
 while ( JetC::oneEnv.weightLimit < JetC::oneEnv.MaxWeight ) {
   JetC::oneEnv.weightLimit++;
   vn = u|v;
   vn = vn.filter( JetC::oneEnv.weightLimit, JetC::oneEnv.weightLimit );
   w = wArg.filter( JetC::oneEnv.weightLimit, JetC::oneEnv.weightLimit );
   v += ( w - vn );
   }
 
 // Correct normalization of the answer ..
 v /= u0;
 
 // .. set the reference point ..
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   v->refPoint[i] = wArg->refPoint[i];
 
 // Determine the maximum weight computed accurately.
 if( wArg->accuWgt < uArg->accuWgt ) v->accuWgt = wArg->accuWgt;
 else                                v->accuWgt = uArg->accuWgt;
 
 // .. and return value.
 v.stacked = 1;
 return v;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC operator^( const JetC& x, const JetC& y ) { 
 JetC z;
 int i;
 int* m = new int [ JetC::oneEnv.NumVar ];
 int* n = new int [ JetC::oneEnv.NumVar ];
 
 if( JetC::oneEnv.PBOK ) {
 
   for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
     m[i] = 0;
     n[i] = 0;
     }
 
   for( i = 0; i < JetC::oneEnv.dof; i++ ) {
     m[ i ] = 1;
     n[ i + JetC::oneEnv.dof ] = 1;
 
     z += ( ( x.D(m) * y.D(n) ) - ( x.D(n) * y.D(m) ) );
 
     m[ i ] = 0;
     n[ i + JetC::oneEnv.dof ] = 0;
     }
 
   delete [] m;
   delete [] n;
   z.stacked = 1;
   return z;
   }
 
 cout << ::form( "\n*** ERROR ***                                \n" );
 cout << ::form(   "*** ERROR *** JLC operator^( JLC&, JLC& )    \n" );
 cout << ::form(   "*** ERROR *** Wrong JLCSetup used; cannot    \n" );
 cout << ::form(   "*** ERROR *** perform bracket.               \n" );
 cout << ::form(   "*** ERROR ***                                \n" );
 cout << endl;
 
 if( 2 + 3 == 5 ) exit(1);       // Done to avoid compiler warnings
 delete [] m;                    // about leaving the function.
 delete [] n;
 z.stacked = 1;
 return z;                       
}


//***************************************************************
//***************************************************************
//***************************************************************
//
//      Overloaded analytic functions

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC acos( const JetC&  ) {   // ??? Write this damned thing!!!  It's trivial!!
 JetC z;
 cout << ::form( "JetC acos( JetC& x ) to be written.\n" ) << endl;
 exit(0);
 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC asin( const JetC& x ) { 
 JetC z;
 JetC dz;
 z = x;
 dz = ( sin(z) - x ) / cos(z);
 int iter = 0;
 while( ( dz != complex_0 ) && ( iter++ < MX_MAXITER ) ) {
  z -= dz;
  dz = ( sin(z) - x ) / cos(z);
 }
 if( iter >= MX_MAXITER ) {
  cout << ::form( "*** WARNING ***                             \n" );
  cout << ::form( "*** WARNING *** JLC& asin( JLC& x ) {       \n" );
  cout << ::form( "*** WARNING *** Over %d iterations used;    \n", MX_MAXITER );
  cout << ::form( "*** WARNING *** result may be incorrect.    \n" );
  cout << ::form( "*** WARNING ***                             \n" );
  cout << endl;
 }
 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC atan( const JetC& x ) { 
 JetC z;
 JetC c, dz;
 z = x;
 c = cos( z );
 dz = c*( sin(z) - x*c );
 int iter = 0;
 while( ( dz != complex_0 ) && ( iter++ < MX_MAXITER ) ) {
  z -= dz;
  c = cos( z );
  dz = c*( sin(z) - x*c );
 }
 if( iter >= MX_MAXITER ) {
  cout << ::form( "*** WARNING ***                             \n" );
  cout << ::form( "*** WARNING *** JLC& atan( JLC& x ) {       \n" );
  cout << ::form( "*** WARNING *** Over %d iterations used;    \n", MX_MAXITER );
  cout << ::form( "*** WARNING *** result may be incorrect.    \n" );
  cout << ::form( "*** WARNING ***                             \n" );
  cout << endl;
 }

 z.stacked = 1;
 return z;
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC cos( const JetC& x ) { 
 
 JetC epsSin( const JetC& );
 JetC epsCos( const JetC& );
 JetC epsilon;
 complex cs, sn;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  exp called with unevaluated JLC.\n")
        << endl;
   return x;
   }
 
 p = (JLCterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   cs = cos( p -> value );
   sn = sin( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return cs*epsCos( epsilon ) - sn*epsSin( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsCos( x );
   }
 
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC cosh( const JetC& x ) { 
 JetC z;
 z = exp(x);
 z = ( z +  (complex_one / z ) ) / complex(2.0,0.0);
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC exp( const JetC& x ) { 
 
 JetC epsExp( const JetC& );
 JetC epsilon;
 complex factor;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  exp called with unevaluated JLC.\n")
        << endl;
   return x;
   }
 
 p = (JLCterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   factor = exp( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return factor*epsExp( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsExp( x );
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC log ( const JetC& x ) { 
 JetC             epsilon;
 JetC             u, w;
 complex          std;
 double           n;
 JLCterm*         p;
 dlist_iterator  getNext( *(dlist*) x.jl );
 
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  log called with unevaluated JLC.\n")
        << endl;
   return x;
   }
 
 p = (JLCterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) != complex_0 )
   )                                  // x has non-zero standard part
   {
   if( x->count == 1 ) {             // x may have no derivatives
     epsilon = log( std );
     return epsilon;
     }
   epsilon = x;                      // x must not be altered by the routine
   p = epsilon.get();                //   pops the standard part off epsilon
   delete p;
   epsilon.scaleBy( - complex_one/std );
   w = epsilon;
   u = epsilon;                      // We use here the formulae
   n = 1.0;                          // ln( s + e )   = ln s + ln( 1 + e/s ), and
   while( u->count > 0 ) {           // ln( 1 + e/s ) = ln( 1 - ( -e/s )) 
     u *= epsilon;                   //               = - sum ( - e/s )^n / n
     w += ( u / ++n );
   }
 
   return ( log(std) - w );
 // -------------------
 
   }
 else                                 // x has zero standard part
   {
   cout << ::form( "\n\n*** ERROR *** \n"                                  );
   cout << ::form(     "*** ERROR *** JLC log called with argument whose\n" );
   cout << ::form(     "*** ERROR *** standard part vanishes.\n"           );
   cout << ::form(     "*** ERROR *** \n\n"                                );
   cout << endl;
   if( 1 + 1 == 2 ) exit(1);
   return x;
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC log10( const JetC& x ) {
 static const double logE = 0.4342944819032518276511289;
 return  logE*log(x);
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC pow( const JetC& x, int n ) {
  JetC z;
  static int i;
  static complex complex_one( 1.0, 0.0 );
  if( n == 0 ) z = complex( 1.0, 0.0 );
  else if( n > 0 ) {
    z = x;
    for( i = 2; i <= n; i++ ) z *= x;
  }
  else {
    JetC xr;
    z = complex_one / x;
    xr = z;
    for( i = -2; i >= n; i-- ) z *= xr;
  }
  z.stacked = 1;
  return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC pow( const JetC& x, const double s ) { 
 JetC epsPow( const JetC&, const double );
 JetC epsilon;
 complex factor;
 complex std;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 int u;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  pow called with unevaluated JLC.\n")
        << endl;
   return x;
 }
 
 p = (JLCterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) !=  complex_0 )
   )                                  // x has non-zero standard part
   {
   if( x->count == 1 ) {             // x may have no derivatives
     epsilon = pow( std, s );
     return epsilon;
     }
   factor = pow( std, s );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 //   pops the standard part off epsilon
   delete p;
   epsilon.scaleBy( complex_one/std );
   epsilon = factor*epsPow( epsilon, s );
   epsilon.stacked = 1;
   return epsilon;
   }
 else                                 // x is pure infinitesimal
   {
   u = nearestInteger( s );
   if( s != (double) u ) {
     cout << ::form( "\n*** ERROR::JLC pow: attempt to use infinitesimal " );
     cout << ::form( "as base with non-integer exponent.\n" );
     cout << endl;
     exit(1);
     }
   epsilon = complex_one;
   if ( u == 0 ) {
     epsilon.stacked = 1;
     return epsilon;
   }
   if ( u > 0 ) {
     while( u-- > 0 ) epsilon *= x;
     epsilon.stacked = 1;
     return epsilon;
     }
   else {
     while( u++ < 0 ) epsilon *= x;
     epsilon = complex_one / epsilon;
     epsilon.stacked = 1;
     return epsilon;
     }
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC sin( const JetC& x ) { 
 
 JetC epsSin( const JetC& );
 JetC epsCos( const JetC& );
 JetC epsilon;
 complex cs, sn;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  exp called with unevaluated JLC.\n")
        << endl;
   return x;
   }
 
 p = (JLCterm*) getNext();
 if( ( p -> weight ) == 0 ) {         // x has non-zero standard part
   cs = cos( p -> value );
   sn = sin( p -> value );
   epsilon = x;                       // x must not be altered by the routine
   p = epsilon.get();                 // pops the standard part off epsilon
   delete p;
   return sn*epsCos( epsilon ) + cs*epsSin( epsilon );
   }
 else {                               // x is pure infinitesimal
   return epsSin( x );
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC sinh( const JetC& x ) {
 JetC z;
 z = exp(x);
 return ( z - (complex_one/z))/2.0;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC sqrt( const JetC& x ) {
 JetC epsSqrt( const JetC& );
 JetC epsilon;
 complex factor;
 complex std;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 
 if( x->count == 0 ) {
   cout << ::form("\n*** WARNING ***  sqrt called with unevaluated JLC.\n")
        << endl;
   return x;
   }
 
 p = (JLCterm*) getNext();
 if( ( ( p -> weight      ) ==  0   ) &&
     ( ( std = p -> value ) !=  complex_0 )
   )                                  // x has non-zero standard part
   {
   factor = sqrt( std );
   epsilon = x;                       // x must not be altered by the routine
   if( epsilon->count == 1 )          // x may have no derivatives
     {
     epsilon = factor;
     return epsilon;
     }
   else                               // normal case
     {
     p = epsilon.get();               // pops the standard part off epsilon
     delete p;
     epsilon.scaleBy( complex_one/std );
     return factor*epsSqrt( epsilon );
     }
   }
 else                                 // x is pure infinitesimal
   {
   cout << ::form( "\n*** ERROR::JLC sqrt: attempt to use infinitesimal argument." )
        << endl;
   if( 1 + 1 == 2 ) exit(1);
   return x;
   }
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC tan( JetC& x ) { 
 return sin(x) / cos(x) ;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC tanh( const JetC& x ) { 
 return sinh(x) / cosh(x);
}
 
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC erf( const JetC& z ) 
{
  // Reset default reference point according to input
  complex* rfp = new complex [ JetC::oneEnv.NumVar ];
  JetC dummyJet;
  dummyJet.getReference( rfp );
  JetC::FixReference( z );


  if( ( fabs(imag(z.standardPart())) > 3.9 ) || 
      ( fabs(real(z.standardPart())) > 3.0 ) ) {
    JetC u;
    u = complex( 0., 1. )*z;
    u = 1.0 - exp(u*u)*w(u);

    // Reset default reference point to original value
    JetC::FixReference( rfp );
    delete [] rfp;

    return u;
  }

         JetC    series;
         JetC    oldseries;
         JetC    arg;
         JetC    term;
  static double  den;
  static double  fctr_x;
  static int     counter;

  series        = 1.0;
  oldseries     = 0.0;
  arg           = - z*z;
  den           = 1.0;
  term          = 1.0;
  fctr_x        = 0.0;

  counter = 0;
  while( ( series != oldseries ) || counter++ < JetC::oneEnv.MaxWeight ) {
    oldseries = series;
    den      += 2.0;
    fctr_x   += 1.0;
    term     *= arg/fctr_x;
    series   += term/den;
    /* DGN cout << "DGN> " << series << endl; */ 
  }  

  // Reset default reference point to original value
  JetC::FixReference( rfp );
  delete [] rfp;

  return (2.0/MATH_SQRTPI)*z*series;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC w( const JetC& z ) 
{
  // Reset default reference point according to input
  complex* rfp = new complex [ JetC::oneEnv.NumVar ];
  JetC dummyJet;
  dummyJet.getReference( rfp );
  JetC::FixReference( z );

  static const complex mi( 0., -1. );
  static double x;
  static double y;
         JetC   answer;
  
  x = real( z.standardPart() );
  y = imag( z.standardPart() );

  if( ( x < 0.0 ) || ( y < 0.0  ) ) {
    cout << "*** ERROR ***                                     \n"
            "*** ERROR *** JetC w( const JetC& )               \n"
            "*** ERROR *** Argument must have positive         \n"
            "*** ERROR *** standard part.                      \n"
            "*** ERROR ***                                     \n"
         << endl;
    exit(1);
  }

  if( ( x > 6.0 ) || ( y > 6.0 ) ) 
    answer = ( - mi * z * (
                          ( 0.5124242  /( z*z - 0.2752551 )) + 
                          ( 0.05176536 /( z*z - 2.724745  ))
                          ) 
             );

  else if( ( x > 3.9 ) || ( y > 3.0 ) ) 
    answer = ( - mi * z * (
                          ( 0.4613135   /( z*z - 0.1901635 )) + 
                          ( 0.09999216  /( z*z - 1.7844927 )) + 
                          ( 0.002883894 /( z*z - 5.5253437 ))
                          ) 
             );

  else answer = exp( -z*z )*( 1.0 - erf( mi*z ) );

  // Reset default reference point to original value
  JetC::FixReference( rfp );
  delete [] rfp;

  return answer;
}



//***************************************************************
//***************************************************************
//***************************************************************
//
//      Auxiliary functions which operate on infinitesimals.
//      It is assumed that the calling program checks to be sure
//      that the argument has zero standard part.
 
JetC epsCos( const JetC& epsilon ) { 
 
 JetC z( complex_one );
 JetC epsq, term;
 double n;
 int i;
 
 epsq = - epsilon*epsilon;
 
 term = epsq / 2.0;
 n = 2.0;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( ( term*epsq ) / ++n ) / ++n;
   }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC epsExp( const JetC& epsilon ) { 
 JetC z( complex_one );
 JetC term;
 double n;
 int i;
 
 term = epsilon;
 n = 1.0;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) 
   z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( term*epsilon ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC epsPow( const JetC& epsilon, const double s ) { 
 JetC z( complex_one );
 JetC term;
 double f, n;
 int i;
 
 f = s;
 n = 1.0;
 term = s*epsilon;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
          z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term = ( (--f) * term * epsilon ) / ++n;
   }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC epsSin( const JetC& epsilon ) { 
 
 JetC z;
 JetC epsq, term;
 double n;
 
 epsq = - epsilon*epsilon;
 z = epsilon;
 term = ( epsilon*epsq ) / 6.0;
 n = 3.0;
 
 while( term->count > 0 ) {
   z += term;
   term = ( ( term*epsq ) / ++n ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC epsSqrt( const JetC& epsilon ) {  // This function is identical to epsPow
                                     // with the substitution  s = 1/2
 JetC     z( complex_one );
 JetC     term;
 double  f, n;
 int     i;
 
 f    = 1.0 / 2.0;
 n    = 1.0;
 term = f*epsilon;
 
 for( i = 0; i < JetC::oneEnv.NumVar; i++ )
          z->refPoint[i] = epsilon->refPoint[i];
 
 while( term->count > 0 ) {
   z += term;
   term *= ( (--f) * epsilon ) / ++n;
 }
 
 z->accuWgt = epsilon->accuWgt;

 z.stacked = 1;
 return z;
}
 

//***************************************************************
//***************************************************************
//***************************************************************
//
//      Implementation of Class JetC

void JetC::fixReference( const JetC& x ) {
 jl->fixReference( *(x.jl) );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

char JetC::isNilpotent() {
 return jl->isNilpotent();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::peekAt() {
 jl->peekAt();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::printCoeffs() {
 jl->printCoeffs();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

dlist JetC::image() {      // ??? Problem: nothing is recorded about the
                          // ??? reference point.
 return jl->image();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::writeToFile( char* fileName ) {
 jl->writeToFile( fileName );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::writeToFile( FILE* filePtr ) {
 jl->writeToFile( filePtr );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::scaleBy( complex y ) {
 jl->scaleBy( y );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JLCterm* JetC::get() {
 return jl->get();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JetC::standardPart() const {
 return jl->standardPart();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

void JetC::clear() {
 jl->clear();
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JetC::weightedDerivative( int* ind ) {
 return jl->weightedDerivative( ind );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JetC::derivative( int* ind ) {
 return jl->derivative( ind );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::filter( const int wgtLo, const int wgtHi ) const { 
 JetC z;
 JLC* zPtr = z.jl;
 JLC* thisPtr = jl;
 dlist_iterator getNext( *(dlist*) thisPtr );
 JLCterm* p;
 JLCterm* q;
 int numTerms;
 int wgt, upperWgt;
 
 if( ( wgtLo <= 0 ) && 
     ( wgtHi >= thisPtr->weight )   
   ) {
   z = *this;
   z.stacked = 1;
   return z;
 } 

 numTerms = 0;
 upperWgt = 0;
 
 while( p = (JLCterm*) getNext() ) {
   wgt = p->weight;
   if( ( wgt >= wgtLo ) && ( wgt <= wgtHi ) ) {
     q = new JLCterm( p );
     zPtr->append( q );
     if( wgt > upperWgt ) upperWgt = wgt;
     numTerms++;
     }
   }
 
 zPtr->count = numTerms;   // ??? Is this necessary?
 zPtr->weight = upperWgt;  // ??? Where is the copy of the re
 zPtr->accuWgt = thisPtr->accuWgt;   // ??? reference point???

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::filter( char (*f) ( const int*, const complex ) ) { 
 JetC z;
 JLC* zPtr =  z.jl;
 JLC* thisPtr = jl;
 dlist_iterator getNext( *(dlist*) thisPtr );
 JLCterm* p;
 JLCterm* q;

 int i;
 int nv = oneEnv.NumVar;
 int* oldIndex = new int[ nv ];

 while( p = (JLCterm*) getNext() ) {
  for( i = 0; i < nv; i++ ) oldIndex[i] = p->index[i];

  if( (*f)( p->index, p->value ) ) {
   q = new JLCterm( p );
   zPtr->append( q );
  }

  for( i = 0; i < nv; i++ ) 
   if( oldIndex[i] != p->index[i] ) {
    cout << ::form( "\n" );
    cout << ::form( "*** ERROR ***                                      \n" );
    cout << ::form( "*** ERROR *** JLC& JLC::filter( char (*f) ( int*, double ) \n" );
    cout << ::form( "*** ERROR *** The test routine is not allowed      \n" );
    cout << ::form( "*** ERROR *** to change the values in the index    \n" );
    cout << ::form( "*** ERROR *** array!!                              \n" );
    cout << ::form( "*** ERROR ***                                      \n" );
    cout << endl;
    exit(0);
   }
 }

 delete [] oldIndex;
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC concat( const JetC& x ) {
 JetC v;
 dlist_iterator getNext( *(dlist*) x.jl );
 JLCterm* p;
 int i;
 
 v = complex_0;
 v.fixReference( JetC::oneEnv.JLmonomial[0] );
 
 while( p = (JLCterm*) getNext() ) {
   if( p->weight > x->accuWgt ) break;
	for( i = 0; i < JetC::oneEnv.NumVar; i++ )
     Jet::oneEnv.exponent[i] = (p->index)[i];
   Jet::monoCode();
   v += ( p->value ) * ( JetC::oneEnv.JLmonomial[ Jet::monoRank() ] );
 }
 
 v.stacked = 1;
 return v;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::operator()( JetC* y ) const { 
 JetC*     u = new JetC [ oneEnv.NumVar ];
 JetC      z, term;
 int      i, j, w;
 JLCterm*  p;

 // Check consistency of reference points.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   for( j = 0; j < oneEnv.NumVar; j++ )
     if( abs( y[i]->refPoint[j] - y[0]->refPoint[j] ) > MX_SMALL*Jet::oneEnv.scale[j] ) {
       cout << ::form( "\n*** JetC JetC::operator( JetC* ) ERROR: Inconsistent reference points.\n" );
       cout << ::form("y[%d].refPoint[%d] != y[0].refPoint[%d]\n", i, j, j);
       cout << ::form("%lf != %lf\n", y[i]->refPoint[j], y[0]->refPoint[j]);
       cout << endl;
       exit(1);
     }
   if( abs( (*this)-> refPoint[i] - y[i].weightedDerivative( oneEnv.AllZeroes ) ) 
       > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << ::form( "\n*** JetC JetC::operator( JetC* ) ERROR: Inconsistent reference points.\n" );
     cout << ::form("(*this)-> refPoint[%d] != y[%d].weightedDerivative( oneEnv.AllZeroes )\n",
             i, i);
     cout << ::form("%lf != %lf\n",
             (*this)-> refPoint[i], y[i].weightedDerivative( oneEnv.AllZeroes ));
     cout << endl;
     exit(1);
   }
 }
 
 // Subtract reference point prior to concatenation.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   u[i] = y[i];
   p    = ( u[i] ).get();
   if( p->weight != 0 ) {
     cout << ::form( "\n*** JLC JLC::operator() ERROR: Argument has no nonstandard part!!\n" )
          << endl;
     exit(0);
   }
   delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 oneEnv.JLmonomial[0] = complex_one;
 oneEnv.JLmonomial[0].fixReference( y[0] );
 
 // For all higher weights ...
 for( w = 1; w <= (*this)->accuWgt; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = oneEnv.JLmonomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     oneEnv.JLmonomial[ Jet::monoRank() ] = term * u[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the JLC variable and evaluate.
 
 z = concat( *this );  // ??? START HERE, LEO!!!
 
 // Set the correct reference point in z.
 for( i = 0; i < oneEnv.NumVar; i++ ) 
   z->refPoint[i] = y[0]->refPoint[i];
 
 delete [] u;

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::operator()( CLieOperator& y ) const { 
 JetC z;
 CLieOperator u;
 JetC term;
 int i;
 int w;
 dlist_iterator getNext( *(dlist*) jl );
 JLCterm* p;
 
 u = y;
 u.reset();

 // Check consistency of reference points.
 for( i = 0; i < oneEnv.NumVar; i++ )
   if( abs( jl->refPoint[i] - ( y.v[i] ).weightedDerivative( oneEnv.AllZeroes ) ) 
       > MX_SMALL*Jet::oneEnv.scale[i] ) {
     cout << ::form( "\n*** JetC JetC::operator( CLieOperator& ) ERROR: Inconsistent reference points.\n" );
     cout << ::form( "jl->.refPoint[%d] != ( y.v[%d] ).weightedDerivative( oneEnv.AllZeroes )\n", i, i);
     cout << ::form( "%lf != %lf\n", 
                    jl->refPoint[i], 
                    ( y.v[i] ).weightedDerivative( oneEnv.AllZeroes ) 
                 );
     cout << endl;
     exit(0);
     }
 
 // In the event that the argument y is null
 if( u.weight <= 0 ) {
  if( ( p = (JLCterm*) getNext() ) == 0 ) {   // ??? Should this be an error
   z.stacked = 1;                            // ??? condition instead?
   return z;                                 
  }
  if( p->weight == 0 ) z = complex_0;
  else                 z = p->value;
  z.stacked = 1;
  return z;
 }

 // Subtract reference point prior to concatenation.
 for( i = 0; i < oneEnv.NumVar; i++ ) {
  p = ( u.v[i] ).get();
  if( p->weight != 0 ) {
    cout << ::form( "\n*** JLC JLC::operator() ERROR: Argument has no nonstandard part!!\n" )
         << endl;
    exit(0);
  }
  delete p;
 }
 
 // Evaluate and store monomials.
 
 // The zeroth one.
 oneEnv.JLmonomial[0] = complex_one;
 oneEnv.JLmonomial[0].fixReferenceAtStart( y );
 
 // For all higher weights ...
 for( w = 1; w <= jl->accuWgt; w++ )
 
   // Get the next set of exponents of weight w.
   while( nexcom( w, oneEnv.NumVar, Jet::oneEnv.exponent ) ) {
 
     // Find the first non-zero exponent.
     i = 0;
     while( !( Jet::oneEnv.exponent[i++] ) ) ;
     i--;
 
     // The value of the monomial associated with this composition
     // is obtained by multiplying a factor into a previously
     // computed monomial.
     ( Jet::oneEnv.exponent[i] )--;
     Jet::monoCode();
     term = oneEnv.JLmonomial[ Jet::monoRank() ];
     ( Jet::oneEnv.exponent[i] )++;
     Jet::monoCode();
     oneEnv.JLmonomial[ Jet::monoRank() ] = term * u.v[i];   // ??? Is this OK???
 
     } // End while loop.
 
 // Monomials have been stored.
 // Now traverse the JLC variable and evaluate.
 
 z = concat( *this );
 
 // Set the correct reference point in z.
 for( i = 0; i < oneEnv.NumVar; i++ ) z->refPoint[i] = y.refPoint[i];
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

complex JetC::operator()( complex* x ) const {
 return jl->operator()( x );
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet real( const JetC& x ) {
 int       i;
 Jet       z;
 JLCterm*  p;
 JLterm*   q;
 JLC*      xPtr;
 JL*       zPtr;

 xPtr = x.jl;
 zPtr = z.jl;

 dlist_iterator gx( *(dlist*) xPtr );
 
 // Check for consistency and set reference point of the sum.
 if( Jet::oneEnv.NumVar != JetC::oneEnv.NumVar ) {
  cerr << "\n*** ERROR ***                                      \n"
          "*** ERROR *** real( JetC& )                        \n"
          "*** ERROR *** Environment variables not consistent!\n"
          "*** ERROR ***                                      \n"
       << endl;
  exit(0);
 }

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
  if( imag( xPtr->refPoint[i] ) != 0.0 ) {
   cerr << "\n*** ERROR ***                                     \n"
           "*** ERROR *** real( JetC& )                       \n"
           "*** ERROR *** Bad reference point.                \n"
           "*** ERROR ***                                     \n"
        << endl;
   exit(0);
  }
  zPtr->refPoint[i] = real( xPtr->refPoint[i] );
 }

 // If the argument is void, then return ...
 if( xPtr->count < 1 ) {
  z = 0.0;
  z.stacked = 1;
  return z;
 }

 // .. otherwise, continue normal operations.
 while( p = (JLCterm*) gx() ) {
  q = new JLterm( p->index, real( p->value ) );
  zPtr->addTerm( q );
 }

 // Set the maximum accurate weight.
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

Jet imag( const JetC& x ) {
 int       i;
 Jet       z;
 JLCterm*  p;
 JLterm*   q;
 JLC*      xPtr;
 JL*       zPtr;

 xPtr = x.jl;
 zPtr = z.jl;

 dlist_iterator gx( *(dlist*) xPtr );
 
 // Check for consistency and set reference point of the sum.
 if( Jet::oneEnv.NumVar != JetC::oneEnv.NumVar ) {
  cerr << "\n*** ERROR ***                                      \n"
          "*** ERROR *** imag( JetC& )                        \n"
          "*** ERROR *** Environment variables not consistent!\n"
          "*** ERROR ***                                      \n"
       << endl;
  exit(0);
 }

 for( i = 0; i < JetC::oneEnv.NumVar; i++ ) {
  if( imag( xPtr->refPoint[i] ) != 0.0 ) {
   cerr << "\n*** ERROR ***                                     \n"
           "*** ERROR *** imag( JetC& )                       \n"
           "*** ERROR *** Bad reference point.                \n"
           "*** ERROR ***                                     \n"
        << endl;
   exit(0);
  }
  zPtr->refPoint[i] = real( xPtr->refPoint[i] );
 }

 // If the argument is void, then return ...
 if( xPtr->count < 1 ) {
  z = 0.0;
  z.stacked = 1;
  return z;
 }

 // .. otherwise, continue normal operations.
 while( p = (JLCterm*) gx() ) {
  q = new JLterm( p->index, imag( p->value ) );
  zPtr->addTerm( q );
 }

 // Set the maximum accurate weight.
 zPtr->accuWgt = xPtr->accuWgt;
 
 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::D( int* n ) const { 
 char noTermAdded;
 int f, i, j, k, w;
 JetC z;
 dlist_iterator getNext( *(dlist*) jl );
 JLCterm* p;
 JLCterm* q;
 
 noTermAdded = 1;
 
 // --- Preliminary check of index set.
 
 w = 0;
 
 for( i = 0; i < oneEnv.NumVar; i++ ) {
   if( n[i] < 0 ) {
     cout << ::form( "\n*** JLC ERROR: Cannot differentiate with negative index." );
     cout << ::form( "\n*******************************************************\n" );
     cout << endl;
     exit(0);
     }
   w += n[i];
 }
 
 if( w > (*this)->accuWgt ) {
   cout << ::form( "\n*** JLC ERROR: Differentiation request beyond accuracy allowed." );
   cout << ::form( "\n**************************************************************\n" );
   cout << endl;
   exit(0);
 }
 
 if( w == 0 ) {
   z = *this;
   z.stacked = 1;
   return z;
 }

 // --- Construct the derivative one link at a time.
 // --- ( See note Obs.4 )
 
 while( p = (JLCterm*) getNext() ) {
 
   q = new JLCterm( p );
 
   char doit = 1;
   // -- Reset the index.
   for( i = 0; i < oneEnv.NumVar; i++ ) 
     doit = doit && ( ( q->index[i] -= n[i] ) >= 0 );
 
   if( doit ) {
     // -- Build factorial multiplier.
     f = 1;
     for( k = 0; k < oneEnv.NumVar; k++ ) {
       j = q->index[k];
       for( i = 0; i < n[k]; i++ ) f *= ++j;
       }
     if( f <= 0 ) {
       cout << ::form( "\n*** JLC ERROR: Horrible, unexplainable error while differentiating!" );
       cout << ::form( "\n******************************************************************\n" );
       cout << endl;
       exit(0);
       }                           // Super fussbudget!! ( f "must" be positive )
   
     // -- Make final changes in private data of the JLCterm and
     //    absorb it into the answer.
     (q->value) *= f;              // OK. So I'm a fussbudget with parentheses again ..
     (q->weight ) -= w;
   
     noTermAdded = 0;
     z->addTerm( q );               // ??? This also may be unnecessarily conservative.
                                    // The append function may be adequate, if the
                                    // ordering is preserved. ???
     } 
   }
 
 if( noTermAdded ) z = complex_0;
 
 
 // --- Finally, adjust accuWgt and return value
 
 z->accuWgt = (*this)->accuWgt - w;       // ??? Is this correct ???

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::inverse() {
 if( oneEnv.SpaceDim != 1 ) {
  cout << ::form( "\n" );
  cout << ::form( "*** ERROR ***                                     \n" );
  cout << ::form( "*** ERROR ***  JLC& JLC::inverse()                  \n" );
  cout << ::form( "*** ERROR ***  This method can be called only     \n" );
  cout << ::form( "*** ERROR ***  if there is one and only one       \n" );
  cout << ::form( "*** ERROR ***  coordinate.  You have %d.          \n", oneEnv.SpaceDim );
  cout << ::form( "*** ERROR ***                                     \n" );
  cout << endl;
  exit(0);
 }

 JetC z;

 JetC x, v;
 JetC g;
 int d = 1;

 x.setVariable(complex_0, 0);
 g = D(&d);
 
 int i = 0;
 z = x;
 v = ( operator()(&z) - x ) / g(&z);
 while( ( i++ < MX_MAXITER ) &&  ( v !=  complex_0 ) ) {
  z -= v;
  v = ( operator()(&z) - x ) / g(&z);
 }

 z.stacked = 1;
 return z;
}


// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::Inverse() const {
 static JetC           z;
 static JLCterm*       p;
 static char           t1, t2;

 if( oneEnv.SpaceDim != 1 ) {
  cout << "\n" 
          "*** ERROR ***                                     \n" 
          "*** ERROR ***  JetC JetC::Inverse()               \n" 
          "*** ERROR ***  This method can be called only     \n" 
          "*** ERROR ***  if there is one and only one       \n" 
          "*** ERROR ***  coordinate.  You have " << oneEnv.SpaceDim << "\n"
          "*** ERROR ***                                     \n" 
       << endl;
  exit(1);
 }

 if( ( t1 = ( jl->refPoint[0]    == complex_0 ) ) && 
     ( t2 = ( jl->standardPart() == complex_0 ) ) 
   ) return EpsInverse();

 z = *this;   // Deep copy needed here.

 if( !t2 ) { 
  p = z.get();
  z->addTerm( new JLCterm( oneEnv.AllZeroes, complex_0 ) );
 }

 if( !t1 ) z.fixReference( &complex_0 );

 z = z.EpsInverse();

 if( t2 ) z.fixReference( &complex_0 );
 else     z.fixReference( &(p->value) );

 delete p;  // ??? OK ???

 if( !t1 ) {
  // A little check
  if( z.standardPart() != complex_0 ) {
   cout << "*** WARNING ***                                              \n"
           "*** WARNING *** JetC::Inverse()                              \n"
           "*** WARNING *** z.standardPart() = " << z.standardPart() << "\n"
           "*** WARNING ***                                              \n"
        << endl;
  }
  z->addTerm( new JLCterm( oneEnv.AllZeroes, jl->refPoint[0] ) );
 }

 z.stacked = 1;
 return z;
}

// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||
// |||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||||

JetC JetC::EpsInverse() const {
 static JetC       z, x, v, g;
 static int        d(1);

 x.setVariable(complex_0, 0);
 g = D(&d);
 
 int i = 0;
 z = x;
 v = ( operator()(&z) - x ) / g(&z);
 while( ( i++ < MX_MAXITER ) &&  ( v != complex_0 ) ) {
  z -= v;
  v = ( operator()(&z) - x ) / g(&z);
 }

 z.stacked = 1;
 return z;
}
